// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "colorbarbase.h"
#include "util.h"
#include "tkpostscript.h"
#include "cbgrid.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "parser.H"

// Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer cbFlexLexer
#include <FlexLexer.h>

YYSTYPE* cblval;
extern int cbparse(ColorbarBase*, cbFlexLexer*);

int cblex(void* vval, cbFlexLexer* ll)
{
  cblval = (YYSTYPE*)vval;
  return ll ? ll->yylex() : 0;
}

void cberror(ColorbarBase* cb, cbFlexLexer* ll, const char* m)
{
  cb->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    cb->error(": ");
    cb->error(cmd);
  }
}

// Public Member Functions

ColorbarBase::ColorbarBase(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : Widget(i, c, item)
{
  // this is needed because of a problem with Tk_ConfigureWidget
  ((ColorbarBaseOptions*)options)->numopts = NULL;

  xmap = NULL;
  colorCells = NULL;
  colorCount = 0;

  numerics = 1;
  grid = NULL;
  gc = NULL;

  cnt = 0;
  lut = NULL;

  invert = 0;
}

ColorbarBase::~ColorbarBase()
{
  if (xmap)
    XDestroyImage(xmap);

  if (colorCells)
    delete [] colorCells;

  if (grid)
    delete grid;

  if (gc)
    XFreeGC(display, gc);

  if (lut)
    delete lut;
}

// Required Virtual Functions

int ColorbarBase::parse(istringstream& istr)
{
  result = TCL_OK;
  cbFlexLexer* ll = new cbFlexLexer(&istr);
  cbparse(this, ll);
  delete ll;

  return result;
}

int ColorbarBase::configure(int argc, const char* argv[], int flags)
{
  if (Widget::configure(argc, argv, flags) == TCL_ERROR)
    return TCL_ERROR;

  // only valid for the initial configuration call
  if (flags != TK_CONFIG_ARGV_ONLY)
    return initColormap();
  else
    return TCL_OK;
}

void ColorbarBase::invalidPixmap()
{
  Widget::invalidPixmap();

  if (xmap)
    XDestroyImage(xmap);
  xmap = NULL;
}

int ColorbarBase::updatePixmap(const BBox& bb)
{
  // create a valid pixmap if needed
  // bb is in canvas coords
  if (!pixmap || !xmap) {
    if (!pixmap) {
      if (!(pixmap = Tk_GetPixmap(display, Tk_WindowId(tkwin), options->width, 
				  options->height, depth))) {
	cerr << "ColorbarBase Internal Error: Unable to Create Pixmap" 
	     << endl;
	exit(1);
      }
      
      // Geometry has changed, redefine our marker GCs including clip regions
      updateGCs();

      XSetForeground(display, gc, getColor("white"));
      XFillRectangle(display, pixmap, gc, 0, 0, 
		     options->width, options->height);
    }

    if (!xmap) {
      if (!(xmap = XGetImage(display, pixmap, 1, 1, options->width -2, 
			     options->height - 
			     ((ColorbarBaseOptions*)options)->numheight -2,
			     AllPlanes, ZPixmap))){
	cerr << "ColorbarBase Internal Error: Unable to Create XImage" 
	     << endl;
	exit(1);
      }
    }
    
    ximageToPixmap();

    if (grid)
      delete grid;
    grid = NULL;

    if (numerics && cnt && lut) {
      grid = new CBGrid(this, cnt, lut, 
			((ColorbarBaseOptions*)options)->numopts);
      grid->render();
    }
  }

  return TCL_OK;
}

int ColorbarBase::postscriptProc(int prepass)
{
  TkPostscriptInfo* psInfo = (TkPostscriptInfo*)(((TkCanvas*)canvas)->psInfo);

  if (!visible)
    return TCL_OK;

  if (prepass)
    return TCL_OK;

  int& width = options->width;
  int& height = options->height;
  int& numheight = ((ColorbarBaseOptions*)options)->numheight;

  float scale = psInfo->scale * psResolution / 72.;
  int ww = (int)((width-1)*scale);
  int hh = (int)((height-numheight)*scale);

  Vector org = psOrigin() + Vector(0,numheight+1);

  // image
  Tcl_AppendResult(interp, "gsave\n", NULL);

  ostringstream str;
  str << org[0] << ' ' << org[1] << " translate" << endl
      << setprecision(12)
      << 1/scale << ' ' << 1/scale << " scale" << endl
      << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);

  switch (psLevel) {
  case 1:
    switch (psColorSpace) {
    case BW:
    case GRAY:
      psLevel1(GRAY, ww, hh, scale);
      break;
    case RGB:
    case CMYK:
      psLevel1(RGB, ww, hh, scale);
      break;
    }
    break;
  case 2:
    psLevel2(psColorSpace, ww, hh, scale);
    break;
  }

  Tcl_AppendResult(interp, "grestore\n", NULL);

  // grid
  Tcl_AppendResult(interp, "gsave\n", NULL);
  {
    Vector ll = Vector(0,0);
    Vector lr = Vector(width,0);
    Vector ur = Vector(width,height);
    Vector ul = Vector(0,height);

    ostringstream str;
    str << "newpath " << endl
	<< ll << "moveto " << endl
	<< lr << "lineto " << endl
	<< ur << "lineto " << endl
	<< ul << "lineto " << endl
	<< ll << "lineto " << endl
      	<< "closepath clip" << endl << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  if (grid)
    grid->ps(psColorSpace, originX, originY);

  Tcl_AppendResult(interp, "grestore\n", NULL);

  return TCL_OK;
}

#ifdef _MACOSX
void ColorbarBase::macosxPrintCmd()
{
  if (!visible)
    return;

  // init
  macosxBegin();

  // clip rect
  Vector v1 = Vector(originX, originY);
  Vector v2 = Vector(options->width, options->height);
  macosxClip(v1,v2);

  int ww = options->width -1;
  int hh = options->height - ((ColorbarBaseOptions*)options)->numheight -1;

  // image
  macosx(1, ww, hh, v1, Vector(ww,hh));

  // grid
  if (grid)
    grid->macosx(originX, originY);

  // end
  macosxEnd();
}
#endif

#ifdef _WIN32
void ColorbarBase::win32PrintCmd()
{
  if (!visible)
    return;

  // init
  win32Begin();

  // clip rect
  Vector v1 = Vector(originX, originY);
  Vector v2 = Vector(options->width, options->height);
  win32Clip(v1,v2);

  int ww = options->width -1;
  int hh = options->height - ((ColorbarBaseOptions*)options)->numheight -1;

  // image
  win32(1, ww, hh, v1, Vector(ww,hh));

  // grid
  if (grid)
    grid->win32(originX, originY);

  // end
  win32End();
}
#endif

void ColorbarBase::updateGCs()
{
  // widget clip region
  XRectangle rectWidget[1];  // clip rect
  rectWidget[0].x = 0;
  rectWidget[0].y = 0;
  rectWidget[0].width = options->width;
  rectWidget[0].height = options->height;

  if (!gc)
    gc = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);
  XSetClipRectangles(display, gc, 0, 0, rectWidget, 1, Unsorted);
}

void ColorbarBase::setColormapLevelCmd(int cc, float* ff)
{
  // check for the same
  if (cnt == cc) {
    int same = 1;
    for (int ii=0; ii<cc; ii++)
      same &= (lut[ii] == ff[ii]);
    if (same)
      return;
  }

  // remove the old
  if (lut)
    delete lut;
  lut = NULL;
  cnt = 0;

  invalidPixmap();
  redraw();

  // now check for repeat values, including zeros
  float* gg = new float[cc];
  int dd = 1;
  gg[0] = ff[0];
  for (int ii=1; ii<cc; ii++)
    if (gg[dd-1]!=ff[ii])
      gg[dd++] = ff[ii];

  // we need at least 3
  if (dd<3)
    return;

  // we have new data
  cnt = dd;
  lut = new double[cnt];
  for (int ii=0; ii<cnt; ii++)
    lut[ii] = gg[ii];

  delete [] gg;
}

void ColorbarBase::getInvertCmd()
{
  if (invert)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void ColorbarBase::invertCmd(int ii)
{
  invert = ii ? 1 : 0;
  updateColors();
}

void ColorbarBase::numericsCmd(int nn)
{
  numerics = nn ? 1 : 0;
  redraw();
}

void ColorbarBase::getNumericsCmd()
{
  if (numerics)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}
