// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <sstream>
#include <iomanip>
using namespace std;

// Note: currently, hcompress will not work with float or double data with 
// BLANK defined, due to the fact that the decompress code takes a int() and 
// not an unsigned int()

#include "hcompress.h"
#include "util.h"

extern "C" {
  int dodecode(unsigned char**, int*, int, int, unsigned char*);
  void undigitize(int*, int, int, int);
  int hinv(int*, int, int, int, int);
}

template<class T> FitsHcompressm<T>::FitsHcompressm(FitsFile* fits) 
  : FitsCompressm<T>(fits)
{
  // hcompress parameters
  scale_ = 1;
  smooth_ = 0;
  char name[] = "ZNAME ";
  char val[] = "ZVAL ";
  for (int ii=0; ii<9; ii++) {
    name[5] = '0'+ii;
    val[4] = '0'+ii;
    if (fits->find(name)) {
      char* which = fits->getString(name);
      if (!strncmp(which,"SCALE",4))
	scale_ = fits->getInteger(val,32);
      else if (!strncmp(which,"SMOOTH",4))
	smooth_ = fits->getInteger(val,4);
    }
  }

  FitsCompressm<T>::uncompress(fits);
}

template <class T> int FitsHcompressm<T>::compressed(T* dest, char* sptr, 
						     char* heap,
						     int kkstart, int kkstop, 
						     int jjstart, int jjstop, 
						     int iistart, int iistop)
{
  double zs = FitsCompressm<T>::bscale_;
  if (FitsCompressm<T>::zscale_)
    zs = FitsCompressm<T>::zscale_->value(sptr,0);

  double zz = FitsCompressm<T>::bzero_;
  if (FitsCompressm<T>::zzero_)
    zz = FitsCompressm<T>::zzero_->value(sptr,0);

  int blank = FitsCompressm<T>::blank_;
  if (FitsCompressm<T>::zblank_)
    blank = (int)FitsCompressm<T>::zblank_->value(sptr,0);

  int icnt=0;
  unsigned char* ibuf = (unsigned char*)((FitsBinColumnArray*)FitsCompressm<T>::compress_)->get(heap, sptr, &icnt);

  // ibuf can be NULL
  if (ibuf && icnt>0) {
    int ocnt = FitsCompressm<T>::ww_ * FitsCompressm<T>::hh_ * FitsCompressm<T>::dd_;
    int obuf[ocnt];

    unsigned char* ptr = ibuf;
    // magic numbers
    if (*ptr++ != 0xDD || *ptr++ != 0x99) {
      if (DebugCompress)
	cerr << "HCOMPRESS Internal Error: bad magic numbers" << endl;
      return 0;
    }
 
    // size,scale,checksum,nbitplanes
    int nx = readint(ptr);
    ptr += sizeof(int);
    int ny = readint(ptr);
    ptr += sizeof(int); // ny
    int scale = readint(ptr);
    ptr += sizeof(int); // scale
    long long sum = readlonglong(ptr);
    ptr += sizeof(long long); // sum
 
    // nbitplanes
    unsigned char nbp[3];
    nbp[0] = *ptr++;
    nbp[1] = *ptr++;
    nbp[2] = *ptr++;

    // decode
    if (!dodecode(&ptr, obuf, nx, ny, nbp)) {
      if (DebugCompress)
	cerr << "HCOMPRESS Internal Error: bad inflate result" << endl;
      return 0;
    }
    obuf[0] = (int)sum;
 
    // undigitize
    undigitize(obuf, nx, ny, scale);

    // hinv
    if (!hinv(obuf, nx, ny, smooth_, scale)) {
      if (DebugCompress)
	cerr << "HCOMPRESS Internal Error: bad transform" << endl;
      return 0;
    }

    int ll=0;
    for (int kk=kkstart; kk<kkstop; kk++)
      for (int jj=jjstart; jj<jjstop; jj++)
	for (int ii=iistart; ii<iistop; ii++,ll++)
	  dest[kk*FitsCompressm<T>::width_*FitsCompressm<T>::height_ + jj*FitsCompressm<T>::width_ + ii] = getValue(obuf+ll,zs,zz,blank);
  }

  return 1;
}

template <class T> int FitsHcompressm<T>::readint(unsigned char* ptr)
{
  int rr;
  if (FitsCompressm<T>::byteswap_)
    swap4((char*)ptr,(char*)(&rr));
  else {
    // do it this way cause solaris has issues with returning the 
    // dereferenced ptr
    // return *((int*)ptr);
    
    char* dest = (char*)&rr;
    char* src  = (char*)ptr;

    *(dest  ) = *(src  );
    *(dest+1) = *(src+1);
    *(dest+2) = *(src+2);
    *(dest+3) = *(src+3);
  }
  return rr;
}

template <class T> long long FitsHcompressm<T>::readlonglong(unsigned char* ptr)
{
  long long rr;
  if (FitsCompressm<T>::byteswap_)
    swap8((char*)ptr,(char*)(&rr));
  else {
    // do it this way cause solaris has issues with returning the 
    // dereferenced ptr
    // return *((long long*)ptr);

    char* dest = (char*)&rr;
    char* src  = (char*)ptr;

    *(dest  ) = *(src  );
    *(dest+1) = *(src+1);
    *(dest+2) = *(src+2);
    *(dest+3) = *(src+3);
    *(dest+4) = *(src+4);
    *(dest+5) = *(src+5);
    *(dest+6) = *(src+6);
    *(dest+7) = *(src+7);
  }
  return rr;
}

template class FitsHcompressm<unsigned char>;
template class FitsHcompressm<short>;
template class FitsHcompressm<unsigned short>;
template class FitsHcompressm<int>;
template class FitsHcompressm<long long>;
template class FitsHcompressm<float>;
template class FitsHcompressm<double>;
