// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "basebox.h"
#include "framebase.h"
#include "fitsimage.h"
#include "util.h"

BaseBox::BaseBox(const BaseBox& a) : BaseMarker(a) {}

BaseBox::BaseBox(FrameBase* p, const Vector& ctr, 
		 double ang,
		 const char* clr, int w, const char* f, 
		 const char* t, unsigned short prop, const char* c,
		 const List<Tag>& tag, const List<CallBack>& cb)
  : BaseMarker(p, ctr, ang, 5, clr, w, f, t, prop, c, tag, cb) {}

int BaseBox::isIn(const Vector& v)
{
  // v is in canvas coords
  Vector s = annuli_[numAnnuli_-1];

  // zero radius
  if (!s[0] || !s[1])
    return 0;

  Matrix mm = bckRefMatrix();
  Vector p = v * mm;

  if (p[0]<-s[0]/2 || p[0]>s[0]/2 || p[1]<-s[1]/2 || p[1]>s[1]/2)
    return 0;
  else
    return 1;
}

// protected

void BaseBox::updateHandles()
{
  // handles are in canvas coords
  // we can't garantee that the radii have been sorted yet
  // also, annuli[i] could be negative

  if (handle)
    delete [] handle;
  handle = new Vector[numHandle];

  Vector max;
  for(int i=0; i<numAnnuli_; i++) {
    Vector foo = annuli_[i].abs();
    if (max[0]<foo[0])
      max = foo;
  }
  Vector s = max;

  Matrix mm = fwdCanvasMatrix();

  handle[0] = Vector(-s[0],-s[1])/2 * mm;
  handle[1] = Vector( s[0],-s[1])/2 * mm;
  handle[2] = Vector( s[0], s[1])/2 * mm;
  handle[3] = Vector(-s[0], s[1])/2 * mm;

  // the rest are radii
  for (int i=4; i<numHandle; i++)
    handle[i] = Vector((annuli_[i-4][0])/2.,0) * mm;
}

Vector BaseBox::intersect(Vector rr, double ang)
{
  // special cases
  if (!rr[0] || !rr[1])
    return Vector();

  double phi = rr.angle();

  if (ang >= 0 && ang < phi)
    return Vector(rr[0],-rr[0]*tan(ang));
  else if (ang >= phi && ang < M_PI-phi)
    return Vector(rr[1]/tan(ang),-rr[1]);
  else if (ang >= M_PI-phi && ang < M_PI+phi)
    return Vector(-rr[0],rr[0]*tan(ang));
  else if (ang >= M_PI+phi && ang < 2*M_PI-phi)
    return Vector(-rr[1]/tan(ang),rr[1]);
  else
    return Vector(rr[0],-rr[0]*tan(ang));
}

void BaseBox::newVertices()
{
  if (vertices_)
    deleteVertices();

  if (startAng_ == 0 && stopAng_ == 360)
    newVerticesA();
  else
    newVerticesB();
}

void BaseBox::newVerticesA()
{
  Matrix mm = fwdRefMatrix();

  numPoints_ = 5;
  vertices_ = new Vector*[numAnnuli_+1];
  for (int i=0; i<numAnnuli_; i++)
    vertices_[i]= new Vector[numPoints_];
  vertices_[numAnnuli_] = new Vector[2];

  for (int i=0; i<numAnnuli_; i++) {
    vertices_[i][0] = Vector(-annuli_[i][0]/2.,-annuli_[i][1]/2.) * mm;
    vertices_[i][1] = Vector( annuli_[i][0]/2.,-annuli_[i][1]/2.) * mm;
    vertices_[i][2] = Vector( annuli_[i][0]/2., annuli_[i][1]/2.) * mm;
    vertices_[i][3] = Vector(-annuli_[i][0]/2., annuli_[i][1]/2.) * mm;
    vertices_[i][4] = Vector(-annuli_[i][0]/2.,-annuli_[i][1]/2.) * mm;
  }

  vertices_[numAnnuli_][0] = vertices_[numAnnuli_-1][0];
  vertices_[numAnnuli_][1] = vertices_[numAnnuli_-1][2];
}

void BaseBox::newVerticesB()
{
  numPoints_ = 0; // we will increment later
  vertices_ = new Vector*[numAnnuli_+1];
  for (int ii=0; ii<numAnnuli_; ii++)
    vertices_[ii]= new Vector[7];
  vertices_[numAnnuli_] = new Vector[2];

  double a1 = degToRad(startAng_);
  double a2 = degToRad(stopAng_);

  if (a1>2*M_PI-FLT_EPSILON)
    a1 -= 2*M_PI;
  if (a2>2*M_PI-FLT_EPSILON)
    a2 -= 2*M_PI;
  if (a2<=a1+FLT_EPSILON)
    a2 += 2*M_PI;
  
  int cnt = 0;
  for (int ii=0; ii<numAnnuli_; ii++) {
    Vector rr = (annuli_[ii]).abs();
    if (rr[0] && rr[1]) {
      double phi = rr.angle();
      int s1 =0;
      int s2 =0;
      cnt = 0;

      vertBTest(&s1, &s2, a1, a2, 0, phi, ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, phi, M_PI-phi, ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, M_PI-phi, M_PI+phi, ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, M_PI+phi, 2*M_PI-phi, ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, 2*M_PI-phi,2*M_PI , ii, &cnt);

      vertBTest(&s1, &s2, a1, a2, 2*M_PI,2*M_PI+phi , ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, 2*M_PI+phi,3*M_PI-phi , ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, 3*M_PI-phi,3*M_PI+phi , ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, 3*M_PI+phi,4*M_PI-phi , ii, &cnt);
      vertBTest(&s1, &s2, a1, a2, 4*M_PI-phi,4*M_PI, ii, &cnt);
    }
  }

  // final number of segments
  numPoints_ = ++cnt;

  // include/exclude
  Matrix mm = fwdRefMatrix();
  vertices_[numAnnuli_][0] = 
    Vector(-annuli_[numAnnuli_-1][0]/2.,-annuli_[numAnnuli_-1][1]/2.) * mm;
  vertices_[numAnnuli_][1] = 
    Vector( annuli_[numAnnuli_-1][0]/2., annuli_[numAnnuli_-1][1]/2.) * mm;
}

void BaseBox::vertBTest(int* s1, int* s2, double a1, double a2, 
			double b1, double b2, int ii, int* cnt)
{
  if ((!(*s1)) && (a1>=b1-FLT_EPSILON) && (a1<b2+FLT_EPSILON))
    *s1 =1;
  if ((!(*s2)) && (a2>b1-FLT_EPSILON) && (a2<=b2+FLT_EPSILON))
    *s2 =1;

  if (((*s1) && (!(*s2))) || ((*s1) && (*s2)))
    vertBPrep(a1,a2,b1,b2,ii,cnt);

  if (*s1&&*s2)
    *s1=*s2=0;
}

void BaseBox::vertBPrep(double a1, double a2, double ll, double ul, 
			int ii, int* cnt)
{
  if (!(a1 >= ll && a1 <= ul))
    a1 = ll;
  if (!(a2 >= ll && a2 <= ul))
    a2 = ul;

  if (a1>a2) {
    vertBSeg(ll,a2,ii,cnt);
    vertBSeg(a1,ul,ii,cnt);
  }
  else
    vertBSeg(a1,a2,ii,cnt);
}

void BaseBox::vertBSeg(double ang1, double ang2, int ii, int* cnt)
{
  Vector rr = (annuli_[ii]/2).abs();
  Matrix mm = fwdRefMatrix();

  if (ang1>2*M_PI-FLT_EPSILON)
    ang1 -= 2*M_PI;
  if (ang2>2*M_PI-FLT_EPSILON)
    ang2 -= 2*M_PI;

  vertices_[ii][(*cnt)++] = intersect(rr,ang1)*mm;
  vertices_[ii][(*cnt)] = intersect(rr,ang2)*mm;
}
