// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "contour.h"
#include "framebase.h"
#include "util.h"

Contour::Contour(FrameBase* p, const char* c, int w, int m) 
  : parent(p), lineWidth(w)
{
  colorName = dupstr(c);
  color = parent->getColor(colorName);
  scale = NULL;
  method = (Method)m;

  previous_ = NULL;
  next_ = NULL;
}

Contour::Contour(FrameBase* p, const char* c, int w, const List<Vertex>& cont) 
  : parent(p), lineWidth(w)
{
  contours_ = cont;
  colorName = dupstr(c);
  color = parent->getColor(colorName);
  scale = NULL;
  method = BLOCK;

  previous_ = NULL;
  next_ = NULL;
}

Contour::~Contour()
{
  if (colorName)
    delete [] colorName;

  if (scale)
    delete scale;
}

void Contour::render(Pixmap pmap, const Matrix& mx, const BBox& bb)
{
  // bb is in ref coords

  if (contours_.head()) {
    XSetForeground(parent->display, parent->contourGC, color);
    int l = lineWidth>=1 ? lineWidth : 1;
    XSetLineAttributes(parent->display, parent->contourGC, 
		       l, LineSolid, CapButt, JoinMiter);

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = (u1 * mx).round();
	Vector v2 = (u2 * mx).round();
	BBox b = bb*mx;
	if (b.isIn(v1) || b.isIn(v2))
	  XDrawLine(parent->display, pmap, parent->contourGC, 
		    (int)v1[0], (int)v1[1], (int)v2[0], (int)v2[1]);
      }

      u1 = u2;
    }
  }
}

void Contour::ps(int mode)
{
  if (contours_.head()) {
    {
      ostringstream str;
      switch ((Widget::PSColorSpace)mode) {
      case Widget::BW:
      case Widget::GRAY:
	psColorGray(parent->getXColor(colorName), str);
	str << " setgray";
	break;
      case Widget::RGB:
	psColorRGB(parent->getXColor(colorName), str);
	str << " setrgbcolor";
	break;
      case Widget::CMYK:
	psColorCMYK(parent->getXColor(colorName), str);
	str << " setcmykcolor";
	break;
      }
      str << endl << ends;
      Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
    }  
  
    {
      ostringstream str;
      str << lineWidth << " setlinewidth" << endl << ends;
      Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
    }

    // output stroke paths

    int first = 1;
    do {
    ostringstream str;
      Vector u = contours_.current()->vector;
      if (u[0] != DBL_MAX) {
	Vector v = u * parent->refToCanvas;
	if (first) {
	  str << "newpath " << endl
	      << v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
	  first = 0;
	}
	else
	  str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;
      }
      else {
	str << "stroke" << endl << ends;
	first = 1;
      }
      Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
    } while (contours_.next());
  }
}

#ifdef _MACOSX
void Contour::macosx()
{
  if (contours_.head()) {
    // color and width
    macosxColor(parent->getXColor(colorName));
    macosxWidth(lineWidth);

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = u1 * parent->refToCanvas;
	Vector v2 = u2 * parent->refToCanvas;
	
	macosxDrawLine(v1,v2);
      }

      u1 = u2;
    }
  }
}
#endif

#ifdef _WIN32
void Contour::win32()
{
  if (contours_.head()) {
    // color and width
    win32Color(parent->getXColor(colorName));
    win32Width(lineWidth);

    Vector u1 = contours_.current()->vector;
    while (contours_.next()) {
      Vector u2 = contours_.current()->vector;
      if (u1[0] != DBL_MAX && u2[0] != DBL_MAX) {
	Vector v1 = u1 * parent->refToCanvas;
	Vector v2 = u2 * parent->refToCanvas;
	
	win32DrawLine(v1,v2);
      }

      u1 = u2;
    }
  }
}
#endif

void Contour::updateCoords(const Matrix& mx)
{
  if (contours_.head())
    do {
      Vector& v = contours_.current()->vector;
      if (v[0] != DBL_MAX && v[0] != DBL_MAX)
	v *= mx;
    } while (contours_.next());
}

void Contour::setLineWidth(int w)
{
  lineWidth = w;
}

void Contour::setColor(const char* clr)
{
  if (colorName)
    delete [] colorName;

  colorName = dupstr(clr);
  color = parent->getColor(colorName);
}

