// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "frame.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"
#include "contour.h"

// Frame Member Functions

Frame::Frame(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item) 
  : FrameBase(i,c,item)
{
  fits = NULL;

  mosaicCount = 0;
  sliceCount = 0;
  mosaicType = NOMOSAIC;
  mosaicSystem = WCS;
  contour = NULL;
  
  keyFits = &fits;
  channelFits = &fits;
  currentFits = NULL;
  
  currentMosaicCount = &mosaicCount;
  currentSliceCount = &sliceCount;
  currentScale = &frScale;
  currentMosaic = &mosaicType;
  currentMosaicSystem = &mosaicSystem;

  currentcontour = &contour;

  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  colorCount = 0;
  colorScale = NULL;
  colorCells = NULL;
  indexCells = NULL;

  iisSaveRoot = 0;
  iisInteractive = 0;
}

Frame::~Frame()
{
  if (colorScale)
    delete colorScale;

  if (colorCells)
    delete [] colorCells;

  if (indexCells)
    delete [] indexCells;
}

// Virtual Routines with base components

void Frame::reset()
{
  cmapID = 1;
  bias = 0.5;
  contrast = 1.0;

  FrameBase::reset();
}

void Frame::setBinCursor(const Vector& v)
{
  if (currentFits)
    currentFits->setBinCursor(cursor);
}

void Frame::updateBin(const Matrix& mx)
{
  // Note: cursor is in USER coords, imageCenter() in IMAGE coords
  cursor = imageCenter();

  // update bin
  crosshair *= mx;

  updateMarkerCoords(&userMarkers, mx);
  updateMarkerCoords(&catalogMarkers, mx);

  // update contours
  // main contour
  if (hasContour())
    (*currentcontour)->updateCoords(mx);

  // aux contours
  {
    Contour* ptr=auxcontours.head();
    while (ptr) {
      ptr->updateCoords(mx);
      ptr=ptr->next();
    }
  }

  // delete any previous slices
  {
    FitsImage* ptr = (*channelFits)->nextSlice();
    (*channelFits)->setNextSlice(NULL);
    while (ptr) {
      FitsImage* tmp = ptr->nextSlice();
      delete ptr;
      ptr = tmp;
    }
  }

  // finish bin
  currentFits = *channelFits;

  *currentMosaicCount = 1;
  *currentSliceCount = 1;
  slice = 1;

  // bin data cube
  {
    int bd = (*channelFits)->binDepth();
    if (bd > 1) {
      FitsImage* ptr = *channelFits;
      for (int i=1; i<bd; i++) {
	FitsImage* next = NULL;
	next = new FitsImageFitsNextHist(this, *channelFits, ptr->imageFile());
	if (next && next->isValid()) {
	  ptr->setNextSlice(next);
	  ptr = next;
	  (*currentSliceCount)++;
	}
	else {
	  if (next)
	    delete next;
	  break;
	}
      }
    }
  }

  updateBinFileNames();

  resetClip(0);
  alignWCS();
  updateNow(MATRIX); // because we have changed zoom

  // update markers call backs
  // wait til matrices have been updated so that any dialogs will print
  // the correct coords

  updateMarkerCBs(&userMarkers);
  updateMarkerCBs(&catalogMarkers);
}

BBox Frame::imageBBox()
{
  // returns imageBBox in IMAGE coords
  //   and extends edge to edge

  BBox rr;
  FitsImage* ptr = fits;
  while (ptr) {
    int* params = ptr->getDataParams(frScale.scanMode());
    int& xmin = params[1];
    int& xmax = params[2];
    int& ymin = params[3];
    int& ymax = params[4];
    Matrix mm = ptr->getWCSToRef() * dataToImage;

    Vector aa =  Vector(xmin,ymin) * mm;
    if (ptr == fits)
      rr = BBox(aa,aa);
    else
      rr.bound(aa);

    rr.bound(Vector(xmax,ymin) * mm);
    rr.bound(Vector(xmax,ymax) * mm);
    rr.bound(Vector(xmin,ymax) * mm);

    ptr = ptr->nextMosaic();
  }

  return rr;
}

void Frame::unload()
{
  unloadFits();
}

void Frame::unloadFits()
{
  if (DebugPerf)
    cerr << "Frame::unloadFits" << endl;

  // clean up from iis if needed
  if (isIIS()) {
    frScale.setClipMode(FrScale::MINMAX);
    frScale.setMMMode(FrScale::SAMPLE);
    frScale.setULow(DEFAULTLOW);
    frScale.setUHigh(DEFAULTHIGH);
    setColorScale(FrScale::LINEARSCALE);
  }

  // reset user mode
  // if (frScale.clipMode() == FrScale::USERCLIP)
  // frScale.setClipMode(FrScale::MINMAX);

  // now, delete any fits
  {
    FitsImage* ptr = fits;
    while (ptr) {
      FitsImage* sptr = ptr->nextSlice();
      while (sptr) {
	FitsImage* stmp = sptr->nextSlice();
	delete sptr;
	sptr = stmp;
      }

      FitsImage* tmp = ptr->nextMosaic();
      delete ptr;
      ptr = tmp;
    }
    fits = NULL;
  }

  // delete any contour
  if (contour)
    delete contour;
  contour = NULL;

  FrameBase::unloadFits();
}

void Frame::alignWCS()
{
  if (!wcsAlignment || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation);
}

void Frame::alignWCS(FitsImage* ptr)
{
  if (!wcsAlignment || !ptr || !currentFits || !hasWCS(wcsSystem)) {
    wcsOrientation = NORMAL;
    wcsOrientationMatrix.identity();
    wcsRotation = 0;
    return;
  }

  calcAlignWCS(ptr, currentFits, wcsSystem, wcsSky,
	       &wcsOrientation, &wcsOrientationMatrix, &wcsRotation, &zoom_);
}

void Frame::updateMatrices()
{
  // very important! do this first
  FrameBase::updateMatrices();

  Matrix rgb; // alway identity

  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMatrices(rgb, refToUser, userToWidget, 
			   widgetToCanvas, userToPanner);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  // now any masks
  FitsMask* fm = mask.tail();
  while(fm) {
    FitsImage* ptr = fm->mask();
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->updateMatrices(rgb, refToUser, userToWidget, 
			     widgetToCanvas, userToPanner);
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }

    fm = fm->previous();
  }
}

void Frame::updateMagnifierMatrices(Matrix& refToMagnifier)
{
  FitsImage* ptr = fits;
  while (ptr) {
    FitsImage* sptr = ptr;
    while (sptr) {
      sptr->updateMagnifierMatrices(refToMagnifier);
      sptr = sptr->nextSlice();
    }
    ptr = ptr->nextMosaic();
  }

  // now any masks
  FitsMask* fm = mask.tail();
  while(fm) {
    FitsImage* ptr = fm->mask();
    while (ptr) {
      FitsImage* sptr = ptr;
      while (sptr) {
	sptr->updateMagnifierMatrices(refToMagnifier);
	sptr = sptr->nextSlice();
      }
      ptr = ptr->nextMosaic();
    }
    fm = fm->previous();
  }
}

void Frame::updateColorCells(unsigned short* index, 
			     unsigned char* cells, int cnt)
{
  // the colorbar widget will pass us a pointer to the indexCells

  colorCount = cnt;
  if (indexCells)
    delete [] indexCells;
  indexCells = new unsigned short[cnt];
  if (!indexCells) {
    cerr << "Frame Internal Error: Unable to Alloc indexCells" << endl;
    exit(1);
  }
  memcpy(indexCells, index, cnt*sizeof(unsigned short));

  // copy the rgb vales to the colorCells array (for postscript printing)

  if (colorCells)
    delete [] colorCells;
  colorCells = new unsigned char[cnt*3];
  if (!colorCells) {
    cerr << "Frame Internal Error: Unable to Alloc colorCells" << endl;
    exit(1);
  }
  memcpy(colorCells, cells, cnt*3);
}

int Frame::isIIS() 
{
  return currentFits && currentFits->isIIS();
}

void Frame::psLevel1(PSColorSpace mode, int width, int height, float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel1Head(mode, width, height);

  if (!isMosaic())
    psLevel1Single(mode, width, height, scale);
  else
    psLevel1Mosaic(mode, width, height, scale);
}

void Frame::psLevel1Single(PSColorSpace mode, int width, int height, 
			   float scale)
{
  if (!currentFits)
    return;

  NoCompressAsciiHex filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  {
    currentFits->updatePS(psMatrix(scale, width, height));

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* currentMask = fm->current(); 
      if (currentMask)
	currentMask->updatePS(psMatrix(scale, width, height));
      fm = fm->previous();
    }
  }

  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;
	    
      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current();
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	  }
	  else {
	    blue = nanColor->blue;
	    green = nanColor->green;
	    red = nanColor->red;
	  }
	} 
	else {
	  blue = maskColor->blue;
	  green = maskColor->green;
	  red = maskColor->red;
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame::psLevel1Mosaic(PSColorSpace mode, int width, int height,
			   float scale)
{
  NoCompressAsciiHex filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  {
    Matrix mx = psMatrix(scale, width, height);

    FitsImage* sptr = currentFits;
    while (sptr) {
      sptr->updatePS(mx);
      sptr = sptr->nextMosaic();
    }

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* mptr = fm->current(); 
      while (mptr) {
	mptr->updatePS(mx);
	mptr = mptr->nextMosaic();
      }
      fm = fm->previous();
    }
  }

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;
	    
      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }
      
      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }
       
	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  } 
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
      case CMYK:
	filter << red << green << blue;
	break;
      }
      str << filter;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame::buildXImage(XImage* xmap, int x0, int y0, int x1, int y1,
			double* (FitsImage::*getMatrix)())
{
  if (!isMosaic())
    fillXImageSingle(xmap, x0, y0, x1, y1, getMatrix);
  else
    fillXImageMosaic(xmap, x0, y0, x1, y1, getMatrix);
}

// Fits Commands

void Frame::sliceCmd(int s)
{
  if (fits && (s>0) && (s<=sliceCount)) {
    slice = s;

    currentFits = fits;
    FitsMask* fm=mask.tail();
    while (fm) {
      fm->initCurrent();
      fm = fm->previous();
    }

    for (int i=1; i<s; i++) {
      currentFits = currentFits->nextSlice();

      FitsMask* fm=mask.tail();
      while (fm) {
	fm->nextSlice();
	fm = fm->previous();
      }
    }
  }

  // execute any update callbacks
  updateCBMarkers();

  update(BASE);
}

void Frame::getWCSAlignCmd()
{
  ostringstream str;
  str << (unsigned short*)*channelFits << ends;

  Tcl_AppendResult(interp, (wcsAlignment ? "1" : "0"), " ", 
		   coordSystemStr_[wcsSystem-IMAGE], " ",
		   skyFrameStr_[wcsSky-FK4], " ",
		   str.str().c_str(), NULL);
}

// Colormap Commands

void Frame::getColormapCmd()
{
  ostringstream str;
  str << cmapID << ' ' << bias << ' ' << contrast << ' ' << invert << ends;
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void Frame::getTypeCmd()
{
  Tcl_AppendResult(interp, "base", NULL);
}

// IIS Commands

void Frame::iisCmd(int width, int height)
{
  unloadAllFits();

  // force clip mode to user between 1 to 200
  frScale.setClipMode(FrScale::USERCLIP);
  frScale.setMMMode(FrScale::SCAN);
  frScale.setULow(IISMIN);
  frScale.setUHigh(IISMAX);
  setColorScale(FrScale::IISSCALE);

  FitsImage* img = new FitsImageIIS(this, width, height);

  load(ALLOC,"",img,IMG);
}

void Frame::iisEraseCmd()
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisErase();
}

void Frame::iisGetCmd(char* ptr, int x, int y, int dx, int dy)
{
  if (currentFits)
    memcpy(ptr, ((FitsImageIIS*)currentFits)->iisGet(x,y), dx*dy);
}

void Frame::iisSetCmd(const char* data, int x, int y, int dx, int dy)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisSet(x, y, data, dx*dy);
}

void Frame::iisWCSCmd(const Matrix& mx, const Vector& z, int zt)
{
  if (currentFits)
    ((FitsImageIIS*)currentFits)->iisWCS(mx, z, zt);
}

// RGB Commands

void Frame::getRGBChannelCmd()
{
  Tcl_AppendResult(interp, "red", NULL);
}

void Frame::getRGBViewCmd()
{
  Tcl_AppendResult(interp, "1 1 1", NULL);
}

void Frame::getRGBSystemCmd()
{
  Tcl_AppendResult(interp, "image", NULL);
}


