// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#ifndef __framebase_h__
#define __framebase_h__

#include "widget.h"

#include "inversescale.h"
#include "marker.h"
#include "hist.h"
#include "list.h"
#include "coord.h"
#include "grid.h"
#include "frscale.h"
#include "fitsmask.h"

#define SCALESIZE 16384
#define HISTEQUSIZE 16384

// We support 28 different WCS's. 'WCS', 'WCSA'-'WCSZ', and internal 'WCS0'
//#define MULTWCS 27
#define MULTWCS 28

#define DEFAULTLOW 1
#define DEFAULTHIGH 100

extern int DebugAST;
extern int DebugMosaic;
extern int DebugPerf;
extern int DebugWCS;
extern int DebugBin;
extern int DebugCompress;
extern int DebugGZ;
extern int DebugRGB;

class FitsImage;
class Contour;
class FVContour;
class Grid;
class Attribute;
class Marker;
class BaseBox;
class BaseEllipse;
class BaseMarker;
class Composite;

class Annulus;
class Box;
class BoxAnnulus;
class Bpanda;
class Circle;
class Compass;
class Cpanda;
class Ellipse;
class EllipseAnnulus;
class Epanda;
class Line;
class Point;
class CirclePoint;
class BoxPoint;
class DiamondPoint;
class CrossPoint;
class ExPoint;
class ArrowPoint;
class BoxCirclePoint;
class Projection;
class Polygon;
class Ruler;
class Text;
class Vect;

// Frame

class FrameBase : public Widget {
  friend class FitsImage;
  friend class Contour;
  friend class FVContour;
  friend class Grid;
  friend class Attribute;
  friend class Marker;
  friend class BaseBox;
  friend class BaseEllipse;
  friend class BaseMarker;
  friend class Composite;

  friend class Annulus;
  friend class Box;
  friend class BoxAnnulus;
  friend class Bpanda;
  friend class Circle;
  friend class Compass;
  friend class Cpanda;
  friend class Ellipse;
  friend class EllipseAnnulus;
  friend class Epanda;
  friend class Line;
  friend class Point;
  friend class CirclePoint;
  friend class BoxPoint;
  friend class DiamondPoint;
  friend class CrossPoint;
  friend class ExPoint;
  friend class ArrowPoint;
  friend class BoxCirclePoint;
  friend class Projection;
  friend class Polygon;
  friend class Ruler;
  friend class Text;
  friend class Vect;

public:
  enum MemType {ALLOC, ALLOCGZ, CHANNEL, MMAP, SMMAP, MMAPINCR, 
		SHARE, SSHARE, SOCKET, SOCKETGZ, VAR, HIST, COMPRESS};
  enum MarkerFormat {DS9, CIAO, SAOTNG, SAOIMAGE, PROS, RAWXY};
  enum MarkerLayer {USER, TEMPLATE, CATALOG};
  enum FileNameType {ROOTBASE, FULLBASE, ROOT, FULL};
  enum ShmType {SHMID,KEY};
  enum ColormapMode {CENTER, CLICK};
  enum UndoMarkerType {NONE, MOVE, EDIT, DELETE};
  enum DisplayMode {SINGLE, MOSAIC};
  enum LoadMethod {LOADALL, INCR};
  enum MosaicType {NOMOSAIC, IRAF, WCSMOSAIC, WFPC2};
  enum CompressType {NOCOMPRESS, GZ};
  enum LayerType {IMG, MASK};

protected:
  enum UpdateType {MATRIX, BASE, PIXMAP, NOUPDATE};

  FitsImage** keyFits;       // primary fits channel
  FitsImage** channelFits;   // current fits channel
  FitsImage* currentFits;    // current fits image

  List <FitsMask> mask;

  int* currentMosaicCount;         // current count of fits images
  int* currentSliceCount;         // current count of fits images

  FrScale* currentScale;     // current scale parameters
  MosaicType* currentMosaic; // mosaic mode
  CoordSystem* currentMosaicSystem; // mosaic wcs system

  Matrix dataToImage;
  Matrix imageToData;

  int slice;                 // current slice

  XImage* baseXImage;        // ximage of image
  Pixmap basePixmap;         // pixmap of image
  UpdateType needsUpdate;    // updateFlags

  Pixmap panPM;              // tmp pixmap for pan motion
  Vector panCursor;          // tmp cursor for interactive panning
  GC panGCXOR;               // GC for interactive panning
  Pixmap pannerPixmap;       // pixmap for panner
  XImage* pannerXImage;      // ximage for panner
  int pannerWidth;           // panner width
  int pannerHeight;          // panner height
  char pannerName[32];       // panner widget name
  CoordSystem pannerSystem;  // panner wcs coord system
  SkyFrame pannerSky;        // panner wcs sky frame
  int usePanner;             // panner status

  Pixmap magnifierPixmap;    // pixmap for magnifier
  XImage* magnifierXImage;   // ximage for magnifier
  int magnifierWidth;        // magnifier width
  int magnifierHeight;       // magnifier height
  double magnifierZoom_;     // magnifier zoom factor
  Vector magnifierCursor;    // we need to save the last cursor used
  char magnifierName[32];    // magnifer widget name
  int useMagnifier;          // magnifer status
  int useMagnifierGraphics;  // magnifer marker status
  int useMagnifierCursor;    // magnifier cursor status
  GC magnifierGC;            // GC for magnifier
				  
  int invert;                // current colormap invert

  Vector cursor;             // current cursor position in USER coords

  Orientation orientation;   // current image orientation
  Orientation irafOrientation; // orientation of first iraf segment
  Matrix orientationMatrix;  // current image orientation matrix
  Vector zoom_;              // current image zoom value
  double rotation;           // current image rotation angle

  double rotateRotation;     // tmp cursor for interactive rotation
  XImage* rotateSrcXM;       // rotate src ximage
  XImage* rotateDestXM;      // rotate dest ximage
  Pixmap rotatePM;           // rotate pixmap
  GC rotateGCXOR;            // GC for interactive rotation

  int preservePan;           // flag to indicate preserve between loads

  int wcsAlignment;          // flag to indicate wcs alignment
  CoordSystem wcsSystem;     // wcs system
  SkyFrame wcsSky;           // wcs sky frame
  Orientation wcsOrientation; // wcs image orientation
  Matrix wcsOrientationMatrix; // wcs image orientation matrix
  double wcsRotation;        // wcs image rotation angle

  char* contourLevel;
  Contour** currentcontour;  // main contour
  List<Contour> auxcontours; // aux contours
  GC contourGC;

  Grid* grid;
  GC gridGC;

  int doSmooth_;
  int smoothFunction_;
  int smoothRadius_;

  FitsHist::Function binFunction_;
  Vector binFactor_;
  int binBufferSize_;
  int binDepth_;

  int cmapArea;              // current colormap update size
  ColormapMode cmapMode;     // current colormap update mode
  Vector cmapOffset;         // current colormap update offset

  Vector crosshair;          // crosshair location in REF coords
  int useCrosshair;          // flag to use crosshair
  GC crosshairGC;            // crosshair gc
  GC crosshairGCXOR;         // crosshair gcxor

  GC highliteGC;             // gc for frame highlite
  int useHighlite;           // flag to draw highlite box

  List<Marker>* markers;     // pointer to current marker stack
  List<Marker>* undoMarkers; // pointer to current marker stack
  List<Marker>* pasteMarkers;// pointer to current marker stack
  int showMarkers;          // display markers
  int showMarkersText;          // display markers

  UndoMarkerType undoMarkerType; // type

  List<Marker> userMarkers;    // foreground marker stack
  List<Marker> catalogMarkers;
  List<Marker> undoUserMarkers;// remember marker to undo
  List<Marker> pasteUserMarkers;// cut/copy/paste

  int preserveMarkers;       // flag to indicate preserve between loads
  
  GC markerGC;               // marker gc
  GC markerGCXOR;               // marker gc
  GC selectGCXOR;            // select gc
  XRectangle rectWidget[1];  // clip rect
  XRectangle rectWindow[1];  // clip rect

  Vector markerBegin;        // marker move begin in image coords
  Marker* editMarker;        // pointer to marker currently being edited
  Marker* rotateMarker;      // pointer to marker currently being rotated
  Vector regionBegin;        // select region begin in canvas coords
  Vector regionEnd;          // select region end in canvas coords
  CoordSystem markerDefaultSystem; // default marker sys (RAWXY)
  SkyFrame markerDefaultSky; // default marker sky (RAWXY)

  Composite* compositeMarker; // pointer to current composite marker

  Vector iisLastCursor;      // iis cursor state info
  Vector iisSaveCursor;
  Window iisSaveRoot;
  int iisInteractive;
				  
  int showMask;
  char* maskColorName;

  char* bgColorName;
  XColor* bgColor;
  char* nanColorName;
  XColor* nanColor;

  Matrix canvasToRef;
  Matrix canvasToUser;
  Matrix canvasToWidget;
  Matrix canvasToWindow;
  Matrix refToCanvas;
  Matrix refToPanner;
  Matrix refToUser;
  Matrix refToWidget;
  Matrix refToWindow;
  Matrix pannerToRef;
  Matrix pannerToUser;
  Matrix pannerToWidget;
  Matrix userToCanvas;
  Matrix userToRef;
  Matrix userToMagnifier;
  Matrix userToMagnifierCursor;
  Matrix userToPanner;
  Matrix userToWidget;
  Matrix widgetToCanvas;
  Matrix widgetToPanner;
  Matrix widgetToRef;
  Matrix widgetToUser;
  Matrix widgetToWindow;
  Matrix windowToCanvas;
  Matrix windowToRef;
  Matrix windowToWidget;

protected:
  Vector imageSize() {return imageBBox().size();}
  Vector imageCenter() {return imageBBox().center();}

  // Virtual routines with base component
  // framebase.C

  virtual void reset();
  virtual void unload() {};
  virtual void unloadFits();
  virtual void updateMatrices();

  // Virtual routines

  virtual void setCurrentFits(FitsImage*) =0;
  virtual void setKeyFits() =0;
  virtual void setBinCursor(const Vector&) =0;
  virtual void updateBin(const Matrix&) =0;
  virtual BBox imageBBox() =0;
  virtual int doRender() =0;
  virtual void unloadAllFits() =0;
  virtual void updateMagnifierMatrices(Matrix&) =0;

  virtual void updateColorScale() =0;
  virtual int validColorScale() =0;

  virtual int isIIS() =0;

  virtual void psLevel1(PSColorSpace, int, int, float) =0;
  virtual void psLevel2(PSColorSpace, int, int, float) =0;

#ifdef _MACOSX
  virtual void macosx(float, int, int, const Vector&, const Vector&) =0;
#endif

#ifdef _WIN32
  virtual void win32(float, int, int, const Vector&, const Vector&) =0;
#endif

  virtual void buildXImage(XImage*, int, int, int, int,
			   double* (FitsImage::*)()) =0;

  virtual void rotateMotion() =0;
  virtual void alignWCS() =0;
  virtual void alignWCS(FitsImage*) =0;

  // framebase.C
  void resetClip(int);
  void resetValues();

  void update(UpdateType);         // schedule update
  void update(UpdateType,BBox);    // schedule update
  void updateNow(UpdateType);      // schedule update
  void updateNow(UpdateType,BBox); // schedule update
  void updateClip(int);

  double calcZoom(Vector,Vector);  // calculate the zoom factor
  void calcHistEqu();
  void setColorScale(FrScale::ColorScaleType);
  InverseScale* inverseScale(int, double, double, FrScale::ColorScaleType);

  void centerImage();
  void setScanModeIncr(LoadMethod);

  void analysisFits();

  // frcommandbase.C
  void printDouble(double, Precision);
  void printInteger(int);
  void printVector(const Vector&, Precision);
  void printCoordSystem(CoordSystem);

  int hasWCS(CoordSystem);
  int hasWCSEqu(CoordSystem); 
  int hasWCSLinear(CoordSystem);
  int hasContour() {return (currentcontour && *currentcontour);}

  void getInfo(const Vector&, InternalSystem,
	       Matrix& (FitsImage::*)(), Matrix& (FitsImage::*)(),
	       SkyFrame, SkyFormat, char*);

  virtual void getInfoSingle(const Vector&, InternalSystem,
			     Matrix& (FitsImage::*)(),Matrix& (FitsImage::*)(),
			     SkyFrame, SkyFormat, char*);
  virtual void getInfoMosaic(const Vector&, InternalSystem,
			     Matrix& (FitsImage::*)(), 
			     Matrix& (FitsImage::*)(),
			     SkyFrame, SkyFormat, char*);

  void getFitsExt(const Vector&, InternalSystem, Matrix& (FitsImage::*)());
  void getColorScaleLevel(int, const Vector&, Matrix& (FitsImage::*)());

  // frload.C
  void load(MemType, const char*, FitsImage*, LayerType);
  void loadDataCube(MemType, const char*, FitsImage*, LayerType);
  void loadMosaicImage(MemType, const char*, FitsImage*, LayerType);
  void loadMosaicWFPC2(MemType, const char*, FitsImage*);
  void loadMosaic(MemType, const char*, FitsImage*,LayerType);
  void loadFinish();
  FitsHead* parseWCS(istream&);

  // frsave.C
  void saveFitsResample(OutFitsStream&);
  void saveFitsResampleSingle(OutFitsStream&);
  void saveFitsResampleMosaic(OutFitsStream&);
  void saveFitsResampleKeyword(OutFitsStream&, FitsHead&);

  // print.C
  Matrix psMatrix(float, int, int);

  void updateMarkerCoords(List<Marker>*, const Matrix&); // update marker coord
  void updateMarkerCBs(List<Marker>*); // update marker call backs
  void updateBinFileNames();

  // frmarker.C
  void updateCBMarkers();           // execute any update callbacks
  void createTemplate(const Vector&, istream&);

  void updatePanner();       // render panner

private:
  Vector getClip(FrScale::ClipMode, float);

  void bltHist(char*, char*);

  void invalidPixmap();      // update all pixmaps
  int updatePixmap(const BBox&); // renders image/graphics into pixmap
  void updateBase(const BBox&);    // generate base pixmap
  void updatePM(const BBox&);      // update part of pixmap
  void updateGCs();          // update gcs
  void ximageToPixmap(Pixmap, XImage*, int, int, int, int, 
		      double* (FitsImage::*)());

  double calcPannerZoom();  // calculate the zoom factor for the panner
  void pannerPush();         // send panner and update message

  void updateMagnifier(); // render magnifier
  void updateMagnifier(const Vector&); // render magnifier
  void updateMagnifierMatrix();
  void updateMagnifierZoom(const Vector&); // render magnifier
  void renderMagnifierCursor(const Vector&); // render magnifier cursor

  FitsImage* isInFits(const Vector&, Matrix& (FitsImage::*)(), Vector*);

  void printFromRef(FitsImage*, const Vector&, CoordSystem, SkyFrame,
		    SkyFormat, Precision);

  void doubleToTclArray(double, Precision, const char*, const char*, 
			const char*);
  void coordToTclArray(FitsImage*, const Vector&, InternalSystem, CoordSystem, 
		       Precision, const char*, const char*);

  char* varcat(char* buf, char* var, char r);

  int hasLTMV();
  int hasATMV();
  int hasDTMV();

  void printSkyFrame(SkyFrame);
  void printSkyFormat(SkyFormat);

  void markerPrintCoord(const Vector&, InternalSystem);
  void markerPrintCoord(const Vector&, const Vector&,
			CoordSystem, SkyFrame, SkyFormat);

  void markerPrintDouble(double, InternalSystem);
  void markerPrintDouble(const Vector&, double, CoordSystem, SkyFormat);

  void markerPrintDouble(const Vector&, const Vector&, InternalSystem);
  void markerPrintDouble(const Vector&, const Vector&, const Vector&, 
			 CoordSystem, SkyFormat);

  void markerPrintVector(const Vector&,InternalSystem);
  void markerPrintVector(const Vector&,const Vector&,CoordSystem,SkyFormat);

  void psMarkers(List<Marker>*, int);
  void macosxMarkers(List<Marker>*);
  void win32Markers(List<Marker>*);

  void renderMarkers(List<Marker>*, const BBox&);
  void renderMagnifierMarkers(List<Marker>*, const Matrix&, const BBox& bb);
  void unselectMarkers(List<Marker>*);
  void unhighliteMarkers();
  void updateMarkers(List<Marker>*);             // update marker bbox
  void markerUndo(Marker*, UndoMarkerType);

  void markerListHeader(ostream&, CoordSystem, SkyFrame, SkyFormat, 
			int, int);
  void markerListCiaoHeader(ostream&, CoordSystem, SkyFrame, SkyFormat, 
			    int, int);
  void markerListSAOtngHeader(ostream&,CoordSystem, SkyFrame, SkyFormat,
			      int, int);

  void createMarker(Marker*);       // process new marker
  void parseMarker(MarkerFormat,istream&);

  void renderHighLite();     // render a hightlite bbox

  void renderCrosshair();    // render the crosshair
  void psCrosshair(PSColorSpace);
#ifdef _MACOSX
  void macosxCrosshair();
#endif
#ifdef _WIN32
  void win32Crosshair();
#endif
  void drawCrosshair();
  void eraseCrosshair();

  void renderContours(Pixmap, const Matrix&, const BBox&);
  void psContours(PSColorSpace);

  void bltProjection(char*, char*, char*, char*, 
		     CoordSystem, SkyFrame,
		     const Vector&, const Vector&, double, int);
  void bltProjSingle(double*, double*, double*, double*, int,
		     CoordSystem, SkyFrame,
		     const Vector&, const Vector&, double, int);
  void bltProjMosaic(double*, double*, double*, double*, int,
		     CoordSystem, SkyFrame,
		     const Vector&, const Vector&, double, int);
  void bltCut(char*, char*, Orientation, const Vector&);
  void bltCutSingle(double*, double*, int, Vector&, Orientation);
  void bltCutMosaic(double*, double*, int, Vector&, Orientation);

  int postscriptProc(int);   // generate postscript

  Vector zoom() {return zoom_;}

  Orientation getOrientation() {return orientation;}
  double getRotation() {return rotation + wcsRotation;}

  double getWCSRotation(CoordSystem, SkyFrame);
  Orientation getWCSOrientation(CoordSystem, SkyFrame);
  Orientation getWCSOrientation() {return wcsOrientation;}

  XRectangle* getRectWidget() {return rectWidget;}
  XRectangle* getRectWindow() {return rectWindow;}

  Matrix& getCanvasToWidget() {return canvasToWidget;}
  Matrix& getRefToCanvas() {return refToCanvas;}
  Matrix& getRefToWidget() {return refToWidget;}
  Matrix& getRefToWindow() {return refToWindow;}
  Matrix& getWidgetToCanvas() {return widgetToCanvas;}

  FitsImage* findFits(const Vector&);
  int findFits(FitsImage*);

  Vector mapFromRef(const Vector&, InternalSystem);
  Vector mapToRef(const Vector&, InternalSystem);
  double mapLenFromRef(double, InternalSystem);
  Vector mapLenFromRef(const Vector&, InternalSystem);
  double mapLenToRef(double, InternalSystem);
  Vector mapLenToRef(const Vector&, InternalSystem);
  double mapDistFromRef(const Vector&, const Vector&, InternalSystem);

public:
  FrameBase(Tcl_Interp*, Tk_Canvas, Tk_Item*);
  virtual ~FrameBase();

  int parse(istringstream&);

  FitsImage* findFits(int);
  double mapAngleFromRef(double, CoordSystem, SkyFrame =FK5);
  double mapAngleToRef(double, CoordSystem, SkyFrame =FK5);

  void calcAlignWCS(FitsImage*, CoordSystem, SkyFrame,
		    Orientation*, Matrix*, double*);
  void calcAlignWCS(FitsImage*, FitsImage*, CoordSystem, SkyFrame,
		    Orientation*, Matrix*, double*, Vector*);
  Matrix calcAlignWCS(FitsImage*, FitsImage*, CoordSystem, SkyFrame);
  Orientation IRAFOrientation(Orientation);

  Matrix& getCanvasToRef() {return canvasToRef;}
  SkyFrame getMarkerDefaultSky() {return markerDefaultSky;}
  CoordSystem getMarkerDefaultSystem() {return markerDefaultSystem;}
  int isMosaic() {return *currentMosaicCount > 1 ? 1 : 0;}
  int isCube()  {return *currentSliceCount > 1 ? 1 : 0;}
  void resetCompositeMarker() {compositeMarker = NULL;}

  // Misc Functions
  void clearCmd();
  void colorScaleCmd(FrScale::ColorScaleType);
  void getBitpixCmd();
  void getHeightCmd();
  void getInfoCmd(char*);
  void getInfoCmd(const Vector&, InternalSystem, SkyFrame, SkyFormat, 
		  char*);
  void getInfoClipCmd();
  virtual void getTypeCmd() =0;
  void getValueCmd(const Vector&, InternalSystem);
  void getWidthCmd();
  void highliteCmd(int);
  void warpCmd(const Vector&);
  void warpToCmd(const Vector&);
#ifdef _MACOSX
  void macosxPrintCmd();
#endif
#ifdef _WIN32
  void win32PrintCmd();
#endif

  // Background Commands
  virtual void bgColorCmd(const char*) =0;
  void getBgColorCmd();

  // Bin Commands
  void binCmd(const Vector&, const char*, const char*, const char*);
  void binCmd(const Vector&, const Vector&, const char*, const char*, 
	      const char*);
  void binCmd(const Vector&, int, const Vector&, 
	      const char*, const char*, const char*, const char*);
  void binCmd(const Vector&, int, const Vector&, const Vector&, 
	      const char*, const char*, const char*, const char*);
  void binAboutCmd(const Vector&);
  void binColsCmd(const char*, const char*, const char*);
  void binDepthCmd(int);
  void binFactorCmd(const Vector&);
  void binFactorAboutCmd(const Vector&, const Vector&);
  void binFactorToCmd(const Vector&);
  void binFactorToAboutCmd(const Vector&, const Vector&);
  void binToFitCmd();
  void binFunctionCmd(FitsHist::Function);
  void binBufferSizeCmd(int);
  void binFilterCmd(const char*);
  void getBinDepthCmd();
  void getBinFactorCmd();
  void getBinFunctionCmd();
  void getBinBufferSizeCmd();
  void getBinCursorCmd();
  void getBinFilterCmd();
  void getBinColsCmd();
  void getBinColsMinMaxCmd(const char*);
  void getBinColsDimCmd(const char*);
  void getBinListCmd();

  // Clip Commands
  void clipMinMaxModeCmd(FrScale::MinMaxMode);
  void clipMinMaxSampleCmd(int);
  void clipModeCmd(float);
  void clipModeCmd(FrScale::ClipMode);
  void clipPreserveCmd(int);
  void clipScopeCmd(FrScale::ClipScope);
  void clipUserCmd(double, double);
  void clipUserLowCmd(double);
  void clipUserHighCmd(double);
  void clipZScaleContrastCmd(float);
  void clipZScaleSampleCmd(int);
  void clipZScaleLineCmd(int);
  void getClipCmd();
  void getClipCmd(float);
  void getClipCmd(FrScale::ClipMode);
  void getClipMinMaxModeCmd();
  void getClipMinMaxSampleCmd();
  void getClipModeCmd();
  void getClipPreserveCmd();
  void getClipScopeCmd();
  void getClipUserCmd();
  void getClipZScaleContrastCmd();
  void getClipZScaleSampleCmd();
  void getClipZScaleLineCmd();
  void getHistogramCmd(char*, char*);
  void getMinMaxCmd();
  void hasDATAMINCmd();
  void hasIRAFMINCmd();

  // Colormap Commands
  void colormapAreaCmd(int s, ColormapMode m) {cmapArea = s; cmapMode = m;}

  // Contour Commands
  void contourCreateCmd(const char*, int, int, float, int);
  void contourCreateCmd(const char*, int, float, int, const char*);
  void contourLoadCmd(const char*, int, const char*, CoordSystem, SkyFrame);
  void contourCopyCmd(CoordSystem, SkyFrame);
  void contourDeleteCmd();
  void contourDeleteAllCmd();
  void contourLevelCmd(const char*);
  void contourPasteCmd(const char*, int, void*, CoordSystem, SkyFrame);
  void contourSaveCmd(const char*, CoordSystem, SkyFrame);
  void contourSetColorCmd(const char*);
  void contourSetLineWidthCmd(int);
  void getContourCmd(CoordSystem sys, SkyFrame sky);
  void getContourLevelCmd();
  void getContourMethodCmd();
  void getContourColorNameCmd();
  void getContourLineWidthCmd();
  void hasContourCmd();

  // Coordinate Commands
  void getCoordCmd(const Vector&, InternalSystem, CoordSystem, 
		   SkyFrame, SkyFormat);

  void getCursorCmd(InternalSystem);
  void getCursorCmd(CoordSystem, SkyFrame, SkyFormat, Precision);

  void getWCSNameCmd(CoordSystem);

  void hasSystemCmd(CoordSystem);
  void hasAmplifierCmd();
  void hasDetectorCmd();
  void hasPhysicalCmd();

  void hasWCSCmd(CoordSystem);
  void hasWCSEquCmd(CoordSystem);
  void hasWCSLinearCmd(CoordSystem);

  virtual void getWCSAlignCmd() =0;

  // CrossHair Commands
  void crosshairCmd(int);
  void crosshairCmd(const Vector&, InternalSystem);
  void crosshairCmd(const Vector&, CoordSystem, SkyFrame);
  void crosshairBeginCmd(const Vector&, InternalSystem);
  void crosshairBeginCmd(const Vector&, CoordSystem);
  void crosshairMotionCmd(const Vector&, InternalSystem);
  void crosshairMotionCmd(const Vector&, CoordSystem);
  void crosshairWarpCmd(const Vector&);
  void getCrosshairCmd(InternalSystem);
  void getCrosshairCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void getCrosshairStatusCmd();

  void getHorzCutCmd(char*, char*, const Vector&, InternalSystem);
  void getVertCutCmd(char*, char*, const Vector&, InternalSystem);

  // Data Values
  void getDataValuesCmd(int, const Vector&, CoordSystem, SkyFrame, 
			const Vector&, char*);
  void getDataValuesCmd(const Vector&, InternalSystem, const Vector&);

  // DATASEC Commands
  void DATASECCmd(int);
  void getDATASECCmd();
  void hasDATASECCmd();

  // Export Commands
  void exportJPEGCmd(const char*, float) {}

  // Fits Commands
  void getFitsExtCmd(const Vector&, InternalSystem);
  void getFitsCountCmd();
  void getFitsMosaicCountCmd();
  void getFitsDepthCmd();
  void getFitsHeaderCmd(int);
  void getFitsHeaderKeywordCmd(int,const char*);
  void getFitsFileNameCmd(FileNameType);
  void getFitsFileNameCmd(int,FileNameType);
  void getFitsFileNameCmd(const Vector&, InternalSystem, FileNameType);
  void getFitsObjectNameCmd();
  void getFitsMasterCmd();
  void getFitsCenterCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void getFitsSizeCmd();
  void getFitsSizeCmd(CoordSystem, SkyFrame, SkyFormat, Precision);
  void getFitsSliceCmd();

  void hasFitsCmd();
  void hasFitsBinCmd();
  void hasFitsCubeCmd();
  void hasFitsMosaicCmd();

  void loadFitsAllocCmd(const char*, const char*, LayerType);
  void loadFitsAllocGZCmd(const char*, const char*, LayerType);
  void loadFitsChannelCmd(const char*, const char*, LayerType);
  void loadFitsMMapCmd(const char* fn, LoadMethod, LayerType);
  void loadFitsSMMapCmd(const char* hdr, const char* fn, LoadMethod,LayerType);
  void loadFitsMMapIncrCmd(const char* fn, LoadMethod, LayerType);
  void loadFitsShareCmd(ShmType, int, const char*, LoadMethod, LayerType);
  void loadFitsSShareCmd(ShmType, int, int, const char*, LoadMethod,LayerType);
  void loadFitsSocketCmd(int, const char* fn, LayerType);
  void loadFitsSocketGZCmd(int, const char* fn, LayerType);
  void loadFitsVarCmd(const char*, const char*, LoadMethod, LayerType);

  void loadArrAllocCmd(const char*, const char*, LayerType);
  void loadArrAllocGZCmd(const char*, const char*, LayerType);
  void loadArrChannelCmd(const char*, const char*, LayerType);
  void loadArrMMapCmd(const char* fn, LayerType);
  void loadArrMMapIncrCmd(const char* fn, LayerType);
  void loadArrShareCmd(ShmType, int, const char*, LayerType);
  void loadArrSocketCmd(int, const char* fn, LayerType);
  void loadArrSocketGZCmd(int, const char* fn, LayerType);
  void loadArrVarCmd(const char*, const char*, LayerType);

  void loadDataCubeAllocCmd(const char*, const char*, LayerType);
  void loadDataCubeAllocGZCmd(const char*, const char*, LayerType);
  void loadDataCubeChannelCmd(const char*, const char*, LayerType);
  void loadDataCubeMMapCmd(const char* fn, LoadMethod, LayerType);
  void loadDataCubeMMapIncrCmd(const char* fn, LoadMethod, LayerType);
  void loadDataCubeShareCmd(ShmType, int, const char*, LoadMethod, LayerType);
  void loadDataCubeSocketCmd(int, const char* fn, LayerType);
  void loadDataCubeSocketGZCmd(int, const char* fn, LayerType);
  void loadDataCubeVarCmd(const char*, const char*, LoadMethod, LayerType);

  void loadMosaicImageIRAFAllocCmd(const char*, const char*, LayerType);
  void loadMosaicImageIRAFAllocGZCmd(const char*, const char*, LayerType);
  void loadMosaicImageIRAFChannelCmd(const char*, const char*, LayerType);
  void loadMosaicImageIRAFMMapCmd(const char*, LoadMethod, LayerType);
  void loadMosaicImageIRAFMMapIncrCmd(const char*, LoadMethod, LayerType);
  void loadMosaicImageIRAFShareCmd(ShmType, int, const char*, LoadMethod, 
				   LayerType);
  void loadMosaicImageIRAFSocketCmd(int, const char*, LayerType);
  void loadMosaicImageIRAFSocketGZCmd(int, const char*, LayerType);
  void loadMosaicImageIRAFVarCmd(const char*,const char*, LoadMethod, 
				 LayerType);

  void loadMosaicIRAFAllocCmd(const char*, const char*, LayerType);
  void loadMosaicIRAFAllocGZCmd(const char*, const char*, LayerType);
  void loadMosaicIRAFChannelCmd(const char*, const char*, LayerType);
  void loadMosaicIRAFMMapCmd(const char*, LoadMethod, LayerType);
  void loadMosaicIRAFSMMapCmd(const char*, const char*, LoadMethod, LayerType);
  void loadMosaicIRAFMMapIncrCmd(const char*, LoadMethod, LayerType);
  void loadMosaicIRAFShareCmd(ShmType, int, const char*, 
			      LoadMethod, LayerType);
  void loadMosaicIRAFSShareCmd(ShmType, int, int, const char*, 
			       LoadMethod, LayerType);
  void loadMosaicIRAFSocketCmd(int, const char*, LayerType);
  void loadMosaicIRAFSocketGZCmd(int, const char*, LayerType);
  void loadMosaicIRAFVarCmd(const char*, const char*, LoadMethod, LayerType);

  void loadMosaicImageWCSAllocCmd(CoordSystem, const char*, const char*, 
				  LayerType);
  void loadMosaicImageWCSAllocGZCmd(CoordSystem, const char*, const char*,
				    LayerType);
  void loadMosaicImageWCSChannelCmd(CoordSystem, const char*, const char*,
				    LayerType);
  void loadMosaicImageWCSMMapCmd(CoordSystem, const char*, LoadMethod,
				 LayerType);
  void loadMosaicImageWCSMMapIncrCmd(CoordSystem, const char*, LoadMethod, 
				     LayerType);
  void loadMosaicImageWCSShareCmd(CoordSystem, ShmType, int, const char*, 
				  LoadMethod, LayerType);
  void loadMosaicImageWCSSocketCmd(CoordSystem, int, const char*, 
				   LayerType);
  void loadMosaicImageWCSSocketGZCmd(CoordSystem, int, const char*, 
				     LayerType);
  void loadMosaicImageWCSVarCmd(CoordSystem, const char*,const char*, 
				LoadMethod, LayerType);

  void loadMosaicWCSAllocCmd(CoordSystem, const char*, const char*, 
			     LayerType);
  void loadMosaicWCSAllocGZCmd(CoordSystem, const char*, const char*, 
			       LayerType);
  void loadMosaicWCSChannelCmd(CoordSystem, const char*, const char*, 
			       LayerType);
  void loadMosaicWCSMMapCmd(CoordSystem, const char*, LoadMethod, 
			    LayerType);
  void loadMosaicWCSSMMapCmd(CoordSystem, const char*, const char*, 
			     LoadMethod, LayerType);
  void loadMosaicWCSMMapIncrCmd(CoordSystem, const char*, LoadMethod, 
				LayerType);
  void loadMosaicWCSShareCmd(CoordSystem, ShmType, int, const char*, 
			     LoadMethod, LayerType);
  void loadMosaicWCSSShareCmd(CoordSystem, ShmType, int, int, const char*, 
			      LoadMethod, LayerType);
  void loadMosaicWCSSocketCmd(CoordSystem, int, const char*, 
			      LayerType);
  void loadMosaicWCSSocketGZCmd(CoordSystem, int, const char*, 
				LayerType);
  void loadMosaicWCSVarCmd(CoordSystem, const char*, const char*, 
			   LoadMethod, LayerType);

  void loadMosaicImageWFPC2AllocCmd(const char*, const char*);
  void loadMosaicImageWFPC2AllocGZCmd(const char*, const char*);
  void loadMosaicImageWFPC2ChannelCmd(const char*, const char*);
  void loadMosaicImageWFPC2MMapCmd(const char* fn, LoadMethod);
  void loadMosaicImageWFPC2MMapIncrCmd(const char* fn, LoadMethod);
  void loadMosaicImageWFPC2ShareCmd(ShmType, int, const char*,LoadMethod);
  void loadMosaicImageWFPC2SocketCmd(int, const char* fn);
  void loadMosaicImageWFPC2SocketGZCmd(int, const char* fn);
  void loadMosaicImageWFPC2VarCmd(const char*, const char*, LoadMethod);

  void loadIncrDataCmd(int, int, int, int, int);
  void loadIncrMinMaxCmd(int, int, int, int, int);
  void loadIncrEndCmd();

  void loadFitsSlaveCmd(const char*, FitsFile*);

  void saveFitsImageFileCmd(const char*, int);
  void saveFitsImageChannelCmd(const char*, int);
  void saveFitsImageSocketCmd(int, int);

  void saveFitsTableFileCmd(const char*, int);
  void saveFitsTableChannelCmd(const char*, int);
  void saveFitsTableSocketCmd(int, int);

  void saveArrayFileCmd(const char*);
  void saveArrayChannelCmd(const char*);
  void saveArraySocketCmd(int);

  void saveFitsResampleFileCmd(const char*, int);
  void saveFitsResampleChannelCmd(const char*, int);
  void saveFitsResampleSocketCmd(int, int);

  virtual void sliceCmd(int) =0;

  void updateFitsCmd(int);
  void updateFitsCmd(int, BBox, int);
  void unloadFitsCmd();

  // Fits RGB
  virtual void loadRGBCubeAllocCmd(const char*, const char*) {}
  virtual void loadRGBCubeAllocGZCmd(const char*, const char*) {}
  virtual void loadRGBCubeChannelCmd(const char*, const char*) {}
  virtual void loadRGBCubeMMapCmd(const char*, LoadMethod) {}
  virtual void loadRGBCubeSMMapCmd(const char*, const char*, LoadMethod) {}
  virtual void loadRGBCubeMMapIncrCmd(const char*, LoadMethod) {}
  virtual void loadRGBCubeShareCmd(ShmType, int, const char*, LoadMethod) {}
  virtual void loadRGBCubeSShareCmd(ShmType,int,int,const char*,LoadMethod) {}
  virtual void loadRGBCubeSocketCmd(int, const char*) {}
  virtual void loadRGBCubeSocketGZCmd(int, const char*) {}
  virtual void loadRGBCubeVarCmd(const char*, const char*, LoadMethod) {}

  virtual void loadRGBCubeSlaveCmd(const char*, FitsFile*) {}

  virtual void loadRGBImageAllocCmd(const char*, const char*) {}
  virtual void loadRGBImageAllocGZCmd(const char*, const char*) {}
  virtual void loadRGBImageChannelCmd(const char*, const char*) {}
  virtual void loadRGBImageMMapCmd(const char*, LoadMethod) {}
  virtual void loadRGBImageMMapIncrCmd(const char*, LoadMethod) {}
  virtual void loadRGBImageShareCmd(ShmType, int, const char*, LoadMethod) {}
  virtual void loadRGBImageSocketCmd(int, const char*) {}
  virtual void loadRGBImageSocketGZCmd(int, const char*) {}
  virtual void loadRGBImageVarCmd(const char*, const char*, LoadMethod) {}

  virtual void loadArrRGBCubeAllocCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeAllocGZCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeChannelCmd(const char*, const char*) {}
  virtual void loadArrRGBCubeMMapCmd(const char*) {}
  virtual void loadArrRGBCubeMMapIncrCmd(const char*) {}
  virtual void loadArrRGBCubeShareCmd(ShmType, int, const char*) {}
  virtual void loadArrRGBCubeSocketCmd(int, const char*) {}
  virtual void loadArrRGBCubeSocketGZCmd(int, const char*) {}
  virtual void loadArrRGBCubeVarCmd(const char*, const char*) {}

  // Fitsy Commands
  void fitsyHasExtCmd(const char*);

  // Frame
  virtual void colormapCmd(int, float, float, int, unsigned short*, 
			     unsigned char*, int) {}
  virtual void colormapMotionCmd(int, float, float, int, unsigned short*,
				   unsigned char*, int) {}
  virtual void colormapEndCmd(int, float, float, int, unsigned short*,
				unsigned char*, int) {}

  // Frame RGB
  virtual void colormapCmd(float, float, float, float, float, float, int,
			     unsigned char*, int) {}
  virtual void colormapMotionCmd(float, float, float, float, float, float, int,
				   unsigned char*, int) {}
  virtual void colormapEndCmd(float, float, float, float, float, float, int,
				unsigned char*, int) {}

  virtual void colormapBeginCmd() =0;
  virtual void colormapBeginCmd(const Vector&) =0;

  virtual void getColormapCmd() =0;
  void getColorScaleCmd();
  void getColorScaleLevelCmd(int);
  void getColorScaleLevelCmd(int,const Vector&, InternalSystem);
  void getColorScaleLevelCmd(int, float, float, FrScale::ColorScaleType);

  // Grid Commands
  void gridCmd(CoordSystem, SkyFrame, SkyFormat, Grid::GridType, const char*);
  void gridDeleteCmd();
  void hasGridCmd();
  void getGridCmd();
  void getGridOptionCmd();

  // IIS Commands
  virtual void iisCmd(int, int) {}
  virtual void iisEraseCmd() {}
  virtual void iisGetCmd(char*, int, int, int, int) {}
  virtual void iisSetCmd(const char*, int, int, int, int) {}
  virtual void iisWCSCmd(const Matrix&, const Vector&, int) {}

  void iisMessageCmd(const char*);
  void iisUpdateCmd() {updateNow(MATRIX);}
  void iisSetCursorCmd(const Vector&, InternalSystem);
  void iisSetCursorCmd(const Vector&, CoordSystem);
  void iisGetCursorCmd();
  void iisCursorModeCmd(int);
  void hasIISCmd();

  void iisGetFileNameCmd();
  void iisGetFileNameCmd(int);
  void iisGetFileNameCmd(const Vector&);
  void iisSetFileNameCmd(const char*);
  void iisSetFileNameCmd(const char*,int);

  // Magnifier Commands
  void magnifierCmd(char*, int, int);
  void magnifierCmd(int);
  void magnifierGraphicsCmd(int);
  void magnifierCursorCmd(int);
  void magnifierZoomCmd(double);

  // Marker Commands
  void createCircleCmd(const Vector&, 
		       double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createAnnulusCmd(const Vector&, 
			double, double, int,
			const char*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createAnnulusCmd(const Vector&, 
			int, double*,
			const char*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createCpandaCmd(const Vector&, 
		      double, double, int, 
		      double, double, int,
		      const char*, int, const char*, 
		      const char*, unsigned short, const char*,
		      const List<Tag>&, const List<CallBack>& cb);
  void createCpandaCmd(const Vector&, 
		      int, double*,
		      int, double*,
		      const char*, int, const char*, 
		      const char*, unsigned short, const char*,
		      const List<Tag>&, const List<CallBack>& cb);


  void createEllipseCmd(const Vector&, 
			const Vector&, 
			double, 
			const char*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createEllipseAnnulusCmd(const Vector&, 
			       const Vector&, const Vector&, int, 
			       double,
			       const char*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);
  void createEllipseAnnulusCmd(const Vector&, 
			       int, Vector*,
			       double, 
			       const char*, int,  const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);
  void createEpandaCmd(const Vector&, 
		       double, double, int, 
		       const Vector&, const Vector&, int, 
		       double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createEpandaCmd(const Vector&, 
		       int, double*,
		       int, Vector*,
		       double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);


  void createBoxCmd(const Vector&, 
		    const Vector&, 
		    double, 
		    const char*, int, const char*, 
		    const char*, unsigned short, const char*,
		    const List<Tag>&, const List<CallBack>& cb);
  void createBoxAnnulusCmd(const Vector&, 
			   const Vector&, const Vector&, int, 
			   double,
			   const char*, int, const char*, 
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBoxAnnulusCmd(const Vector&, 
			   int, Vector*,
			   double, 
			   const char*, int, const char*, 
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBpandaCmd(const Vector&, 
		       double, double, int, 
		       const Vector&, const Vector&, int, 
		       double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);
  void createBpandaCmd(const Vector&, 
		       int, double*,
		       int, Vector*,
		       double,
		       const char*, int, const char*, 
		       const char*, unsigned short, const char*,
		       const List<Tag>&, const List<CallBack>& cb);


  void createLineCmd(const Vector&, 
		     const Vector&, 
		     int, int,
		     const char*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createVectCmd(const Vector&, 
		     const Vector&, 
		     int,
		     const char*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createVectCmd(const Vector&, 
		     double mag, double ang, 
		     int,
		     const char*, int, const char*,
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);
  void createProjectionCmd(const Vector&, 
			   const Vector&,
			   double, 
			   const char*, const char*,
			   const char*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createRulerCmd(const Vector&, 
		      const Vector&, 
		      CoordSystem, SkyFrame, CoordSystem, SkyFormat,
		      const char*, int, const char*,
		      const char*, unsigned short, const char*,
		      const List<Tag>&, const List<CallBack>& cb);

  void createCirclePointCmd(const Vector&, int,
			    const char*, int, const char*,
			    const char*, unsigned short, const char*,
			    const List<Tag>&, const List<CallBack>& cb);
  void createBoxPointCmd(const Vector&, int,
			 const char*, int, const char*,
			 const char*, unsigned short, const char*,
			 const List<Tag>&, const List<CallBack>& cb);
  void createDiamondPointCmd(const Vector&, int,
			     const char*, int, const char*, 
			     const char*, unsigned short, const char*,
			     const List<Tag>&, const List<CallBack>& cb);
  void createCrossPointCmd(const Vector&, int,
			   const char*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createExPointCmd(const Vector&, int,
			const char*, int, const char*,
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createArrowPointCmd(const Vector&, int,
			   const char*, int, const char*,
			   const char*, unsigned short, const char*,
			   const List<Tag>&, const List<CallBack>& cb);
  void createBoxCirclePointCmd(const Vector&, int,
			       const char*, int, const char*,
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);

  void createTextCmd(const Vector&, 
		     double, int,
		     const char*, int, const char*, 
		     const char*, unsigned short, const char*,
		     const List<Tag>&, const List<CallBack>& cb);

  void createPolygonCmd(const Vector&, 
			const Vector&,
			const char*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createPolygonCmd(const List<Vertex>&,
			const char*, int, const char*, 
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);
  void createContourPolygonCmd(const char*, int, const char*, 
			       const char*, unsigned short, const char*,
			       const List<Tag>&, const List<CallBack>& cb);

  void createCompassCmd(const Vector&, 
			double, 
			const char*, const char*,
			int, int,
			CoordSystem, SkyFrame,
			const char*, int, const char*,
			const char*, unsigned short, const char*,
			const List<Tag>&, const List<CallBack>& cb);

  void createCompositeCmd(const Vector&, double, int,
			  const char*, int, const char*, 
			  const char*, unsigned short, const char*,
			  const List<Tag>&, const List<CallBack>& cb);
  void createCompositeCmd(
			  const char*, int, const char*, 
			  const char*, unsigned short, const char*,
			  const List<Tag>&, const List<CallBack>& cb);

  void createTemplateCmd(const Vector&, const char*);
  void createTemplateVarCmd(const Vector&, const char* var);

  void getMarkerAngleCmd(int);
  void getMarkerAngleCmd(int, CoordSystem, SkyFrame);
  void getMarkerAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBoxAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerBoxAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBoxRadiusCmd(int, InternalSystem);
  void getMarkerBoxRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerBpandaAnglesCmd(int);
  void getMarkerBpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerBpandaRadiusCmd(int, InternalSystem);
  void getMarkerBpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerCenterCmd(int, InternalSystem);
  void getMarkerCenterCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerCircleRadiusCmd(int, InternalSystem);
  void getMarkerCircleRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerColorCmd();
  void getMarkerColorCmd(const char*);
  void getMarkerColorCmd(int);
  void getMarkerCompassArrowCmd(int);
  void getMarkerCompassLabelCmd(int);
  void getMarkerCompassRadiusCmd(int, InternalSystem);
  void getMarkerCompassRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerCompassSystemCmd(int);
  void getMarkerCompositeCmd(int);
  void getMarkerCpandaAnglesCmd(int);
  void getMarkerCpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerCpandaRadiusCmd(int, InternalSystem);
  void getMarkerCpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEllipseRadiusCmd(int, InternalSystem);
  void getMarkerEllipseRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEllipseAnnulusRadiusCmd(int, InternalSystem);
  void getMarkerEllipseAnnulusRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerEpandaAnglesCmd(int);
  void getMarkerEpandaAnglesCmd(int, CoordSystem, SkyFrame);
  void getMarkerEpandaRadiusCmd(int, InternalSystem);
  void getMarkerEpandaRadiusCmd(int, CoordSystem, SkyFormat);
  void getMarkerFontCmd();
  void getMarkerFontCmd(const char*);
  void getMarkerFontCmd(int);
  void getMarkerHandleCmd(const Vector&);
  void getMarkerIdCmd(const char*);
  void getMarkerIdCmd(const Vector&);
  void getMarkerIdAllCmd();
  void getMarkerLineCmd(int, InternalSystem);
  void getMarkerLineCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerLineArrowCmd(int);
  void getMarkerLineLengthCmd(int, InternalSystem);
  void getMarkerLineLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerLineWidthCmd();
  void getMarkerLineWidthCmd(int);
  void getMarkerMapLenFromRefCmd(int, double, CoordSystem, SkyFormat);
  void getMarkerPointSizeCmd(int);
  void getMarkerPolygonSegmentCmd(const Vector&);
  void getMarkerPreserveCmd();
  void getMarkerProjectionCmd(int, char*, char*, char*, char*);
  void getMarkerProjectionPointsCmd(int, InternalSystem);
  void getMarkerProjectionPointsCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerProjectionLengthCmd(int, InternalSystem);
  void getMarkerProjectionLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerProjectionMethodCmd(int);
  void getMarkerProjectionWidthCmd(int, InternalSystem);
  void getMarkerProjectionWidthCmd(int, CoordSystem, SkyFormat);
  void getMarkerProjectionSystemCmd(int);
  void getMarkerPropertyCmd(unsigned short);
  void getMarkerPropertyCmd(const char*, unsigned short);
  void getMarkerPropertyCmd(int, unsigned short);
  void getMarkerRulerLengthCmd(int, InternalSystem);
  void getMarkerRulerLengthCmd(int, CoordSystem, SkyFormat);
  void getMarkerRulerPointCmd(int, InternalSystem);
  void getMarkerRulerPointCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerRulerSystemCmd(int);
  void getMarkerSelectedCmd();
  void getMarkerSelectedCmd(int);
  void getMarkerSelectedCmd(const Vector&);
  void getMarkerSelectedNumberCmd();
  void getMarkerHighlitedCmd();
  void getMarkerHighlitedCmd(int);
  void getMarkerHighlitedCmd(const Vector&);
  void getMarkerHighlitedNumberCmd();
  void getMarkerShowCmd();
  void getMarkerShowTextCmd();
  void getMarkerTagCmd(const char*);
  void getMarkerTagCmd(int);
  void getMarkerTagCmd(int,int);
  void getMarkerTagDefaultNameCmd();
  void getMarkerTagNumberCmd(const char*);
  void getMarkerTagsCmd();
  void getMarkerTextCmd(int);
  void getMarkerTextRotateCmd(int);
  void getMarkerTypeCmd(int);
  void getMarkerVectorCmd(int, InternalSystem);
  void getMarkerVectorCmd(int, CoordSystem, SkyFrame, SkyFormat);
  void getMarkerVectorArrowCmd(int);
  void getMarkerVectorLengthCmd(int, InternalSystem);
  void getMarkerVectorLengthCmd(int, CoordSystem, SkyFormat);

  void hasMarkerSelectedCmd();
  void hasMarkerHighlitedCmd();
  void hasMarkerPasteCmd();
  void hasMarkerUndoCmd();

  void markerLayerCmd(MarkerLayer);

  void markerAngleCmd(int, double);
  void markerAngleCmd(int, double, CoordSystem, SkyFrame);
  void markerAnnulusCreateRadiusCmd(int, const Vector&);
  void markerAnnulusDeleteRadiusCmd(int, int);
  void markerAnnulusRadiusCmd(int, double, double, int, InternalSystem);
  void markerAnnulusRadiusCmd(int, double, double, int, CoordSystem,SkyFormat);
  void markerAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat);
  void markerBackCmd();
  void markerBackCmd(const char*);
  void markerBackCmd(int);
  void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 InternalSystem);
  void markerBoxAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				 CoordSystem, SkyFormat);
  void markerBoxAnnulusRadiusCmd(int, const char*, CoordSystem, SkyFormat);
  void markerBoxAnnulusCreateRadiusCmd(int, const Vector&);
  void markerBoxAnnulusDeleteRadiusCmd(int, int);
  void markerBoxRadiusCmd(int, const Vector&, InternalSystem);
  void markerBoxRadiusCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerBpandaCreateAnglesCmd(int, const Vector&);
  void markerBpandaCreateRadiusCmd(int, const Vector&);
  void markerBpandaDeleteCmd(int, int);
  void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int);
  void markerBpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame);
  void markerBpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat);
  void markerCallBackCmd(int, CallBack::Type, const char*, const char*);
  void markerCircleRadiusCmd(int, double, InternalSystem);
  void markerCircleRadiusCmd(int, double, CoordSystem, SkyFormat);
  void markerColorCmd(const char*);
  void markerColorCmd(const char*, const char*);
  void markerColorCmd(int, const char*);
  void markerCompassArrowCmd(int, int, int);
  void markerCompassLabelCmd(int, const char*, const char*);
  void markerCompassRadiusCmd(int, double, InternalSystem);
  void markerCompassRadiusCmd(int, double, CoordSystem, SkyFormat);
  void markerCompassSystemCmd(int, CoordSystem, SkyFrame);
  void markerCompositeCmd(int, int);
  void markerCompositeDeleteCmd();
  void markerCommandCmd(MarkerFormat, const char*);
  void markerCommandVarCmd(MarkerFormat, const char*);
  void markerCpandaCreateAnglesCmd(int, const Vector&);
  void markerCpandaCreateRadiusCmd(int, const Vector&);
  void markerCpandaDeleteCmd(int, int);
  void markerCpandaEditCmd(int, double, double, int, double, double, int);
  void markerCpandaEditCmd(int, double, double, int, double, double, int,
			  CoordSystem, SkyFrame);
  void markerCpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,CoordSystem, SkyFormat);
  void markerDeleteCmd();
  void markerDeleteCmd(const char*);
  void markerDeleteCmd(int);
  void markerDeleteAllCmd();
  void markerDeleteLastCmd();
  void markerDeleteCallBackCmd(int, CallBack::Type, const char*);
  void markerDeleteTagCmd(int);
  void markerDeleteTagCmd(int, const char*);
  void markerDeleteTagCmd(int, int);
  void markerEditBeginCmd(int, int);
  void markerEditBeginCmd(const Vector&, int);
  void markerEditMotionCmd(const Vector&, int h);
  void markerEditEndCmd();
  void markerEllipseRadiusCmd(int, const Vector&, InternalSystem);
  void markerEllipseRadiusCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     InternalSystem);
  void markerEllipseAnnulusRadiusCmd(int, const Vector&, const Vector&, int, 
				     CoordSystem, SkyFormat);
  void markerEllipseAnnulusRadiusCmd(int, const char*, 
				     CoordSystem, SkyFormat);
  void markerEllipseAnnulusCreateRadiusCmd(int, const Vector&);
  void markerEllipseAnnulusDeleteRadiusCmd(int, int);
  void markerEpandaCreateAnglesCmd(int, const Vector&);
  void markerEpandaCreateRadiusCmd(int, const Vector&);
  void markerEpandaDeleteCmd(int, int);
  void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int);
  void markerEpandaEditCmd(int, double, double, int, 
			   const Vector&, const Vector&, int,
			   CoordSystem, SkyFrame);
  void markerEpandaEditCmd(int, const char*, const char*, 
			   CoordSystem, SkyFrame,
			   CoordSystem, SkyFormat);
  void markerFontCmd(const char*);
  void markerFontCmd(const char*, const char*);
  void markerFontCmd(int, const char*);
  void markerFrontCmd();
  void markerFrontCmd(const char*);
  void markerFrontCmd(int);
  void markerKeyCmd();
  void markerKeyCmd(const Vector&);
  void markerLineCmd(int, const Vector&, const Vector&, InternalSystem);
  void markerLineCmd(int, const Vector&, const Vector&, CoordSystem, SkyFrame);
  void markerLineArrowCmd(int, int, int);
  void markerLineWidthCmd(int);
  void markerLineWidthCmd(int, int);
  void markerListCmd(MarkerFormat, CoordSystem, SkyFrame, SkyFormat, 
		     int strip, int listwcs, int select,
		     unsigned short, unsigned short, List<Tag>&);
  void markerLoadCmd(MarkerFormat,const char*);
  void markerLoadCmd(MarkerFormat,const char*,CoordSystem,SkyFrame);
  void markerLoadCmd(MarkerFormat,int);
  void markerLoadCmd(MarkerFormat,int,CoordSystem,SkyFrame);
  void markerLoadFitsCmd(const char*, const char*, int, const char*);
  void markerMoveCmd(const Vector&);
  void markerMoveCmd(const char*, const Vector&);
  void markerMoveCmd(int id, const Vector&);
  void markerMoveBeginCmd(const Vector&);
  void markerMoveBeginCmd(int, const Vector&);
  void markerMoveMotionCmd(const Vector&);
  void markerMoveMotionCmd(int, const Vector&);
  void markerMoveEndCmd();
  void markerMoveEndCmd(int);
  void markerMoveToCmd(const Vector&, InternalSystem);
  void markerMoveToCmd(const Vector&, CoordSystem, SkyFrame);
  void markerMoveToCmd(const char*, const Vector&, InternalSystem);
  void markerMoveToCmd(const char*, const Vector&, CoordSystem, SkyFrame);
  void markerMoveToCmd(int, const Vector&, InternalSystem);
  void markerMoveToCmd(int, const Vector&, CoordSystem, SkyFrame);
  void markerPointSizeCmd(int,int);
  void markerPolygonCreateVertexCmd(int, int, const Vector&);
  void markerPolygonDeleteVertexCmd(int, int);
  void markerPolygonResetCmd(int, const Vector&, InternalSystem);
  void markerPolygonResetCmd(int, const Vector&, CoordSystem, SkyFormat);
  void markerPreserveCmd(int r) {preserveMarkers = r;}
  void markerPropertyCmd(unsigned short, int);
  void markerPropertyCmd(const char*, unsigned short, int);
  void markerPropertyCmd(int, unsigned short, int);
  void markerPropertyCmd(unsigned short, int, const Vector&);
  void markerProjectionCmd(int, const Vector&, const Vector&, 
			   InternalSystem, double, int);
  void markerProjectionCmd(int, const Vector&, const Vector&, 
			   CoordSystem, SkyFrame, 
			   double, int, CoordSystem, SkyFormat);
  void markerProjectionSystemCmd(int, CoordSystem, SkyFrame);
  void markerRotateBeginCmd(int);
  void markerRotateBeginCmd(const Vector&);
  void markerRotateMotionCmd(const Vector&, int h);
  void markerRotateEndCmd();
  void markerRulerPointCmd(int, const Vector&, const Vector&, 
			   CoordSystem, SkyFrame);
  void markerRulerPointCmd(int, const Vector&, const Vector&, InternalSystem);
  void markerRulerSystemCmd(int, CoordSystem, SkyFrame, 
			    CoordSystem, SkyFormat);
  void markerSaveCmd(const char*, MarkerFormat, CoordSystem, SkyFrame,
		     SkyFormat, int strip, int listwcs);
  void markerSaveTemplateCmd(const char*);

  void markerVectorCmd(int, const Vector&, InternalSystem, double, double);
  void markerVectorCmd(int, const Vector&, CoordSystem, SkyFrame, 
		       double, CoordSystem, SkyFormat, double);
  void markerVectorArrowCmd(int, int);

  void markerSelectAllCmd();
  void markerSelectCmd(int); // not used
  void markerSelectCmd(const char*);
  void markerSelectToggleCmd();
  void markerSelectToggleCmd(const Vector&);
  void markerSelectOnlyCmd(int);
  void markerSelectOnlyCmd(const char*);
  void markerSelectOnlyCmd(const Vector&);
  void markerUnselectAllCmd();
  void markerUnselectCmd(int); // not used
  void markerUnselectCmd(const char*); // not used

  void markerHighliteAllCmd(); // not used
  void markerHighliteCmd(int); // not used
  void markerHighliteCmd(const char*); // not used
  void markerHighliteToggleCmd(const Vector&);
  void markerHighliteOnlyCmd(int);
  void markerHighliteOnlyCmd(const char*); // not used
  void markerHighliteOnlyCmd(const Vector&);  // not used
  void markerUnhighliteAllCmd();
  void markerUnhighliteCmd(int); // not used
  void markerUnhighliteCmd(const char*);

  void markerShowCmd(int);
  void markerShowTextCmd(int);
  void markerTagCmd(const char*);
  void markerTagCmd(int, const char*);
  void markerTagEditCmd(const char*, const char*);
  void markerTagDeleteCmd(const char*);
  void markerTagDeleteAllCmd();
  void markerTagUpdateCmd(const char*);
  void markerTextCmd(int, const char*);
  void markerTextRotateCmd(int, int);

  void markerCopyCmd();
  void markerCopyCmd(const char*);
  void markerCutCmd();
  void markerCutCmd(const char*);
  void markerPasteCmd();
  void markerPasteCmd(CoordSystem, SkyFrame);
  void markerUndoCmd();

  void regionHighliteBeginCmd(const Vector& v) {regionSelectBeginCmd(v);}
  void regionHighliteMotionCmd(const Vector& v) {regionSelectMotionCmd(v);}
  void regionHighliteEndCmd();
  void regionHighliteShiftEndCmd();

  void regionSelectBeginCmd(const Vector&);
  void regionSelectMotionCmd(const Vector&);
  void regionSelectEndCmd();
  void regionSelectShiftEndCmd();

  // Mask Commands
  void getMaskCmd();
  void getMaskColorCmd();
  void maskCmd(int);
  void maskColorCmd(const char*);
  void maskClearCmd();

  // NaN Commands
  void getNANColorCmd();
  virtual void nanColorCmd(const char*) =0;

  // Pan Zoom Rotate Orient Commands
  void centerCmd();
  void getOrientCmd();
  void getRotateCmd(Precision);
  void getPanPreserveCmd();
  void getZoomCmd(Precision);
  void orientCmd(Orientation);

  void panCmd(const Vector&, const Vector&);
  void panCmd(const Vector&, InternalSystem);
  void panCmd(const Vector&, CoordSystem, SkyFrame);
  void panBBoxCmd(const Vector&);
  void panMotionBeginCmd(const Vector&);
  void panMotionCmd(const Vector&);
  void panMotionEndCmd(const Vector&);
  void panToCmd(const Vector&, InternalSystem);
  void panToCmd(const Vector&, CoordSystem, SkyFrame);
  void panPreserveCmd(int r) {preservePan = r;}

  void rotateCmd(double);
  void rotateToCmd(double);
  void rotateMotionBeginCmd();
  void rotateMotionCmd(double);
  void rotateMotionEndCmd();
  void wcsAlignCmd(int, CoordSystem, SkyFrame);
  void wcsAlignCmd(int, CoordSystem, SkyFrame, FitsImage*);

  void zoomCmd(const Vector&);
  void zoomAboutCmd(const Vector&, const Vector&, InternalSystem);
  void zoomAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame);
  void zoomToCmd(const Vector&);
  void zoomToAboutCmd(const Vector&, const Vector&, InternalSystem);
  void zoomToAboutCmd(const Vector&, const Vector&, CoordSystem, SkyFrame);
  void zoomToFitCmd(double =1);

  // Panner
  void pannerCmd(char*, int, int);
  void pannerCmd(int);
  void pannerCmd(CoordSystem, SkyFrame);
  void updatePannerCmd();
  void updateMagnifierCmd();
  void updateMagnifierCmd(const Vector& v);

  // Pixel Table
  void getPixelTableCmd(const Vector&, InternalSystem, int, char*);
  void getPixelTable(const Vector&, Matrix& (FitsImage::*matrixToData)(),
		     int, char*);
  void getPixelTable(FitsImage*, const Vector&, int, char*);


  // RGB Commands
  virtual void getRGBChannelCmd() =0;
  virtual void setRGBChannelCmd(const char*) {}
  virtual void getRGBViewCmd() =0;
  virtual void setRGBViewCmd(int, int, int) {}
  virtual void getRGBSystemCmd() =0;
  virtual void setRGBSystemCmd(CoordSystem) {}

  // Smooth Commands
  void smoothCmd(int,int);
  void smoothDeleteCmd();
  void hasSmoothCmd();
  void getSmoothFunctionCmd();
  void getSmoothRadiusCmd();

  // WCS
  void wcsResetCmd();
  void wcsReplaceCmd(int);
  void wcsReplaceCmd(const char*);
  void wcsAppendCmd(int);
  void wcsAppendCmd(const char*);
};

#endif
