// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framepseudocolor.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

FramePseudoColor::FramePseudoColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i,c,item), Frame(i,c,item), FramePseudo(i,c,item)
{
}

void FramePseudoColor::fillXImageSingle(XImage* xmap, 
					int x0, int y0, int x1, int y1,
					double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!currentFits)
    return;

  int& width = xmap->width;
  int& height = xmap->height;
  int& bytesPerLine = xmap->bytes_per_line;
  char* data = XImageData(xmap);

  double* mm = (currentFits->*getMatrix)();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  // clear ximage
  memset(data, bgColor->pixel, bytesPerLine * height);

  // test as doubles, convert to int's at last minute.
  // use double to maintain precision
  // This solves a host of problems.

  for (long jj=y0; jj<y1; jj++) {
    // the line may be padded at the end
    char* dest = data + jj*bytesPerLine + x0;

    for (long ii=x0; ii<x1; ii++,dest++) {
      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      *dest = table[0];
	    else if (value >= hh)
	      *dest = table[length];
	    else
	      *dest = table[(int)(((value - ll)/diff * length) + .5)];
	  }
	  else
	    *dest = nanColor->pixel;
	}
	else
	  *dest = maskColor->pixel;
      }
    }
  }
}

void FramePseudoColor::fillXImageMosaic(XImage* xmap, 
					 int x0, int y0, int x1, int y1,
					 double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int& width = xmap->width;
  int& height = xmap->height;
  int& bytesPerLine = xmap->bytes_per_line;
  char* data = XImageData(xmap);

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();

  // clear ximage
  memset(data, bgColor->pixel, bytesPerLine * height);

  // test as doubles, convert to int's at last minute.
  // use double to maintain precision
  // This solves a host of problems.

  for (long jj=y0; jj<y1; jj++) {
    // the line may be padded at the end
    char* dest = data + jj*bytesPerLine + x0;

    for (long ii=x0; ii<x1; ii++,dest++) {
      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = table[0];
	      else if (value >= hh)
		*dest = table[length];
	      else
		*dest = table[(int)(((value - ll)/diff * length) + .5)];
	    }
	    else
	      *dest = nanColor->pixel;
	  }
	  else
	    *dest = maskColor->pixel;

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }
    }
  }
}

void FramePseudoColor::colormapCmd(int id, float b, float c, int i, 
				    unsigned short* index, 
				    unsigned char* cells, int cnt)
{
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();
}

void FramePseudoColor::psLevel2(PSColorSpace mode, int width, int height, 
				float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel2HeadPseudo(mode, width, height, colorCount, indexCells, colorCells);

  if (!isMosaic())
    psLevel2Single(mode, width, height, scale);
  else
    psLevel2Mosaic(mode, width, height, scale);
}

void FramePseudoColor::psLevel2Single(PSColorSpace mode,
				       int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned short* table = colorScale->psIndex();

  // update PS Matrix
  currentFits->updatePS(psMatrix(scale, width, height));

  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {
      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      filter << table[0];
	    else if (value >= hh)
	      filter << table[length];
	    else
	      filter << table[(int)(((value - ll)/diff * length) + .5)];
	  }
	  else
	    filter << nanColor->pixel;
	}
	else
	  filter << maskColor->pixel;
      }
      else
	filter << bgColor->pixel;

      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FramePseudoColor::psLevel2Mosaic(PSColorSpace mode,
				       int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned short* table = colorScale->psIndex();

  // update PS Matrix
  {
    Matrix mx = psMatrix(scale, width, height);

    FitsImage* sptr = currentFits;
    while (sptr) {
      sptr->updatePS(mx);
      sptr = sptr->nextMosaic();
    }

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* mptr = fm->current(); 
      while (mptr) {
	mptr->updatePS(mx);
	mptr = mptr->nextMosaic();
      }
      fm = fm->previous();
    }
  }

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		filter << table[0];
	      else if (value >= hh)
		filter << table[length];
	      else
		filter << table[(int)(((value - ll)/diff * length) + .5)];
	    }
	    else
	      filter << nanColor->pixel;
	  }
	  else
	    filter << maskColor->pixel;

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }

      if (!sptr)
	filter << bgColor->pixel;

      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}
