// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "frametrue.h"
#include "util.h"

FrameTrue::FrameTrue(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i, c, item)
{
  skipUpdate = 0;

  colormapXM = NULL;
  colormapPM = 0;
  colormapGCXOR = 0;

  memset(bgTrueColor_,255,4);
  memset(nanTrueColor_,255,4);
}

FrameTrue::~FrameTrue()
{
  if (colormapXM)
    XDestroyImage(colormapXM);

  if (colormapPM)
    Tk_FreePixmap(display, colormapPM);

  if (colormapGCXOR)
    XFreeGC(display, colormapGCXOR);
}

void FrameTrue::bgColorCmd(const char* color)
{
  if (bgColorName)
    delete [] bgColorName;
  bgColorName = dupstr(color);
  bgColor = getXColor(bgColorName);
  encodeTrueColor(bgColor, bgTrueColor_);
  update(BASE);
}

void FrameTrue::nanColorCmd(const char* color)
{
  if (nanColorName)
    delete [] nanColorName;
  nanColorName = dupstr(color);
  nanColor = getXColor(nanColorName);
  encodeTrueColor(nanColor, nanTrueColor_);
  update(BASE);
}

void FrameTrue::colormapBeginCmd()
{
  colormapBeginCmd(Vector(options->width,options->height)/2*widgetToCanvas);
}

void FrameTrue::rotateMotion()
{
  // be sure we have the correct color encoded
  encodeTrueColor(bgColor, bgTrueColor_);
  encodeTrueColor(nanColor, nanTrueColor_);

  // Rotate from src to dest
  Vector cc = Vector(options->width,options->height)/2;
  Matrix m = (Translate(-cc) * Rotate(rotation-rotateRotation) * 
	      Translate(cc)).invert();
  double* mm = m.mm();

  int& width = options->width;
  int& height = options->height;
  char* src = XImageData(rotateSrcXM);

  int bytesPerPixel = rotateDestXM->bits_per_pixel/8;

  for (int j=0; j<height; j++) {
    // the line may be padded at the end
    char* dest = XImageData(rotateDestXM) + j*rotateDestXM->bytes_per_line;

    for (int i=0; i<width; i++, dest+=bytesPerPixel) {
      double x = i*mm[0] + j*mm[3] + mm[6];
      double y = i*mm[1] + j*mm[4] + mm[7];

      if (x >= 0 && x < width && y >= 0 && y < height)
	memcpy(dest, 
	       src + ((int)y)*rotateDestXM->bytes_per_line+
	       ((int)x)*bytesPerPixel, 
	       bytesPerPixel);
      else
	memcpy(dest, bgTrueColor_, bytesPerPixel);
    }
  }

  // XImage to Pixmap
  XPutImage(display, rotatePM, gc, rotateDestXM,
	    0, 0, 0, 0, options->width, options->height);

  // Display Pixmap
  Vector d = Vector(0,0) * widgetToWindow;
  XCopyArea(display, rotatePM, Tk_WindowId(tkwin), rotateGCXOR, 0, 0, 
	    options->width, options->height, (int)d[0], (int)d[1]);
}

void FrameTrue::encodeTrueColor8(XColor* src, char* dest)
{
  int rs,gs,bs;
  unsigned char rm = decodeMask((unsigned char)visual->red_mask, &rs);
  unsigned char gm = decodeMask((unsigned char)visual->green_mask, &gs);
  unsigned char bm = decodeMask((unsigned char)visual->blue_mask, &bs);

  *dest = 
    ((((unsigned char)src->blue)  & bm) >> bs) | 
    ((((unsigned char)src->green) & gm) >> gs) | 
    ((((unsigned char)src->red)   & rm) >> rs);
}

void FrameTrue::encodeTrueColor16(XColor* src, char* dest)
{
  if (!baseXImage)
    return;

  int msb = baseXImage->byte_order;
  int rs,gs,bs;
#if !(_WIN32 || _MACOSX)
  unsigned short rm = decodeMask((unsigned short)visual->red_mask, &rs);
  unsigned short gm = decodeMask((unsigned short)visual->green_mask, &gs);
  unsigned short bm = decodeMask((unsigned short)visual->blue_mask, &bs);
#endif
#if _MACOSX
  // tk visual masks are bogus
  unsigned short rm = decodeMask((unsigned short)baseXImage->red_mask, &rs);
  unsigned short gm = decodeMask((unsigned short)baseXImage->green_mask, &gs);
  unsigned short bm = decodeMask((unsigned short)baseXImage->blue_mask, &bs);
#endif
#if _WIN32
  // windows masks are plain wrong
  unsigned short rm = decodeMask((unsigned short)0x7C00, &rs);
  unsigned short gm = decodeMask((unsigned short)0x03E0, &gs);
  unsigned short bm = decodeMask((unsigned short)0x001F, &bs);
#endif

  unsigned short r = (unsigned char)src->red;
  unsigned short g = (unsigned char)src->green;
  unsigned short b = (unsigned char)src->blue;
  unsigned short a = 0;
  a |= rs>0 ? ((r & rm) << rs) : ((r & rm) >> -rs);
  a |= gs>0 ? ((g & gm) << gs) : ((g & gm) >> -gs);
  a |= bs>0 ? ((b & bm) << bs) : ((b & bm) >> -bs);

  // we need to check to byteswap when we have cross platforms
  if ((!msb && lsb()) || (msb && !lsb()))
    memcpy(dest, &a, 2);
  else {
    unsigned char* rr = (unsigned char*)(&a);
    *(dest) = *(rr+1);
    *(dest+1) = *(rr);
  }
}

void FrameTrue::encodeTrueColor24_32(XColor* src, char* dest)
{
  if (!baseXImage)
    return;

  switch (baseXImage->bits_per_pixel) {
  case 24:
    encodeTrueColor24(src,dest);
    break;
  case 32:
    encodeTrueColor32(src,dest);
    break;
  }
}

void FrameTrue::encodeTrueColor24(XColor* src, char* dest)
{
  int msb = baseXImage->byte_order;
#if !(_WIN32 || _MACOSX)
  int rs = decodeMask((unsigned long)visual->red_mask);
  int gs = decodeMask((unsigned long)visual->green_mask);
  int bs = decodeMask((unsigned long)visual->blue_mask);
#endif
#if _MACOSX
  // tk visual masks are bogus
  int rs = decodeMask((unsigned long)baseXImage->red_mask);
  int gs = decodeMask((unsigned long)baseXImage->green_mask);
  int bs = decodeMask((unsigned long)baseXImage->blue_mask);
#endif
#if _WIN32
  // windows masks are plain wrong
  int rs = decodeMask((unsigned long)0x00FF0000);
  int gs = decodeMask((unsigned long)0x0000FF00);
  int bs = decodeMask((unsigned long)0x000000FF);
#endif
      
  unsigned int r = (unsigned char)src->red;
  unsigned int g = (unsigned char)src->green;
  unsigned int b = (unsigned char)src->blue;
  unsigned int a = 0;
  a |= r << rs;
  a |= g << gs;
  a |= b << bs;

  // we need to check to byteswap when we have cross platforms
  if ((!msb && lsb()) || (msb && !lsb()))
    memcpy(dest, &a, 3);
  else {
    unsigned char* rr = (unsigned char*)(&a);
    *(dest) = *(rr+3);
    *(dest+1) = *(rr+2);
    *(dest+2) = *(rr+1);
  }
}

void FrameTrue::encodeTrueColor32(XColor* src, char* dest)
{
  int msb = baseXImage->byte_order;
#if !(_WIN32 || _MACOSX)
  int rs = decodeMask((unsigned long)visual->red_mask);
  int gs = decodeMask((unsigned long)visual->green_mask);
  int bs = decodeMask((unsigned long)visual->blue_mask);
#endif
#if _MACOSX
  // tk visual masks are bogus
  int rs = decodeMask((unsigned long)baseXImage->red_mask);
  int gs = decodeMask((unsigned long)baseXImage->green_mask);
  int bs = decodeMask((unsigned long)baseXImage->blue_mask);
#endif
#if _WIN32
  // windows masks are plain wrong
  int rs = decodeMask((unsigned long)0x00FF0000);
  int gs = decodeMask((unsigned long)0x0000FF00);
  int bs = decodeMask((unsigned long)0x000000FF);
#endif
      
  unsigned int r = (unsigned char)src->red;
  unsigned int g = (unsigned char)src->green;
  unsigned int b = (unsigned char)src->blue;
  unsigned int a = 0;
  a |= r << rs;
  a |= g << gs;
  a |= b << bs;

  // we need to check to byteswap when we have cross platforms
  if ((!msb && lsb()) || (msb && !lsb()))
    memcpy(dest, &a, 4);
  else {
    unsigned char* rr = (unsigned char*)(&a);
    *(dest) = *(rr+3);
    *(dest+1) = *(rr+2);
    *(dest+2) = *(rr+1);
    *(dest+3) = *(rr);
  }
}

