// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "tcl.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "frametruecolor.h"
#include "fitsimage.h"
#include "util.h"
#include "ps.h"
#include "NaN.h"

FrameTrueColor::FrameTrueColor(Tcl_Interp* i, Tk_Canvas c, Tk_Item* item)
  : FrameBase(i,c,item), Frame(i,c,item), FrameTrue(i,c,item)
{
  colormapData = NULL;
}

FrameTrueColor::~FrameTrueColor()
{
  if (colormapData)
    delete [] colormapData;
}

void FrameTrueColor::fillXImageSingle(XImage* xmap,
				      int x0, int y0, int x1, int y1,
				      double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!currentFits)
    return;

  int& height = xmap->height;
  int bytesPerPixel = xmap->bits_per_pixel/8;
  char* data = XImageData(xmap);

  double* mm = (currentFits->*getMatrix)();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;
  const unsigned char* table = colorScale->colors();

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (long jj=y0; jj<y1; jj++) {
    // the line may be padded at the end
    char* dest = data + jj*xmap->bytes_per_line + x0*bytesPerPixel;

    for (long ii=x0; ii<x1; ii++, dest+=bytesPerPixel) {

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	char maskTrueColor[4];
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		encodeTrueColor(fm->color(), maskTrueColor);
	    }
	    fm = fm->previous();
	  }
	}

	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      memcpy(dest, table, bytesPerPixel);
	    else if (value >= hh)
	      memcpy(dest, table+last, bytesPerPixel);
	    else
	      memcpy(dest, table+((int)(((value - ll)/diff * length) + .5))*
		     bytesPerPixel,bytesPerPixel);
	  }
	  else
	    memcpy(dest, nanTrueColor_, bytesPerPixel);
	} else
	  memcpy(dest, maskTrueColor, bytesPerPixel);
      }
      else {
	memcpy(dest, bgTrueColor_, bytesPerPixel);
      }
    }
  }
}

void FrameTrueColor::fillXImageMosaic(XImage* xmap, 
				      int x0, int y0, int x1, int y1,
				      double* (FitsImage::*getMatrix)())
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int& height = xmap->height;
  int bytesPerPixel = xmap->bits_per_pixel/8;
  char* data = XImageData(xmap);

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->colors();
  int last = length * bytesPerPixel;

  for (long jj=y0; jj<y1; jj++) {

    // the line may be padded at the end
    char* dest = data + jj*xmap->bytes_per_line + x0*bytesPerPixel;

    for (long ii=x0; ii<x1; ii++, dest+=bytesPerPixel) {
      memcpy(dest, bgTrueColor_, bytesPerPixel);

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = (sptr->*getMatrix)();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  char maskTrueColor[4];
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  encodeTrueColor(fm->color(), maskTrueColor);
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		memcpy(dest, table, bytesPerPixel);
	      else if (value >= hh)
		memcpy(dest, table+last, bytesPerPixel);
	      else
		memcpy(dest,
	          table+((int)(((value-ll)/diff*length)+.5))*bytesPerPixel,
		  bytesPerPixel);
	    }
	    else 
	      memcpy(dest, nanTrueColor_, bytesPerPixel);
	  }
	  else
	      memcpy(dest, maskTrueColor, bytesPerPixel);

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }
    }
  }
}

void FrameTrueColor::colormapCmd(int id, float b, float c, int i, 
				 unsigned short* index, unsigned char* cells, 
				 int cnt)
{
  // first check for change
  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  if (mosaicCount) {
    update(BASE); // always update
    skipUpdate = 0;
  }
  else {
    if (!skipUpdate) {
      update(BASE); // update once
      skipUpdate = 1;
    }      
  }
}

void FrameTrueColor::colormapBeginCmd(const Vector& w)
{
  // we need some fits data
  // we assume the colorScale length will not change during motion calls
  if (!currentFits)
    return;

  int width = options->width;
  int height = options->height;
  cmapOffset = Vector();
  if (cmapArea) {
    width = height = cmapArea;

    switch (cmapMode) {
    case CENTER:
      cmapOffset = (Vector(options->width,options->height) - 
		    Vector(width,height))/2;
      break;
    case CLICK: 
      {
	Vector where = w * canvasToWidget;
	if (where[0] >= 0 && where[0] < options->width &&
	    where[1] >= 0 && where[1] < options->height)
	  cmapOffset = where - Vector(width,height)/2;
	else
	  cmapOffset = (Vector(options->width,options->height) - 
			Vector(width,height))/2;
      }
      break;
    }

    if (cmapOffset[0] < 0) {
      width += (int)cmapOffset[0];
      cmapOffset[0] = 0;
    }
    if (cmapOffset[0]+width >= options->width)
      width = options->width - (int)cmapOffset[0];

    if (cmapOffset[1] < 0) {
      height += (int)cmapOffset[1];
      cmapOffset[1] = 0;
    }

    if (cmapOffset[1]+height >= options->height)
      height = options->height - (int)cmapOffset[1];
  }

  // Create XImage
  if (!(colormapXM = XGetImage(display, pixmap, 0, 0, 
			       width, height, AllPlanes, ZPixmap))) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap XImage" 
	 << endl;
    exit(1);
  }

  // Create Pixmap
  colormapPM = 
    Tk_GetPixmap(display, Tk_WindowId(tkwin), width, height, depth);
  if (!colormapPM) {
    cerr << "FrameTrueColor Internal Error: Unable to Create Colormap Pixmap" 
	 << endl;
    exit(1);
  }

  // colormapGCXOR
  colormapGCXOR = XCreateGC(display, Tk_WindowId(tkwin), 0, NULL);

  // Create table index array
  if (colormapData)
    delete [] colormapData;
  colormapData = new long[width*height];
  if (!colormapData) {
    cerr << "FrameTrueColor Internal Error: Unable to alloc tmp data array" 
	 << endl;
    exit(1);
  }

  // fill data array
  if (!isMosaic()) 
    colormapBeginSingle(width, height);
  else
    colormapBeginMosaic(width, height);
}

void FrameTrueColor::colormapBeginSingle(int width, int height)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  double* mm = currentFits->getmWidgetToData();
  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  long* dest = colormapData;
  for (long jj=yoff; jj<yoff+height; jj++) {
    for (long ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -3; // initialize

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask)
	      msk = currentMask->getValueMask(xx,yy);
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll)
	      *dest = 0;
	    else if (value >= hh)
	      *dest = last;
	    else
	      *dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	  }
	  else
	    *dest = -1;
	}
	else
	  *dest = -2;
      }
    }
  }
}

void FrameTrueColor::colormapBeginMosaic(int width, int height)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  int xoff = (int)cmapOffset[0];
  int yoff = (int)cmapOffset[1];

  int bytesPerPixel = colormapXM->bits_per_pixel/8;

  int length = colorScale->size() - 1;
  int last = length * bytesPerPixel;

  // scan fits

  long* dest = colormapData;
  for (long jj=yoff; jj<yoff+height; jj++) {
    for (long ii=xoff; ii<xoff+width; ii++, dest++) {
      *dest = -3;

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = sptr->getmWidgetToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr)
		msk = mptr->getValueMask(xx,yy);
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll)
		*dest = 0;
	      else if (value >= hh)
		*dest = last;
	      else
		*dest = (int)(((value - ll)/diff * length) + .5)*bytesPerPixel;
	    }
	    else
	      *dest = -1;
	  }
	  else
	    *dest = -2;

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }
    }
  }
}

void FrameTrueColor::colormapMotionCmd(int id, float b, float c, int i, 
				       unsigned short* index, 
				       unsigned char* cells, int cnt)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  // first check for change
  if (cmapID == id && bias == b && contrast == c && invert == i && colorCells)
    return;

  // we got a change
  cmapID = id;
  bias = b;
  contrast = c;
  invert = i;

  updateColorCells(index, cells, cnt);
  updateColorScale();

  // if we have no data, stop now
  if (!currentFits) 
    return;

  // clear ximage
  int& width = colormapXM->width;
  int& height = colormapXM->height;
  char* data = XImageData(colormapXM);
  int bytesPerPixel = colormapXM->bits_per_pixel/8;
  int& bytesPerLine = colormapXM->bytes_per_line;

  const unsigned char* table = colorScale->colors();

  long* src = colormapData;
  for (long jj=0; jj<height; jj++) {
    // line may be padded at end
    char* dest = data + jj*bytesPerLine;

    for (long ii=0; ii<width; ii++, src++, dest+=bytesPerPixel)
      switch (*src) {
      case -1:
	memcpy(dest, nanTrueColor_, bytesPerPixel);
	break;
      case -2:
	// can't do anything here, mask can be any color
	break;
      case -3:
	memcpy(dest, bgTrueColor_, bytesPerPixel);
	break;
      default:
	memcpy(dest, table+(*src), bytesPerPixel);
	break;
      }
  }

  // XImage to Pixmap
  XPutImage(display, colormapPM, gc, colormapXM, 0, 0, 0, 0, width, height);

  // Display Pixmap
  Vector d = cmapOffset * widgetToWindow;
  XCopyArea(display, colormapPM, Tk_WindowId(tkwin), colormapGCXOR, 0, 0, 
	    width, height, (int)d[0], (int)d[1]);

  // update panner
  updatePanner();
}

void FrameTrueColor::colormapEndCmd(int id, float b, float c, int i, 
				    unsigned short* index, 
				    unsigned char* cells, int cnt)
{
  if (colormapXM) {
    XDestroyImage(colormapXM);
    colormapXM = NULL;
  }

  if (colormapPM) {
    Tk_FreePixmap(display, colormapPM);
    colormapPM = 0;
  }

  if (colormapGCXOR) {
    XFreeGC(display, colormapGCXOR);
    colormapGCXOR = 0;
  }

  if (colormapData) {
    delete [] colormapData;
    colormapData = NULL;
  }

  update(BASE); // always update
}

void FrameTrueColor::psLevel2(PSColorSpace mode, int width, int height, 
			      float scale)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  psLevel2HeadTrue(mode, width, height);

  if (!isMosaic())
    psLevel2Single(mode, width, height, scale);
  else
    psLevel2Mosaic(mode, width, height, scale);
}

void FrameTrueColor::psLevel2Single(PSColorSpace mode,
				    int width, int height, float scale)
{
  if (!currentFits)
    return;

  //NoCompressAsciiHex filter;
  //NoCompressAscii85 filter;

  //RLEAsciiHex filter;
  RLEAscii85 filter;

  //GZIPAsciiHex filter;
  //GZIPAscii85 filter;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  currentFits->updatePS(psMatrix(scale, width, height));

  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	  }
	  else {
	    blue = nanColor->blue;
	    green = nanColor->green;
	    red = nanColor->red;
	  }
	}
	else {
	  blue = maskColor->blue;
	  green = maskColor->green;
	  red = maskColor->red;
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FrameTrueColor::psLevel2Mosaic(PSColorSpace mode, 
				    int width, int height, float scale)
{
  if (!currentFits)
    return;

  RLEAscii85 filter;
  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  {
    Matrix mx = psMatrix(scale, width, height);

    FitsImage* sptr = currentFits;
    while (sptr) {
      sptr->updatePS(mx);
      sptr = sptr->nextMosaic();
    }

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* mptr = fm->current(); 
      while (mptr) {
	mptr->updatePS(mx);
	mptr = mptr->nextMosaic();
      }
      fm = fm->previous();
    }
  }

  for (long jj=0; jj<height; jj++) {
    ostringstream str;

    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  }
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }

      switch (mode) {
      case BW:
      case GRAY:
	filter << RGB2Gray(red, green, blue);
	break;
      case RGB:
	filter << red << green << blue;
	break;
      case CMYK:
	{
	  unsigned char cyan, magenta, yellow, black;
	  RGB2CMYK(red, green, blue, &cyan, &magenta, &yellow, &black);
	  filter << cyan << magenta << yellow << black;
	}
	break;
      }
      str << filter;
    }
    str << ends;
    psFix(str);
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  ostringstream str;
  filter.flush(str);
  psFix(str);
  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

#ifdef _MACOSX
void FrameTrueColor::macosx(float scale, int width, int height, 
			    const Vector& v, const Vector& s)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!isMosaic())
    macosxSingle(scale, width, height, v, s);
  else
    macosxMosaic(scale, width, height, v, s);
}

void FrameTrueColor::macosxSingle(float scale, int width, int height, 
				  const Vector& v, const Vector& s)
{
  if (!currentFits)
    return;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  currentFits->updatePS(psMatrix(scale, width, height));
  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  unsigned char* dst = new unsigned char[width*height*4];
  unsigned char* dptr = dst;

  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	  }
	  else {
	    blue = nanColor->blue;
	    green = nanColor->green;
	    red = nanColor->red;
	  }
	}
	else {
	  blue = maskColor->blue;
	  green = maskColor->green;
	  red = maskColor->red;
	}
      }

      *dptr++ = red;
      *dptr++ = green;
      *dptr++ = blue;
      *dptr++ = 0;
    }
  }

  macosxBitmapCreate(dst, width, height, v, s);

  if (dst)
    delete dst;
}

void FrameTrueColor::macosxMosaic(float scale, int width, int height, 
				  const Vector& v, const Vector& s)
{
  if (!currentFits)
    return;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  {
    Matrix mx = psMatrix(scale, width, height);

    FitsImage* sptr = currentFits;
    while (sptr) {
      sptr->updatePS(mx);
      sptr = sptr->nextMosaic();
    }

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* mptr = fm->current(); 
      while (mptr) {
	mptr->updatePS(mx);
	mptr = mptr->nextMosaic();
      }
      fm = fm->previous();
    }
  }

  unsigned char* dst = new unsigned char[width*height*4];
  unsigned char* dptr = dst;

  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  }
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }

      *dptr++ = red;
      *dptr++ = green;
      *dptr++ = blue;
      *dptr++ = 0;
    }
  }

  macosxBitmapCreate(dst, width, height, v, s);

  if (dst)
    delete dst;
}

#endif

#ifdef _WIN32
void FrameTrueColor::win32(float scale, int width, int height, 
			    const Vector& v, const Vector& s)
{
  // we need a colorScale before we can render
  if (!validColorScale())
    return;

  if (!isMosaic())
    win32Single(scale, width, height, v, s);
  else
    win32Mosaic(scale, width, height, v, s);
}

void FrameTrueColor::win32Single(float scale, int width, int height, 
				  const Vector& v, const Vector& s)
{
  if (!currentFits)
    return;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  currentFits->updatePS(psMatrix(scale, width, height));
  double* mm = currentFits->getmPSToData();

  int* params = currentFits->getDataParams(frScale.scanMode());
  int& srcw = params[0];
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double ll = currentFits->getLowDouble();
  double hh = currentFits->getHighDouble();
  double diff = hh - ll;

  unsigned char* dst = new unsigned char[width*height*3];
  unsigned char* dptr = dst;

  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      double xx = ii*mm[0] + jj*mm[3] + mm[6];
      double yy = ii*mm[1] + jj*mm[4] + mm[7];

      if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	double value = currentFits->getValueDouble(long(yy)*srcw + long(xx));

	int msk =0;
	XColor* maskColor;
	if (showMask) {
	  FitsMask* fm = mask.tail();
	  while (fm && !msk) {
	    FitsImage* currentMask = fm->current(); 
	    if (currentMask) {
	      msk = currentMask->getValueMask(xx,yy);
	      if (msk)
		maskColor = fm->color();
	    }
	    fm = fm->previous();
	  }
	}
       
	if (!msk) {
	  if (!isnand(value)) {
	    if (value <= ll) {
	      blue = table[0];
	      green = table[1];
	      red = table[2];
	    }
	    else if (value >= hh) {
	      blue = table[length*3];
	      green = table[length*3+1];
	      red = table[length*3+2];
	    }
	    else {
	      int l = (int)(((value - ll)/diff * length) + .5);
	      blue = table[l*3];
	      green = table[l*3+1];
	      red = table[l*3+2];
	    }
	  }
	  else {
	    blue = nanColor->blue;
	    green = nanColor->green;
	    red = nanColor->red;
	  }
	}
	else {
	  blue = maskColor->blue;
	  green = maskColor->green;
	  red = maskColor->red;
	}
      }

      *dptr++ = blue;
      *dptr++ = green;
      *dptr++ = red;
    }
  }

  win32BitmapCreate(dst, width, height, v, s);

  if (dst)
    delete dst;
}

void FrameTrueColor::win32Mosaic(float scale, int width, int height, 
				  const Vector& v, const Vector& s)
{
  if (!currentFits)
    return;

  int length = colorScale->size() - 1;
  const unsigned char* table = colorScale->psColors();

  // update PS Matrix
  {
    Matrix mx = psMatrix(scale, width, height);

    FitsImage* sptr = currentFits;
    while (sptr) {
      sptr->updatePS(mx);
      sptr = sptr->nextMosaic();
    }

    FitsMask* fm = mask.tail();
    while (fm) {
      FitsImage* mptr = fm->current(); 
      while (mptr) {
	mptr->updatePS(mx);
	mptr = mptr->nextMosaic();
      }
      fm = fm->previous();
    }
  }

  unsigned char* dst = new unsigned char[width*height*3];
  unsigned char* dptr = dst;

  for (long jj=0; jj<height; jj++) {
    for (long ii=0; ii<width; ii++) {
      unsigned char red = bgColor->red;
      unsigned char green = bgColor->green;
      unsigned char blue = bgColor->blue;

      FitsImage* sptr = currentFits;
      FitsMask* fm = mask.tail();
      while(fm) {
	fm->initMosaic();
	fm = fm->previous();
      }

      while (sptr) {
	double* mm = sptr->getmPSToData();
	double xx = ii*mm[0] + jj*mm[3] + mm[6];
	double yy = ii*mm[1] + jj*mm[4] + mm[7];

	int* params = sptr->getDataParams(frScale.scanMode());
	int& srcw = params[0];
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (xx>=xmin && xx<xmax && yy>=ymin && yy<ymax) {
	  double value = sptr->getValueDouble(long(yy)*srcw + long(xx));

	  int msk =0;
	  XColor* maskColor;
	  if (showMask) {
	    FitsMask* fm = mask.tail();
	    while(fm && !msk) {
	      FitsImage* mptr = fm->mptr();
	      if (mptr) {
		msk = mptr->getValueMask(xx,yy);
		if (msk)
		  maskColor = fm->color();
	      }
	      fm = fm->previous();
	    }
	  }

	  if (!msk) {
	    if (!isnand(value)) {
	      double ll = sptr->getLowDouble();
	      double hh = sptr->getHighDouble();
	      double diff = hh - ll;

	      if (value <= ll) {
		blue = table[0];
		green = table[1];
		red = table[2];
	      }
	      else if (value >= hh) {
		blue = table[length*3];
		green = table[length*3+1];
		red = table[length*3+2];
	      }
	      else {
		int l = (int)(((value - ll)/diff * length) + .5);
		blue = table[l*3];
		green = table[l*3+1];
		red = table[l*3+2];
	      }
	    }
	    else {
	      blue = nanColor->blue;
	      green = nanColor->green;
	      red = nanColor->red;
	    }
	  }
	  else {
	    blue = maskColor->blue;
	    green = maskColor->green;
	    red = maskColor->red;
	  }

	  break;
	}
	else {
	  sptr = sptr->nextMosaic();
	  FitsMask* fm = mask.tail();
	  while(fm) {
	    fm->nextMosaic();
	    fm = fm->previous();
	  }
	}
      }

      *dptr++ = blue;
      *dptr++ = green;
      *dptr++ = red;
    }
  }

  win32BitmapCreate(dst, width, height, v, s);

  if (dst)
    delete dst;
}

#endif
