// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framebase.h"
#include "fitsimage.h"
#include "NaN.h"
#include "blt.h"

void FrameBase::bltHist(char* xname, char* yname)
{
  int num = 256;

  if (!currentFits)
    return;

  if (DebugPerf)
    cerr << "bltHist...";

  double* histogramX = currentScale->histogramX();
  double* histogramY = currentScale->histogramY();

  if (!histogramX || !histogramY) {
    histogramX = currentScale->initHistogramX(num);
    histogramY = currentScale->initHistogramY(num);

    for (int i=0; i<num; i++)
      histogramX[i] = (double)i/(num-1)*
	(currentScale->max()-currentScale->min()) + currentScale->min();
    
    for (int j=0; j<num; j++)
      histogramY[j] = 0;

    if (currentScale->max() > currentScale->min()) {
      FitsImage* ptr = *channelFits;
      while (ptr) {
	FitsImage* sptr = ptr;
	while (sptr) {
	  sptr->bin(histogramY, num, currentScale->min(), currentScale->max(), 
		    currentScale->scanMode());
	  sptr = sptr->nextSlice();
	}
	ptr = ptr->nextMosaic();
      }
    }
  }

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, histogramX, num, num*sizeof(double), TCL_STATIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, histogramY, num, num*sizeof(double), TCL_STATIC) 
      != TCL_OK)
    goto error;

  if (DebugPerf)
    cerr << "end" << endl;

  return;

 error:
  result = TCL_ERROR;
  return;
}

void FrameBase::bltProjection(char* xname, char* yname, 
			      char* xcname, char* ycname,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  if (!currentFits)
    return;

  // the number of pixels sampled is the greater of either dimension
  Vector ap1 = p1;
  Vector ap2 = p2;
  int xnum = ::abs(ap2[0]-ap1[0]);
  int ynum = ::abs(ap2[1]-ap1[1]);
  int num = (xnum > ynum ? xnum : ynum)+1;

  double* x = (double*)malloc(num*sizeof(double));
  double* y = (double*)malloc(num*sizeof(double));
  double* xc = (double*)malloc(num*sizeof(double));
  double* yc = (double*)malloc(num*sizeof(double));

  if (!isMosaic()) 
    bltProjSingle(x, y, xc, yc, num, sys, sky, p1, p2, width, avg);
  else
    bltProjMosaic(x, y, xc, yc, num, sys, sky, p1, p2, width, avg);

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yy, y, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* xxc;
  if (Blt_GetVector(interp, xcname, &xxc) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xxc, xc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  Blt_Vector* yyc;
  if (Blt_GetVector(interp, ycname, &yyc) != TCL_OK) 
    goto error;

  if (Blt_ResetVector(yyc, yc, num, num*sizeof(double), TCL_DYNAMIC) 
      != TCL_OK)
    goto error;

  return;

 error:
  result = TCL_ERROR;
  return;
}

void FrameBase::bltProjSingle(double* x, double* y, 
			      double* xc, double* yc, int num,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  int* cnt = new int[num];
  Vector s = (p2-p1)/(num-1);

  FitsImage* ptr = currentFits;
  Matrix mm = ptr->getRefToData();

  for (long ii=0; ii<num; ii++) {
    Vector t = p1 + s*ii;

    x[ii] = ii+1;
    Vector tt = ptr->mapFromRef(t,sys,sky);
    xc[ii] = tt[0];
    yc[ii] = tt[1];

    // check for nan
    double v = ptr->getValueDouble(t*mm);
    if (!isnand(v)) {
      y[ii] = v;
      cnt[ii] = 1;
    }
    else
      y[ii] = getnand();

    Vector ps = (p2-p1).normalize();
    Vector ss = Vector(-ps[1],ps[0]);

    for (long jj=1; jj<width; jj++) {
      Vector tt = p1 + s*ii + ss*jj;
      double value = ptr->getValueDouble(tt*mm);
      // check for nan
      if (!isnand(value) && !isnand(y[ii])) {
	y[ii] += value;
	cnt[ii]++;
      }
      else
	y[ii] = getnand();
    }
  }

  // average if needed
  if (avg)
    for (long ii=0; ii<num; ii++)
      if (!isnand(y[ii]))
	y[ii] /= cnt[ii];

  // convert nand's to low
  for (long jj=0; jj<num; jj++)
    if (isnand(y[jj]))
      y[jj] = ptr->getLowDouble();

  delete [] cnt;
}

void FrameBase::bltProjMosaic(double* x, double* y, 
			      double* xc, double* yc, int num,
			      CoordSystem sys, SkyFrame sky,
			      const Vector& p1, const Vector& p2, 
			      double width, int avg)
{
  int* cnt = new int[num];
  Vector s = (p2-p1)/(num-1);

  for (long ii=0; ii<num; ii++) {
    Vector t = p1 + s*ii;

    x[ii] = ii+1;

    FitsImage* sptr = currentFits;
    while (sptr) {
      Vector tt = sptr->mapFromRef(t,sys,sky);
      xc[ii] = tt[0];
      yc[ii] = tt[1];

      int* params = sptr->getDataParams(currentScale->scanMode());
      int& srcw = params[0];
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      Vector z = t * sptr->getRefToData();

      if (z[0]>=xmin && z[0]<xmax && z[1]>=ymin && z[1]<ymax) {

	// check for nan
	double v = sptr->getValueDouble(z);
	if (!isnand(v)) {
	  y[ii] = v;
	  cnt[ii] = 1;
	}
	else
	  y[ii] = getnand();

	Vector ps = (p2-p1).normalize();
	Vector ss = Vector(-ps[1],ps[0]);

	for (long jj=1; jj<width; jj++) {
	  Vector tt = p1 + s*ii + ss*jj;

	  FitsImage* spptr = *channelFits;
	  while (spptr) {
	    int* pparams = spptr->getDataParams(currentScale->scanMode());
	    int& ssrcw = pparams[0];
	    int& xxmin = pparams[1];
	    int& xxmax = pparams[2];
	    int& yymin = pparams[3];
	    int& yymax = pparams[4];

	    Vector zz = tt * spptr->getRefToData();
		
	    if (zz[0]>=xxmin && zz[0]<xxmax && 
		zz[1]>=yymin && zz[1]<yymax) {

	      double vvalue = spptr->getValueDouble(zz);
	      // check for nan
	      if (!isnand(vvalue) && !isnand(y[ii])) {
		y[ii] += vvalue;
		cnt[ii]++;
	      }
	      else
		y[ii] = getnand();

	      break;
	    }
	    else
	      spptr = spptr->nextMosaic();
	  }

	  if (!spptr)
	    y[ii] = getnand();
	}

	break;
      }
      else
	sptr = sptr->nextMosaic();
    }

    if (!sptr)
      y[ii] = getnand();
  }

  // average if needed
  if (avg)
    for (long ii=0; ii<num; ii++)
      if (!isnand(y[ii]))
	y[ii] /= cnt[ii];

      // convert nand's to low
  for (int j=0; j<num; j++)
    if (isnand(y[j]))
      y[j] = currentScale->low();

  delete [] cnt;
}

void FrameBase::bltCut(char* xname, char* yname, 
		       Orientation axis, const Vector& rr)
{
  if (!currentFits)
    return;

  // Vector r is in Widget coords
  Vector r = rr;

  int size;
  if (axis == XX)
    size = options->width;
  else
    size = options->height;

  long length = (size+1) * 2;
  double* x = (double*)malloc(sizeof(double)*length);
  double* y = (double*)malloc(sizeof(double)*length);

  if (*currentMosaicCount == 0 && *currentSliceCount == 0) {
    for (int i=0; i<=size; i++) {
      x[i*2] = i;
      x[i*2+1] = i;
      y[i*2] = 0;
      y[i*2+1] = 0;
    }
  }
  else {
    if (!isMosaic()) 
      bltCutSingle(x, y, size, r, axis);
    else
      bltCutMosaic(x, y, size, r, axis);
  }

  Blt_Vector* xx;
  if (Blt_GetVector(interp, xname, &xx) != TCL_OK)
    goto error;

  if (Blt_ResetVector(xx, x, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  Blt_Vector* yy;
  if (Blt_GetVector(interp, yname, &yy) != TCL_OK)
    goto error;

  if (Blt_ResetVector(yy, y, length, length*sizeof(double), TCL_DYNAMIC) != 
      TCL_OK)
    goto error;

  return;

 error:
    result = TCL_ERROR;
    return;
}

void FrameBase::bltCutSingle(double* x, double* y, int size, Vector& r, 
			 Orientation axis)
{
  Matrix m = currentFits->getWidgetToData();
  int* params = currentFits->getDataParams(currentScale->scanMode());
  int& xmin = params[1];
  int& xmax = params[2];
  int& ymin = params[3];
  int& ymax = params[4];

  double prev = currentScale->low();

  for (int i=0; i<=size; i++) {
    double v = currentScale->low();

    Vector img;
    if (axis == XX)
      img = Vector(1+i,r[1]) * m;
    else
      img = Vector(r[0],1+i) * m;

    if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
      double value = currentFits->getValueDouble(img);

      if (!isnand(value))
	v = value;
    }

    x[2*i] = i;
    x[2*i +1] = i;
    
    y[2*i] = prev;
    y[2*i +1] = v;
    prev = v;
  }
}

void FrameBase::bltCutMosaic(double* x, double* y, int size, Vector& r, 
			 Orientation axis)
{
  double prev = currentScale->low();

  for (int ii=0; ii<=size; ii++) {
    double v = currentScale->low();

    // find y first
    FitsImage* sptr = currentFits;
    while (sptr) {
      Matrix& mx = sptr->getWidgetToData();
      int* params = sptr->getDataParams(currentScale->scanMode());
      int& xmin = params[1];
      int& xmax = params[2];
      int& ymin = params[3];
      int& ymax = params[4];

      Vector img;
      if (axis == XX)
	img = Vector(1+ii,r[1]) * mx;
      else
	img = Vector(r[0],1+ii) * mx;

      if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	double value = sptr->getValueDouble(img);

	if (!isnand(value))
	  v = value;

	break;
      }
      else
	sptr = sptr->nextMosaic();
    }

    x[2*ii] = ii;
    x[2*ii +1] = ii;
    
    y[2*ii] = prev;
    y[2*ii +1] = v;
    prev = v;
  }
}

