// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framebase.h"
#include "fitsimage.h"
#include "util.h"

// Map Point

Vector FrameBase::mapFromRef(const Vector& v, InternalSystem sys)
{
  switch (sys) {
  case CANVAS:
    return v * refToCanvas;
  case PANNER:
    return v * refToPanner;
  }
}

Vector FrameBase::mapToRef(const Vector& v, InternalSystem sys)
{
  switch (sys) {
  case CANVAS:
    return v * canvasToRef;
  case PANNER:
    return v * pannerToRef;
  }
}

// Map Length

double FrameBase::mapLenFromRef(double d, InternalSystem sys)
{
  Vector r = mapLenFromRef(Vector(d,0),sys);
  return r[0];
}

Vector FrameBase::mapLenFromRef(const Vector& v, InternalSystem sys)
{
  switch (sys) {
  case CANVAS:
    return mapLen(v,refToCanvas);
  case PANNER:
    return mapLen(v,refToPanner);
  }
}

double FrameBase::mapLenToRef(double d, InternalSystem sys)
{
  Vector r = mapLenToRef(Vector(d,0),sys);
  return r[0];
}

Vector FrameBase::mapLenToRef(const Vector& v, InternalSystem sys)
{
  switch (sys) {
  case CANVAS:
    return mapLen(v,canvasToRef);
  case PANNER:
    return mapLen(v,pannerToRef);
  }
}

// Map Distance

double FrameBase::mapDistFromRef(const Vector& v1, const Vector& v2, 
				 InternalSystem sys)
{
  switch (sys) {
  case CANVAS:
    return (v2*refToCanvas - v1*refToCanvas).length();
  case PANNER:
    return (v2*refToPanner - v1*refToPanner).length();
  }
}

// Map Angle

double FrameBase::mapAngleFromRef(double angle, CoordSystem sys, SkyFrame sky)
{
  double r = angle;

  switch (sys) {
  case IMAGE: 
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    break;
  default:
    switch (getWCSOrientation(sys,sky)) {
    case NORMAL:
      r = angle + getWCSRotation(sys,sky);
      break;
    case XX:
      r = -angle - getWCSRotation(sys,sky) + M_PI;
      break;
    }
  }

  if (r>0)
    while (r>=2*M_PI)
      r -= 2*M_PI;
  else
    while (r<0)
      r += 2*M_PI;

  return r;
}

double FrameBase::mapAngleToRef(double angle, CoordSystem sys, SkyFrame sky)
{
  double r = angle;

  switch (sys) {
  case IMAGE: 
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    break;
  default:
    switch (getWCSOrientation(sys,sky)) {
    case NORMAL:
      r = angle - getWCSRotation(sys,sky);
      break;
    case XX:
      r = -angle - getWCSRotation(sys,sky) + M_PI;
      break;
    }
  }

  if (r>0)
    while (r>=2*M_PI)
      r -= 2*M_PI;
  else
    while (r<0)
      r += 2*M_PI;

  return r;
}

void FrameBase::calcAlignWCS(FitsImage* fits1, 
			     CoordSystem sys, SkyFrame sky,
			     Orientation* orientation, Matrix* oo,
			     double* rotation)
{
  // init
  *orientation = NORMAL;
  oo->identity();
  *rotation = 0;

  if (!fits1 || !fits1->hasWCS(sys))
    return;

  // orientation
  *orientation = fits1->getWCSOrientation(sys,sky);
  switch (*orientation) {
  case NORMAL:
    oo->identity();
    break;
  case XX:
    *oo = FlipX();
    break;
  case YY:
    *oo = FlipY();
    break;
  case XY:
    *oo = FlipXY();
    break;
  }

  // rotation
  *rotation = fits1->getWCSRotation(sys,sky);
}

void FrameBase::calcAlignWCS(FitsImage* fits1, FitsImage* fits2, 
			     CoordSystem sys, SkyFrame sky,
			     Orientation* orientation, Matrix* oo,
			     double* rotation, Vector* zoom)
{
  // init
  *orientation = NORMAL;
  oo->identity();
  *rotation = 0;

  if ((!fits1 || !fits2) || 
      (fits1 == fits2) ||
      !(fits1->hasWCS(sys)) ||
      !(fits2->hasWCS(sys)))
    return;

  Vector org1 = fits1->getWCScrpix(sys);
  Vector orval1 = fits1->pix2wcs(org1,sys,sky);

  // orientation
  *orientation = fits2->getWCSOrientation(sys,sky);
  switch (*orientation) {
  case NORMAL:
    oo->identity();
    break;
  case XX:
    *oo = FlipX();
    break;
  case YY:
    *oo = FlipY();
    break;
  case XY:
    *oo = FlipXY();
    break;
  }

  // rotation
  Vector orpix = fits2->wcs2pix(orval1, sys, sky);
  Vector delta = fits2->getWCScdelt(sys).abs();
  Vector npix = 
    fits2->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
  Vector north = npix - orpix;

  if ((fits2->getWCS(sys))->imflip)
    *rotation = (north.angle()-M_PI/2);
  else
    *rotation = -(north.angle()-M_PI/2);

  // zoom
  Vector cd1 = fits1->getWCScdelt(sys);
  Vector cd2 = fits2->getWCScdelt(sys);
  *zoom /= Vector(cd1[0]/cd2[0], cd1[1]/cd2[1]).abs();
}

Matrix FrameBase::calcAlignWCS(FitsImage* fits1, FitsImage* fits2, 
			       CoordSystem sys, SkyFrame sky)
{
  if ((!fits1 || !fits2) || 
      (fits1 == fits2) ||
      !(fits1->hasWCS(sys)) ||
      !(fits2->hasWCS(sys)))
    return Matrix();

  Vector org1 = fits1->getWCScrpix(sys);
  Vector orval1 = fits1->pix2wcs(org1,sys,sky);

  Vector org2 = fits2->getWCScrpix(sys);
  Vector orval2 = fits2->pix2wcs(org2,sys,sky);

  // orientation
  Orientation orientation =NORMAL;
  {
    Orientation o1 = fits1->getWCSOrientation(sys,sky);
    Orientation o2 = fits2->getWCSOrientation(sys,sky);

    switch (o1) {
    case NORMAL:
      {
	switch (o2) {
	case NORMAL:
	  orientation = NORMAL;
	  break;
	case XX:
	  orientation = XX;
	  break;
	}
      }
      break;
    case XX:
      {
	switch (o2) {
	case NORMAL:
	  orientation = XX;
	  break;
	case XX:
	  orientation = NORMAL;
	  break;
	}
      }
      break;
    }
  }

  // zoom
  Vector zoom =Vector(1,1);
  {
    Vector cd1 = fits1->getWCScdelt(sys);
    Vector cd2 = fits2->getWCScdelt(sys);
    zoom = Vector(cd2[0]/cd1[0], cd2[1]/cd1[1]).abs();
  }
  
  // rotation
  double rotation =0;
  {
    double r1 =0;
    {
      Vector orpix = fits1->wcs2pix(orval1, sys, sky);
      Vector delta = fits1->getWCScdelt(sys).abs();
      Vector npix = 
	fits1->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
      Vector north = npix - orpix;

      if ((fits1->getWCS(sys))->imflip)
	r1 = (north.angle()-M_PI/2);
      else
	r1 = -(north.angle()-M_PI/2);
    }

    double r2 =0;
    {
      Vector orpix = fits2->wcs2pix(orval1, sys, sky);
      Vector delta = fits2->getWCScdelt(sys).abs();
      Vector npix = 
	fits2->wcs2pix(Vector(orval1[0],orval1[1]+delta[1]),sys,sky);
      Vector north = npix - orpix;

      if ((fits2->getWCS(sys))->imflip)
	r2 = (north.angle()-M_PI/2);
      else
	r2 = -(north.angle()-M_PI/2);
    }
    rotation = r1-r2;
  }

  // origin
  Vector origin;
  {
    Vector orpix1 = fits1->wcs2pix(orval2,sys,sky) * imageToData;
    Vector orpix2 = fits2->wcs2pix(orval2,sys,sky) * imageToData;
    origin = orpix1 - orpix2;
  }

  // matrix
  {
    // if the segment is flipped, we can have a discontinuity at
    // the edges, due to round off errors, so we 'nudge' it

    double rr = degToRad(180);
    if ((rotation > rr-FLT_EPSILON && rotation < rr+FLT_EPSILON) || 
	(-rotation > rr-FLT_EPSILON && -rotation < rr+FLT_EPSILON)) {
      switch (orientation) {
      case NORMAL:
	origin -= Vector(0,FLT_EPSILON);
	break;
      case XX:
	origin -= Vector(FLT_EPSILON,FLT_EPSILON);
	break;
      }
    }
    else {
      switch (orientation) {
      case NORMAL:
	break;
      case XX:
	origin -= Vector(FLT_EPSILON,0);
	break;
      }
    }

    Matrix flip;
    switch (orientation) {
    case NORMAL:
      break;
    case XX:
      flip = FlipX();
      break;
    }

    Vector orpix2 = fits2->wcs2pix(orval2,sys,sky) * imageToData;
    return
      Translate(-orpix2) *
      flip *
      Scale(zoom) *
      Rotate(rotation) *
      Translate(orpix2) *
      Translate(origin);
  }
}

Orientation FrameBase::IRAFOrientation(Orientation oo)
{
  if (irafOrientation == (Orientation)-1)
    irafOrientation = oo;
  return irafOrientation;
}
