// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <stdlib.h>
#include <time.h>

#include <fstream>
#include "fdstream.hpp"

#include "framebase.h"
#include "fitsimage.h"
#include "util.h"
#include "mmap.h"
#include "contour.h"

#include "marker.h"
#include "basemarker.h"

#include "annulus.h"
#include "box.h"
#include "boxannulus.h"
#include "bpanda.h"
#include "circle.h"
#include "compass.h"
#include "composite.h"
#include "cpanda.h"
#include "ellipse.h"
#include "ellipseannulus.h"
#include "epanda.h"
#include "line.h"
#include "ruler.h"
#include "point.h"
#include "polygon.h"
#include "projection.h"
#include "text.h"
#include "vect.h"

#include "parser.H"

#define LISTBUFSIZE 8192

// NOTE: all marker traversal routines use a local ptr as opposed to the
// list current() because marker call backs may invoke another traversal 
// routine or the event loop, which may process pending events

FrameBase* fr;

// DS9/Funtools Marker Parser Stuff
// Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer mkFlexLexer
#include <FlexLexer.h>

YYSTYPE* mklval;
mkFlexLexer* mklexx;
extern int mkparse(FrameBase*, mkFlexLexer*);

int mklex(void* vval, mkFlexLexer* ll)
{
  mklval = (YYSTYPE*)vval;
  mklexx = ll;
  return ll ? ll->yylex() : 0;
}

void mkerror(FrameBase* fr, mkFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// CIAO Marker Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer ciaoFlexLexer
#include <FlexLexer.h>

YYSTYPE* ciaolval;
ciaoFlexLexer* ciaolexx;
extern int ciaoparse(FrameBase*, ciaoFlexLexer*);

int ciaolex(void* vval, ciaoFlexLexer* ll)
{
  ciaolval = (YYSTYPE*)vval;
  ciaolexx = ll;
  return ll ? ll->yylex() : 0;
}

void ciaoerror(FrameBase* fr, ciaoFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// PROS Marker Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer prosFlexLexer
#include <FlexLexer.h>

YYSTYPE* proslval;
prosFlexLexer* proslexx;
extern int prosparse(FrameBase*, prosFlexLexer*);

int proslex(void* vval, prosFlexLexer* ll)
{
  proslval = (YYSTYPE*)vval;
  proslexx = ll;
  return ll ? ll->yylex() : 0;
}

void proserror(FrameBase* fr, prosFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// TNG Marker Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer tngFlexLexer
#include <FlexLexer.h>

YYSTYPE* tnglval;
tngFlexLexer* tnglexx;
extern int tngparse(FrameBase*, tngFlexLexer*);

int tnglex(void* vval, tngFlexLexer* ll)
{
  tnglval = (YYSTYPE*)vval;
  tnglexx = ll;
  return ll ? ll->yylex() : 0;
}

void tngerror(FrameBase* fr, tngFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// SAO Marker Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer saoFlexLexer
#include <FlexLexer.h>

YYSTYPE* saolval;
saoFlexLexer* saolexx;
extern int saoparse(FrameBase*, saoFlexLexer*);

int saolex(void* vval, saoFlexLexer* ll)
{
  saolval = (YYSTYPE*)vval;
  saolexx = ll;
  return ll ? ll->yylex() : 0;
}

void saoerror(FrameBase* fr, saoFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// XY Marker Parser Stuff
#undef yyFlexLexer
#define yyFlexLexer xyFlexLexer
#include <FlexLexer.h>

YYSTYPE* xylval;
xyFlexLexer* xylexx;
extern int xyparse(FrameBase*, xyFlexLexer*);

int xylex(void* vval, xyFlexLexer* ll)
{
  xylval = (YYSTYPE*)vval;
  xylexx = ll;
  return ll ? ll->yylex() : 0;
}

void xyerror(FrameBase* fr, xyFlexLexer* ll, const char* m)
{
  fr->error(m);
  const char* cmd = ll ? ll->YYText() : (const char*)NULL;
  if (cmd && cmd[0] != '\n') {
    fr->error(": ");
    fr->error(cmd);
  }
}

// Marker Commands

void FrameBase::createCircleCmd(const Vector& center, 
				double radius,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop, 
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Circle(this, center, 
			  radius, 
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createAnnulusCmd(const Vector& center,
				 double start, double stop, int num,
				 const char* color, int width,const char* font,
				 const char* text, unsigned short prop,
				 const char* comment, 
				 const List<Tag>& tag,const List<CallBack>& cb)
{
  createMarker(new Annulus(this, center, 
			   start, stop, num, 
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createAnnulusCmd(const Vector& center,
				 int num, double* radii,
				 const char* color, int width,const char* font,
				 const char* text, unsigned short prop,
				 const char* comment, 
				 const List<Tag>& tag,const List<CallBack>& cb)
{
  createMarker(new Annulus(this, center, 
			   num, radii,
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createCpandaCmd(const Vector& center,
			       double ang1, double ang2, int an, 
			       double rad1, double rad2, int rn, 
			       const char* color, int width, const char* font,
			       const char* text, unsigned short prop,
			       const char* comment, 
			       const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Cpanda(this, center, 
			 ang1, ang2, an, 
			 rad1, rad2, rn, 
			 color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createCpandaCmd(const Vector& center,
			       int an, double* a,
			       int rn, double* r,
			       const char* color, int width, const char* font,
			       const char* text, unsigned short prop,
			       const char* comment, 
			       const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Cpanda(this, center, 
			 an, a,
			 rn, r,
			 color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createEllipseCmd(const Vector& center, 
				 const Vector& radius,
				 double angle,
				 const char* color, int width,const char* font,
				 const char* text, unsigned short prop,
				 const char* comment, 
				 const List<Tag>& tag,const List<CallBack>& cb)

{
  createMarker(new Ellipse(this,center, 
			   radius, 
			   angle, 
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createEllipseAnnulusCmd(const Vector& center,
			const Vector& inner, const Vector& outer, int num, 
			double angle,
			const char* color, int width, const char* font, 
  			const char* text, unsigned short prop,
			const char* comment, 
			const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new EllipseAnnulus(this, center, 
		        inner, outer, num,
		        angle, 
			color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createEllipseAnnulusCmd(const Vector& center, 
			int num, Vector* radii,
			double angle,
			const char* color, int width, const char* font, 
			const char* text, unsigned short prop,
			const char* comment, 
			const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new EllipseAnnulus(this, center, 
			  num, radii, 
			  angle, 
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createEpandaCmd(const Vector& center,
				double ang1, double ang2, int an,
				const Vector& rad1, const Vector& rad2, int rn,
				double angle,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Epanda(this, center, 
			  ang1, ang2, an, 
			  rad1, rad2, rn, 
			  angle,
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createEpandaCmd(const Vector& center,
				int an, double* a,
				int rn, Vector* r,
				double angle,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Epanda(this, center, 
			  an, a,
			  rn, r,
			  angle,
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createBoxCmd(const Vector& center, 
			     const Vector& size, 
			     double angle,
			     const char* color, int width, const char* font,
			     const char* text, unsigned short prop,
			     const char* comment, 
			     const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Box(this, center, 
		       size, 
		       angle, 
		       color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createBoxAnnulusCmd(const Vector& center,
			    const Vector& inner, const Vector& outer, int num,
			    double angle,
			    const char* color, int width, const char* font,
			    const char* text, unsigned short prop,
			    const char* comment, 
			    const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new BoxAnnulus(this, center, 
		      inner, outer, num, 
		      angle, 
		      color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createBoxAnnulusCmd(const Vector& center, 
			    int num, Vector* size,
			    double angle, 
			    const char* color, int width, const char* font, 
			    const char* text, unsigned short prop,
			    const char* comment, 
			    const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new BoxAnnulus(this, center, 
		      num, size, 
		      angle, 
		      color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createBpandaCmd(const Vector& center,
				double ang1, double ang2, int an,
				const Vector& rad1, const Vector& rad2, int rn,
				double angle,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Bpanda(this, center, 
			  ang1, ang2, an, 
			  rad1, rad2, rn, 
			  angle,
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createBpandaCmd(const Vector& center,
				int an, double* a,
				int rn, Vector* r,
				double angle,
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Bpanda(this, center, 
			  an, a,
			  rn, r,
			  angle,
			  color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createLineCmd(const Vector& center, const Vector& p2, 
			      int arrow1, int arrow2,
			      const char* color, int width, const char* font,
			      const char* text, unsigned short prop,
			      const char* comment, 
			      const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Line(this, center, p2, arrow1, arrow2,
			color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createVectCmd(const Vector& center, const Vector& p2, 
			      int arrow,
			      const char* color, int width, const char* font,
			      const char* text, unsigned short prop,
			      const char* comment, 
			      const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Vect(this, center, p2, arrow,
			color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createVectCmd(const Vector& center, double mag, double ang,
			      int arrow,
			      const char* color, int width, const char* font,
			      const char* text, unsigned short prop,
			      const char* comment, 
			      const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Vect(this, center, mag, ang, arrow,
			color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createProjectionCmd(const Vector& center, const Vector& p2, 
				    double w, const char* mvcb, 
				    const char* delcb,
				    const char* color, int width, 
				    const char* font,
				    const char* text, unsigned short prop,
				    const char* comment, 
				    const List<Tag>& tag, 
				    const List<CallBack>& cb)
{
  createMarker(new Projection(this, center, p2, w, mvcb, delcb, color, width,
			      font, text, prop, comment, tag, cb));
}

void FrameBase::createRulerCmd(const Vector& center, const Vector& p2,
			       CoordSystem sys, SkyFrame sky,
			       CoordSystem distsys, SkyFormat distformat,
			       const char* color, int width, const char* font,
			       const char* text, unsigned short prop,
			       const char* comment, 
			       const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Ruler(this, center, p2, 
			 sys, sky, distsys, distformat, 
			 color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createCirclePointCmd(const Vector& center, int size,
				     const char* color, int width,
				     const char* font,
				     const char* text, unsigned short prop,
				     const char* comment, 
				     const List<Tag>& tag, 
				     const List<CallBack>& cb)
{
  createMarker(new CirclePoint(this, center, size, color, width, font, 
			       text, prop, comment, tag, cb));
}

void FrameBase::createBoxPointCmd(const Vector& center, int size,
				  const char* color, int width, 
				  const char* font,
				  const char* text, unsigned short prop,
				  const char* comment, 
				  const List<Tag>& tag, 
				  const List<CallBack>& cb)
{
  createMarker(new BoxPoint(this, center, size, color, width, font, text, prop, 
			    comment, tag, cb));
}

void FrameBase::createDiamondPointCmd(const Vector& center, int size,
				      const char* color, int width,
				      const char* font,
				      const char* text, unsigned short prop,
				      const char* comment, 
				      const List<Tag>& tag,
				      const List<CallBack>& cb)
{
  createMarker(new DiamondPoint(this, center, size, color, width, font, 
				text, prop, comment, tag, cb));
}

void FrameBase::createCrossPointCmd(const Vector& center, int size,
				    const char* color, int width, 
				    const char* font,
				    const char* text, unsigned short prop,
				    const char* comment, 
				    const List<Tag>& tag, 
				    const List<CallBack>& cb)
{
  createMarker(new CrossPoint(this, center, size, color, width, font, 
			      text, prop, comment, tag, cb));
}

void FrameBase::createExPointCmd(const Vector& center, int size,
				 const char* color, int width,
				 const char* font,
				 const char* text, unsigned short prop,
				 const char* comment, 
				 const List<Tag>& tag, 
				 const List<CallBack>& cb)
{
  createMarker(new ExPoint(this, center, size, color, width, font, 
			   text, prop, comment, tag, cb));
}

void FrameBase::createArrowPointCmd(const Vector& center, int size,
				    const char* color, int width, 
				    const char* font,
				    const char* text, unsigned short prop,
				    const char* comment, 
				    const List<Tag>& tag, 
				    const List<CallBack>& cb)
{
  createMarker(new ArrowPoint(this, center, size, color, width, font, 
			      text, prop, comment, tag, cb));
}

void FrameBase::createBoxCirclePointCmd(const Vector& center, int size,
					const char* color, int width,
					const char* font,
					const char* text, unsigned short prop,
					const char* comment,
					const List<Tag>& tag, 
					const List<CallBack>& cb)
{
  createMarker(new BoxCirclePoint(this, center, size, color, width, font, 
				  text, prop, comment, tag, cb));
}

void FrameBase::createTextCmd(const Vector& center, double angle, int rotate,
			      const char* color, int width, const char* font,
			      const char* text, unsigned short prop,
			      const char* comment, 
			      const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Text(this, center, angle, rotate,
			color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createPolygonCmd(const Vector& center, 
			 const Vector& bb,
			 const char* color, int width, const char* font,
			 const char* text, unsigned short prop,
			 const char* comment, 
			 const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Polygon(this, center, 
			   bb, 
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createPolygonCmd(const List<Vertex>& list,
			 const char* color, int width, const char* font,
			 const char* text, unsigned short prop,
			 const char* comment, 
			 const List<Tag>& tag, const List<CallBack>& cb)
{
  createMarker(new Polygon(this, list, 
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createContourPolygonCmd(
				const char* color, int width, const char* font,
				const char* text, unsigned short prop,
				const char* comment, 
				const List<Tag>& tag, const List<CallBack>& cb)
{
  // only create in the USER layer
  markerLayerCmd(USER);

  // main contour
  if (hasContour()) {
    List<Vertex>& c = (List<Vertex>&)((*currentcontour)->contours());

    // build from back to front
    if (c.tail()) {
      List<Vertex> list;
      while (c.previous()) {
	if (c.current()->vector[0] != DBL_MAX)
	  list.append(new Vertex(c.current()->vector));
	else {
	  if (list.count()>0) 
	    createMarker(new Polygon(this, list, color, width, font, text, 
				     prop, NULL, tag, cb));
	  list.deleteAll();
	}
      }
      if (list.count()>0) 
	createMarker(new Polygon(this, list, color, width, font, text, prop, 
				 comment, tag, cb));
    }
  }

  // aux contours
  Contour* cptr = auxcontours.head();
  while (cptr) {
    List<Vertex>& c = (List<Vertex>&)(cptr->contours());

    // build from back to front
    if (c.tail()) {
      List<Vertex> list;
      while (c.previous()) {
	if (c.current()->vector[0] != DBL_MAX)
	  list.append(new Vertex(c.current()->vector));
	else {
	  if (list.count()>0) 
	    createMarker(new Polygon(this, list, color, width, font, 
				     text, prop, NULL, tag, cb));
	  list.deleteAll();
	}
      }
      if (list.count()>0) 
	createMarker(new Polygon(this, list, color, width, font, 
				 text, prop, comment, tag, cb));
    }

    cptr = cptr->next();
  }
}

void FrameBase::createCompassCmd(const Vector& center, double r,
				 const char* north, const char* east, 
				 int na, int ea,
				 CoordSystem sys, SkyFrame sky,
				 const char* color, int width, 
				 const char* font,
				 const char* text, unsigned short prop,
				 const char* comment, 
				 const List<Tag>& tag, 
				 const List<CallBack>& cb)
{
  createMarker(new Compass(this, center, r, north, east, na, ea, sys, sky, 
			   color, width, font, text, prop, comment, tag, cb));
}

void FrameBase::createCompositeCmd(const Vector& center, double angle, 
				   int global,
				   const char* color, int width, 
				   const char* font,
				   const char* text, unsigned short prop, 
				   const char* comment, 
				   const List<Tag>& tag, 
				   const List<CallBack>& cb)
{
  Composite* m = new Composite(this, center, angle, global, color, width, font, 
			       text, prop, comment, tag, cb);
  createMarker(m);
  compositeMarker = m;
}

void FrameBase::createCompositeCmd(
				   const char* color, int width, 
				   const char* font,
				   const char* text, unsigned short prop, 
				   const char* comment, 
				   const List<Tag>& tag, 
				   const List<CallBack>& cb)
{
  // find center
  Vector cc;
  int cnt=0;
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && strcmp(m->getType(),"composite")) {
      cc += m->getCenter();
      cnt++;
    }
    m=m->next();
  }
  cc /= cnt;

  // create composite
  Composite* mk = new Composite(this, cc, 0, 1, color, width, font, 
				text, prop, comment, tag, cb);
  createMarker(mk);

  // append members
  m=markers->head();
  while (m) {
    if (m->isSelected() && strcmp(m->getType(),"composite")) {
      m->unselect();
      Marker* next = markers->extractNext(m);
      m->doCallBack(CallBack::DELETECB);
      m->deleteCBs();
      mk->append(m);
      m = next;
    }
    else
      m=m->next();
  }

  mk->select();
  update(PIXMAP);
}

void FrameBase::createTemplateCmd(const Vector& center, const char* fn)
{
  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  

  createTemplate(center,str);
}

void FrameBase::createTemplateVarCmd(const Vector& center, const char* var)
{
  Tcl_Obj* obj = Tcl_GetVar2Ex(interp, (char*)var, NULL, 
			       TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (!obj)
    return;

  // just in case
  Tcl_ConvertToType(interp, obj, Tcl_GetObjType("bytearray"));

  typedef struct ByteArray {
    int used;			/* The number of bytes used in the byte
				 * array. */
    int allocated;		/* The amount of space actually allocated
				 * minus 1 byte. */
    unsigned char bytes[4];	/* The array of bytes.  The actual size of
				 * this field depends on the 'allocated' field
				 * above. */
  } ByteArray;

  Tcl_IncrRefCount(obj);

  // only make command string as long as needed
  // or the rest will be processed as garbage
  ByteArray* ba = (ByteArray*)(obj->internalRep.otherValuePtr);
  int len = ba->used+2;
  char* buf = new char[len];
  memcpy(buf, (char*)ba->bytes, ba->used);

  Tcl_DecrRefCount(obj);

  // add terminator to make parser happy
  buf[len-2] = '\n';
  buf[len-1] = NULL;

  string x(buf);
  istringstream istr(x);

  createTemplate(center,istr);

  delete [] buf;
}

void FrameBase::createTemplate(const Vector& center, istream& str)
{
  FitsImage* ptr = currentFits;
  while (ptr) {
    ptr->initWCS0(center);
    ptr = ptr->nextMosaic();
  }

  mkFlexLexer* ll = new mkFlexLexer(&str);
  mkparse(this, ll);
  delete ll;

  Marker* mk = compositeMarker;
  resetCompositeMarker();

  ptr = *channelFits;
  while (ptr) {
    ptr->resetWCS0();
    ptr = ptr->nextMosaic();
  }

  if (mk) {
    mk->moveTo(center);
    update(PIXMAP, mk->getAllBBox());
  }
}

void FrameBase::createMarker(Marker* m)
{
  if (compositeMarker)
    compositeMarker->append(m);
  else {
    markers->append(m);

    // now update new marker
    update(PIXMAP, m->getAllBBox());

    // exectute any Edit CallBacks
    m->doCallBack(CallBack::EDITCB);

    // and return id
    printInteger(m->getId());
  }
}

void FrameBase::getMarkerAngleCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printDouble(radToDeg(m->getAngle()), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerAngleCmd(int id, CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printDouble(radToDeg(mapAngleFromRef(m->getAngle(), sys, sky)), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerAnnulusRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Annulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Annulus*)m)->annuli(i);
	markerPrintDouble(r[0], sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerAnnulusRadiusCmd(int id, CoordSystem sys,
					  SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Annulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Annulus*)m)->annuli(i);
	markerPrintDouble(m->getCenter(), r[0], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBoxAnnulusRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((BoxAnnulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	markerPrintVector(((BoxAnnulus*)m)->annuli(i), sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBoxAnnulusRadiusCmd(int id, CoordSystem sys, 
				       SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((BoxAnnulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	markerPrintVector(m->getCenter(), ((BoxAnnulus*)m)->annuli(i), 
			  sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBoxRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVector(((Box*)m)->annuli(0), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBoxRadiusCmd(int id, CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVector(m->getCenter(), ((Box*)m)->annuli(0), sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBpandaAnglesCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Bpanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = radToDeg(((Bpanda*)m)->angles(ii));
	if (!ii)
	  first = ang;
	else 
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;
	
	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerBpandaAnglesCmd(int id, CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Bpanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = 
	  radToDeg(mapAngleFromRef(((Bpanda*)m)->angles(ii),sys,sky));
	if (!ii)
	  first = ang;
	else 
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;
	
	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerBpandaRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Bpanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Bpanda*)m)->annuli(i);
	markerPrintVector(((EllipseAnnulus*)m)->annuli(i), sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerBpandaRadiusCmd(int id, CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Bpanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Bpanda*)m)->annuli(i);
	markerPrintVector(m->getCenter(), ((EllipseAnnulus*)m)->annuli(i),
			  sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCenterCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), sys);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerCenterCmd(int id, CoordSystem sys, SkyFrame sky, 
				   SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), m->getCenter(), sys, sky, format);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerCircleRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Vector r = ((Circle*)m)->annuli(0);
      markerPrintDouble(r[0],sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCircleRadiusCmd(int id, CoordSystem sys, 
					 SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Vector r = ((Circle*)m)->annuli(0);
      markerPrintDouble(m->getCenter(),r[0],sys,format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCpandaAnglesCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Cpanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = radToDeg(((Cpanda*)m)->angles(ii));
	if (!ii) 
	  first = ang;
	else
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;
	
	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCpandaAnglesCmd(int id, CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Cpanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = 
	  radToDeg(mapAngleFromRef(((Cpanda*)m)->angles(ii),sys,sky));
	if (!ii) 
	  first = ang;
	else
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;
	
	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCpandaRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Cpanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Cpanda*)m)->annuli(i);
	markerPrintDouble(r[0], sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCpandaRadiusCmd(int id, CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Cpanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Cpanda*)m)->annuli(i);
	markerPrintDouble(m->getCenter(), r[0], sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerColorCmd()
{
  // return first found

  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerColorCmd(const char* tag)
{
  // return first found

  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerColorCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp,  m->getColorName(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompassArrowCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (((Compass*)m)->getNorthArrow())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);

      if (((Compass*)m)->getEastArrow())
	Tcl_AppendResult(interp, " 1", NULL);
      else
	Tcl_AppendResult(interp, " 0", NULL);

      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompassLabelCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendElement(interp, ((Compass*)m)->getNorthText());
      Tcl_AppendElement(interp, ((Compass*)m)->getEastText());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompassRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      double r = ((Compass*)m)->getRadius();
      markerPrintDouble(((Compass*)m)->getRadius(), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompassRadiusCmd(int id, CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(),((Compass*)m)->getRadius(),sys,format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompassSystemCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printCoordSystem(((Compass*)m)->getSystem());
      Tcl_AppendResult(interp, " ", NULL);
      printSkyFrame(((Compass*)m)->getSkyFrame());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerCompositeCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (((Composite*)m)->getGlobal())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerEllipseRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVector(((Ellipse*)m)->annuli(0),sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerEllipseRadiusCmd(int id, CoordSystem sys, 
					  SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintVector(m->getCenter(),((Ellipse*)m)->annuli(0),sys,format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerEllipseAnnulusRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((EllipseAnnulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	markerPrintVector(((EllipseAnnulus*)m)->annuli(i), sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerEllipseAnnulusRadiusCmd(int id, CoordSystem sys, 
						 SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((EllipseAnnulus*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	markerPrintVector(m->getCenter(), ((EllipseAnnulus*)m)->annuli(i),
			  sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerEpandaAnglesCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Epanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = radToDeg(((Epanda*)m)->angles(ii));
	if (!ii)
	  first = ang;
	else
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;

	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerEpandaAnglesCmd(int id, CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Epanda*)m)->numAngles();
      double first=0;
      for (int ii=0; ii<cnt; ii++) {
	double ang = 
	  radToDeg(mapAngleFromRef(((Epanda*)m)->angles(ii),sys,sky));
	if (!ii)
	  first = ang;
	else
	  if (ang<=first+FLT_EPSILON)
	    ang += 360;

	printDouble(ang, DEFAULT);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerEpandaRadiusCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Epanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Epanda*)m)->annuli(i);
	markerPrintVector(((EllipseAnnulus*)m)->annuli(i), sys);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerEpandaRadiusCmd(int id, CoordSystem sys, 
					 SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      int cnt = ((Epanda*)m)->numAnnuli();
      for (int i=0; i<cnt; i++) {
	Vector r = ((Epanda*)m)->annuli(i);
	markerPrintVector(m->getCenter(), ((EllipseAnnulus*)m)->annuli(i),
			  sys, format);
	Tcl_AppendResult(interp, "\n", NULL);
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerFontCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerFontCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerFontCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getFont(), NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerHandleCmd(const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      int h = m->onHandle(v);
      if (h) {
	ostringstream str;
	str << m->getId() << ' ' << h << ends;
	Tcl_AppendResult(interp, str.str().c_str(), NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}

void FrameBase::getMarkerIdCmd(const char* tag)
{
  // return first found

  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      printInteger(m->getId());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerIdCmd(const Vector& v)
{
  // v is in canvas coords

  Marker* m=markers->head();
  while (m) {
    if (m->isIn(v)) {
      printInteger(m->getId());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerIdAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    ostringstream str;
    str << m->getId() << ' ' << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
    m=m->next();
  }
}

void FrameBase::getMarkerLineCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(((Line*)m)->getP1(), sys);
      markerPrintCoord(((Line*)m)->getP2(), sys);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerLineCmd(int id, CoordSystem sys, SkyFrame sky, 
				 SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), ((Line*)m)->getP1(), sys, sky, format);
      markerPrintCoord(m->getCenter(), ((Line*)m)->getP2(), sys, sky, format);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerLineArrowCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (((Line*)m)->getP1Arrow())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);

      if (((Line*)m)->getP2Arrow())
	Tcl_AppendResult(interp, " 1", NULL);
      else
	Tcl_AppendResult(interp, " 0", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerLineLengthCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(((Line*)m)->getP2(),((Line*)m)->getP1(), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerLineLengthCmd(int id, CoordSystem sys, 
				       SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(),((Line*)m)->getP2(),((Line*)m)->getP1(),
			sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerLineWidthCmd()
{
  // return first found
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      ostringstream str;
      str << m->getLineWidth() << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerLineWidthCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      ostringstream str;
      str << m->getLineWidth() << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerMapLenFromRefCmd(int id, double d, CoordSystem sys, 
					  SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printDouble(m->mapLenFromRef(d,sys,format), DEFAULT);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerPointSizeCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printInteger(((Point*)m)->size());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerPolygonSegmentCmd(const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->isPolygon()) {
      int s = ((Polygon*)(m))->getSegment(v);
      if (s) {
	ostringstream str;
	str << m->getId() << ' ' << s << ends;
	Tcl_AppendResult(interp, str.str().c_str(), NULL);
	return;
      }
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0 0", NULL);
}

void FrameBase::getMarkerPreserveCmd()
{
  if (preserveMarkers)
      Tcl_AppendResult(interp, "1", NULL);
  else
      Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerProjectionCmd(int id, 
				       char* xName, char* yName,
				       char* xcName, char* ycName)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Projection* p = (Projection*)m;
      bltProjection(xName, yName, xcName, ycName, 
		    p->getSystem(), p->getSkyFrame(),
		    p->getP1(), p->getP2(), p->getWidth(), p->getMethod());
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerProjectionPointsCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(((Projection*)m)->getP1(), sys);
      markerPrintCoord(((Projection*)m)->getP2(), sys);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerProjectionPointsCmd(int id, CoordSystem sys, 
					     SkyFrame sky, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), ((Projection*)m)->getP1(), sys, 
		       sky, format);
      markerPrintCoord(m->getCenter(), ((Projection*)m)->getP2(), sys, 
		       sky, format);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerProjectionLengthCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(((Projection*)m)->getP2(),
			((Projection*)m)->getP1(), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerProjectionLengthCmd(int id, CoordSystem sys,
					   SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(), ((Projection*)m)->getP2(),
			((Projection*)m)->getP1(), sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerProjectionMethodCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printInteger(((Projection*)m)->getMethod());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerProjectionWidthCmd(int id, InternalSystem sys) 
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(((Projection*)m)->getWidth(), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerProjectionWidthCmd(int id, CoordSystem sys, 
					    SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(), ((Projection*)m)->getWidth(),
			sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerProjectionSystemCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printCoordSystem(((Projection*)m)->getSystem());
      Tcl_AppendResult(interp, " ", NULL);
      printSkyFrame(((Projection*)m)->getSkyFrame());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerPropertyCmd(unsigned short which)
{
  // return first selected found

  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL): 
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerPropertyCmd(const char* tag, unsigned short which)
{
  // return first selected found

  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL): 
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerPropertyCmd(int id, unsigned short which)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->getProperty(which) ?
	Tcl_AppendResult(interp, "1", NULL):
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  // else, return 'off'

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerRulerLengthCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(((Ruler*)m)->getP2(),((Ruler*)m)->getP1(),sys);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDouble(((Ruler*)m)->getP3(),((Ruler*)m)->getP1(),sys);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDouble(((Ruler*)m)->getP3(),((Ruler*)m)->getP2(),sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerRulerLengthCmd(int id, CoordSystem sys, 
				      SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(), 
			((Ruler*)m)->getP2(), ((Ruler*)m)->getP1(),
			sys, format);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDouble(m->getCenter(), 
			((Ruler*)m)->getP3(), ((Ruler*)m)->getP1(),
			sys, format);
      Tcl_AppendResult(interp, " ", NULL);
      markerPrintDouble(m->getCenter(), 
			((Ruler*)m)->getP3(), ((Ruler*)m)->getP2(),
			sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerRulerPointCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(((Ruler*)m)->getP1(), sys);
      markerPrintCoord(((Ruler*)m)->getP2(), sys);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerRulerPointCmd(int id, CoordSystem sys, SkyFrame sky,
				       SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), ((Ruler*)m)->getP1(), sys, sky, format);
      markerPrintCoord(m->getCenter(), ((Ruler*)m)->getP2(), sys, sky, format);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerRulerSystemCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      printCoordSystem(((Ruler*)m)->getSystem());
      Tcl_AppendResult(interp, " ", NULL);
      printSkyFrame(((Ruler*)m)->getSkyFrame());
      Tcl_AppendResult(interp, " ", NULL);
      printCoordSystem(((Ruler*)m)->getDistSystem());
      Tcl_AppendResult(interp, " ", NULL);
      printSkyFormat(((Ruler*)m)->getDistFormat());
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerSelectedCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      ostringstream str;
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }
    m=m->next();
  }
}

void FrameBase::getMarkerSelectedCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->isSelected())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerSelectedCmd(const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isIn(v) && m->isSelected()) {
      ostringstream str;
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerSelectedNumberCmd()
{
  int count=0;
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected())
      count++;
    m=m->next();
  }

  printInteger(count);
}

void FrameBase::getMarkerHighlitedCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isHighlited()) {
      ostringstream str;
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }
    m=m->next();
  }
}

void FrameBase::getMarkerHighlitedCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->isHighlited())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerHighlitedCmd(const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isIn(v) && m->isHighlited()) {
      ostringstream str;
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerHighlitedNumberCmd()
{
  int count=0;
  Marker* m=markers->head();
  while (m) {
    if (m->isHighlited())
      count++;
    m=m->next();
  }

  printInteger(count);
}

void FrameBase::getMarkerShowCmd()
{
  if (showMarkers)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerShowTextCmd()
{
  if (showMarkersText)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::getMarkerTagCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      ostringstream str;
      str << m->getId() << ' ' << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }
    m=m->next();
  }
}

void FrameBase::getMarkerTagCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      const char* r = m->getTag();
      while (r) {
	Tcl_AppendElement(interp, r);
	r = m->getNextTag();
      }
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerTagCmd(int id, int num)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp,  m->getTag(num), NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerTagDefaultNameCmd()
{
  int cnt = 1;

 again:
  ostringstream str;
  str << "Group " << cnt << ends;
  Marker* m = markers->head();
  while (m) {
    if (m->hasTag(str.str().c_str())) {
      cnt++;
      goto again;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, str.str().c_str(), NULL);
}

void FrameBase::getMarkerTagNumberCmd(const char* tag)
{
  // return number of markers with tag
  int count=0;
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag))
      count++;
    m=m->next();
  }

  printInteger(count);
}

void FrameBase::getMarkerTagsCmd()
{
  // return all tags
  List<Tag> tags;

  // loop thru all markers
  Marker* m=markers->head();
  while (m) {

    // loop thru all tags in markers
    const char* t = m->getTag();
    while (t) {
      // loop thru our current list
      int found = 0;
      Tag* tt = tags.head();
      while (tt) {
	if (!strcmp(tt->tag(), t)) {
	  found =1;
	  break;
	}
	tt = tt->next();
      }
      // didn't find it, add it to the list
      if (!found) {
	tags.append(new Tag(t));
      }

      t=m->getNextTag();
    }
    m=m->next();
  }

  // now sort

  // ok, dump the tags
  Tag* tt = tags.head();
  while (tt) {
    Tcl_AppendElement(interp, tt->tag());
    tt=tt->next();
  }
}

void FrameBase::getMarkerTextCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getText(), NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerTextRotateCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (((Text*)m)->getRotate())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerTypeCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      Tcl_AppendResult(interp, m->getType(), NULL);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerVectorCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(((Line*)m)->getP1(), sys);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerVectorCmd(int id, CoordSystem sys, SkyFrame sky, 
				   SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintCoord(m->getCenter(), ((Line*)m)->getP1(), sys, sky, format);
      return;
    }
    m=m->next();
  }
}

void FrameBase::getMarkerVectorArrowCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (((Vect*)m)->getArrow())
	Tcl_AppendResult(interp, "1", NULL);
      else
	Tcl_AppendResult(interp, "0", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerVectorLengthCmd(int id, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(((Vect*)m)->getP2(), ((Vect*)m)->getP1(), sys);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::getMarkerVectorLengthCmd(int id, CoordSystem sys, 
					 SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      markerPrintDouble(m->getCenter(), ((Vect*)m)->getP2(), 
			((Vect*)m)->getP1(), sys, format);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::hasMarkerSelectedCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::hasMarkerHighlitedCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isHighlited()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::hasMarkerPasteCmd()
{
  if (!pasteMarkers->isEmpty())
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::hasMarkerUndoCmd()
{
  if (undoMarkers && !undoMarkers->isEmpty()) 
    switch (undoMarkerType) {
    case MOVE:
      Tcl_AppendResult(interp, "move", NULL);
      return;
    case EDIT:
      Tcl_AppendResult(interp, "edit", NULL);
      return;
    case DELETE:
      Tcl_AppendResult(interp, "delete", NULL);
      return;
    default:
      Tcl_AppendResult(interp, "", NULL);
      return;
    }
  else
    Tcl_AppendResult(interp, "", NULL);
}

void FrameBase::markerLayerCmd(MarkerLayer layer) {
  switch (layer) {
  case USER:
    markers = &userMarkers;
    undoMarkers = &undoUserMarkers;
    pasteMarkers = &pasteUserMarkers;
    break;
  case CATALOG:
    markers = &catalogMarkers;
    undoMarkers = NULL;
    pasteMarkers = NULL;
    break;
  }
}

void FrameBase::markerAngleCmd(int id, double angle)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canRotate()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	m->setAngle(angle);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerAngleCmd(int id, double angle, CoordSystem sys, 
			       SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canRotate()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	m->setAngle(mapAngleToRef(angle,sys,sky));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Annulus*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Annulus*)m)->deleteAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerAnnulusRadiusCmd(int id, double inner, double outer,
				       int num, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	double r1 = mapLenToRef(inner,sys);
	double r2 = mapLenToRef(outer,sys);
	((Annulus*)m)->setAnnuli(Vector(r1,r1),Vector(r2,r2),num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerAnnulusRadiusCmd(int id, double inner, double outer,
				       int num,
				       CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	double r1 = ptr->mapLenToRef(inner, sys, format);
	double r2 = ptr->mapLenToRef(outer, sys, format);
	((Annulus*)m)->setAnnuli(Vector(r1,r1),Vector(r2,r2),num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerAnnulusRadiusCmd(int id, const char* lev, 
				       CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int cnt = 0;
	Vector radii[MAXANNULI];
	string x(lev);
	istringstream str(x);
	while ((cnt<MAXANNULI) && (str >> radii[cnt][0])) {
	  radii[cnt][1] = radii[cnt][0];
	  ++cnt;
	}	

	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<cnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], sys, format);

	((Annulus*)m)->setAnnuli(radii,cnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBackCmd()
{
  Marker* m = markers->tail();
  while (m) {
    if (m->isSelected()) {
      Marker* prev = markers->extractPrev(m);
      markers->append(m);
      update(PIXMAP, m->getAllBBox());
      m = prev;
    }
    else
      m=m->previous();
  }
}

void FrameBase::markerBackCmd(const char* tag)
{
  Marker* m = markers->tail();
  while (m) {
    if (m->hasTag(tag)) {
      Marker* prev = markers->extractPrev(m);
      markers->append(m);
      update(PIXMAP, m->getAllBBox());
      m = prev;
    }
    else
      m=m->previous();
  }
}

void FrameBase::markerBackCmd(int id)
{
  Marker* m = markers->tail();
  while (m) {
    if (m->getId() == id) {
      markers->extractPrev(m);
      markers->append(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->previous();
  }
}

void FrameBase::markerBoxAnnulusRadiusCmd(int id, const Vector& inner,
					const Vector& outer, int num,
					InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	Vector s1 = mapLenToRef(inner, sys);
	Vector s2 = mapLenToRef(outer, sys);
	((BoxAnnulus*)(m))->setAnnuli(s1,s2,num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBoxAnnulusRadiusCmd(int id, const Vector& inner,
					const Vector& outer, int num,
					CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	Vector s1 = ptr->mapLenToRef(inner, sys, format);
	Vector s2 = ptr->mapLenToRef(outer, sys, format);
	((BoxAnnulus*)(m))->setAnnuli(s1,s2,num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBoxAnnulusRadiusCmd(int id,const char* lev, 
					CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int cnt = 0;
	Vector size[MAXANNULI];
	string x(lev);
	istringstream str(x);
	while ((cnt<MAXANNULI) && (str >> size[cnt][0]))
	  str >> size[cnt++][1];
	
	// verify proper ratios
	for (int i=0; i<cnt; i++)
	  size[i][1] = size[i][0]*size[cnt-1][1]/size[cnt-1][0];

	// map to ref coord sys
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<cnt; i++)
	  size[i] = ptr->mapLenToRef(size[i], sys, format);

	((BoxAnnulus*)m)->setAnnuli(size,cnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBoxAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((BoxAnnulus*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBoxAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((BoxAnnulus*)m)->deleteAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerBoxRadiusCmd(int id, const Vector& size, 
				   InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Box*)(m))->setAnnuli(mapLenToRef(size, sys));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBoxRadiusCmd(int id, const Vector& size, 
				 CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Box*)(m))->setAnnuli(ptr->mapLenToRef(size, sys, format));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBpandaCreateAnglesCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Bpanda*)m)->addAngles(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBpandaCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Bpanda*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBpandaDeleteCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Bpanda*)m)->deleteAnglesAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerBpandaEditCmd(int id, 
				    double a1, double a2, int an, 
				    const Vector& r1, const Vector& r2, int rn)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Bpanda*)m)->setAnglesAnnuli(a1,a2,an,r1,r2,rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBpandaEditCmd(int id, 
				    double a1, double a2, int an, 
				    const Vector& r1, const Vector& r2, int rn,
				    CoordSystem sys, SkyFrame sky)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Bpanda*)m)->setAnglesAnnuli(mapAngleToRef(a1,sys,sky),
				      mapAngleToRef(a2,sys,sky),an,r1,r2,rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerBpandaEditCmd(int id, 
				    const char* a, const char* r,
				    CoordSystem sys, SkyFrame sky, 
				    CoordSystem rsys, SkyFormat rformat)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int acnt = 0;
	double angles[MAXANGLES];
	{
	  string x(a);
	  istringstream astr(x);
	  while ((acnt<MAXANGLES) && (astr >> angles[acnt]))
	    ++acnt;
	}	
	for (int i=0; i<acnt; i++)
	  angles[i] = mapAngleToRef(degToRad(angles[i]),sys,sky);

	int rcnt = 0;
	Vector radii[MAXANNULI];
	{
	  string x(r);
	  istringstream rstr(x);
	  while ((rcnt<MAXANNULI) && (rstr >> radii[rcnt][0]))
	    rstr >> radii[rcnt++][1];
	}	
	// verify proper ratios
	for (int i=0; i<rcnt; i++)
	  radii[i][1] = radii[i][0]*radii[rcnt-1][1]/radii[rcnt-1][0];

	// map to ref coord sys
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<rcnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], rsys, rformat);

	((Bpanda*)m)->setAnglesAnnuli(angles,acnt,radii,rcnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCallBackCmd(int id, CallBack::Type cb, const char* p, 
			      const char* a)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      result = m->addCallBack(cb, p, a);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCircleRadiusCmd(int id, double radius,InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	double r = mapLenToRef(radius, sys);
	((Circle*)m)->setAnnuli(Vector(r,r));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCircleRadiusCmd(int id, double radius,
				      CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	double r = ptr->mapLenToRef(radius, sys, format);
	((Circle*)m)->setAnnuli(Vector(r,r));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerColorCmd(const char* clr)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerColorCmd(const char* tag, const char* clr)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerColorCmd(int id, const char* clr)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->setColor(clr);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerCommandCmd(MarkerFormat fm, const char* ccmd)
{
  // only make command string as long as needed
  // or the rest will be processed as garbage
  int len = strlen(ccmd)+2;
  char* buf = new char[len];
  memcpy(buf, ccmd, len);

  // add terminator to make parser happy
  buf[len-2] = '\n';
  buf[len-1] = NULL;

  string x(buf);
  istringstream istr(x);
  parseMarker(fm, istr);
  delete [] buf;
}

void FrameBase::markerCommandVarCmd(MarkerFormat fm, const char* var)
{
  const char* ccmd = Tcl_GetVar(interp, var,
				TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (!ccmd) {
    result = TCL_ERROR;
    return;
  }

  // only make command string as long as needed
  // or the rest will be processed as garbage
  int len = strlen(ccmd)+2;
  char* buf = new char[len];
  memcpy(buf, ccmd, len);

  // add terminator to make parser happy
  buf[len-2] = '\n';
  buf[len-1] = NULL;

  string x(buf);
  istringstream istr(x);
  parseMarker(fm, istr);
  delete [] buf;
}

void FrameBase::markerCompassArrowCmd(int id, int n, int e)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	((Compass*)(m))->setArrows(n, e);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompassLabelCmd(int id, const char* n, const char* e)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Compass*)(m))->setLabels(n, e);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompassRadiusCmd(int id, double r, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	double rr = mapLenToRef(r, sys);
	((Compass*)m)->setRadius(rr);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompassRadiusCmd(int id, double r, CoordSystem sys, 
				       SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	double rr = ptr->mapLenToRef(r, sys, format);
	((Compass*)m)->setRadius(rr);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompassSystemCmd(int id, CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Compass*)(m))->setCoordSystem(sys, sky);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompositeCmd(int id, int gl)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	((Composite*)(m))->setGlobal(gl);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCompositeDeleteCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && !strcmp(m->getType(),"composite")) {

      Marker* mm = ((Composite*)m)->extract();
      while (mm) {
	markers->append(mm);
	mm = ((Composite*)m)->extract();
      }
      Marker* next = markers->extractNext(m);
      delete m;
      m = next;

      update(PIXMAP);
    }
    else
      m=m->next();
  }
}

void FrameBase::markerCpandaCreateAnglesCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Cpanda*)m)->addAngles(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCpandaCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Cpanda*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCpandaDeleteCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Cpanda*)m)->deleteAnglesAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerCpandaEditCmd(int id, double a1, double a2, int an, 
				    double r1, double r2, int rn)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	((Cpanda*)m)->setAnglesAnnuli(a1,a2,an,Vector(r1,r1),Vector(r2,r2),rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCpandaEditCmd(int id, double a1, double a2, int an, 
			       double r1, double r2, int rn, 
			       CoordSystem sys, SkyFrame sky)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	((Cpanda*)m)->setAnglesAnnuli(mapAngleToRef(a1,sys,sky),
				     mapAngleToRef(a2,sys,sky),
				     an,
				     Vector(r1,r1),Vector(r2,r2),rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCpandaEditCmd(int id, const char* a, const char* r,
				   CoordSystem sys, SkyFrame sky, 
				   CoordSystem rsys, SkyFormat rformat)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int acnt = 0;
	double angles[MAXANGLES];
	{
	  string x(a);
	  istringstream astr(x);
	  while ((acnt<MAXANGLES) && (astr >> angles[acnt]))
	    ++acnt;
	}	
	{
	  for (int i=0; i<acnt; i++)
	    angles[i] = mapAngleToRef(degToRad(angles[i]),sys,sky);
	}
	int rcnt = 0;
	Vector radii[MAXANNULI];
	{
	  string x(r);
	  istringstream rstr(x);
	  while ((rcnt<MAXANNULI) && (rstr >> radii[rcnt][0])) {
	    radii[rcnt][1] = radii[rcnt][0];
	    ++rcnt;
	  }
	}	
	{
	  FitsImage* ptr = findFits(m->getCenter());
	  for (int i=0; i<rcnt; i++)
	    radii[i] = ptr->mapLenToRef(radii[i], rsys, rformat);
	}

	((Cpanda*)m)->setAnglesAnnuli(angles,acnt,radii,rcnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerDeleteCallBackCmd(int id, CallBack::Type cb, 
					const char* p)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      result = m->deleteCallBack(cb, p);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerDeleteCmd()
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers->extractNext(m);
      update(PIXMAP);

      m->doCallBack(CallBack::DELETECB);
      m->deleteCBs();
      if (undoMarkers) {
	undoMarkers->append(m);
	undoMarkerType = DELETE;
      }
      else
	delete m;
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerDeleteCmd(const char* tag)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->canDelete() && m->hasTag(tag)) {
      Marker* next = markers->extractNext(m);
      update(PIXMAP);

      m->doCallBack(CallBack::DELETECB);
      m->deleteCBs();
      if (undoMarkers) {
	undoMarkers->append(m);
	undoMarkerType = DELETE;
      } 
      else
	delete m;
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerDeleteCmd(int id)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canDelete()) {
	markers->extractNext(m);
	update(PIXMAP);

	m->doCallBack(CallBack::DELETECB);
	m->deleteCBs();
	if (undoMarkers) {
	  undoMarkers->append(m);
	  undoMarkerType = DELETE;
	}
	else
	  delete m;
      }
      return;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerDeleteAllCmd()
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->canDelete()) {
      update(PIXMAP);
      Marker* next = markers->extractNext(m);

      m->doCallBack(CallBack::DELETECB);
      m->deleteCBs();
      if (undoMarkers) {
	undoMarkers->append(m);
	undoMarkerType = DELETE;
      }
      else
	delete m;
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerDeleteLastCmd()
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->tail();
  if (m && m->canDelete()) {
    markers->extractNext(m);
    update(PIXMAP);

    m->doCallBack(CallBack::DELETECB);
    m->deleteCBs();
    if (undoMarkers) {
      undoMarkers->append(m);
      undoMarkerType = DELETE;
    }
    else
      delete m;
    return;
  }
}

void FrameBase::markerDeleteTagCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canDelete())
	m->deleteTags();
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerDeleteTagCmd(int id, const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canDelete()) 
	m->deleteTag(tag);
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerDeleteTagCmd(int id, int which)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canDelete())
	m->deleteTag(which);
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerEditBeginCmd(int id, int h)
{
  // remember which marker is being edited
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id && m->canEdit()) {
      markerUndo(m, EDIT);

      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void FrameBase::markerEditBeginCmd(const Vector& v, int h)
{
  // remember which marker is being edited
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canEdit()) {
      markerUndo(m, EDIT);

      editMarker = m;
      editMarker->editBegin(h);
      return;
    }
    m=m->next();
  }

  editMarker = NULL;
}

void FrameBase::markerEditMotionCmd(const Vector& v, int h)
{
  if (editMarker) {

    // erase current marker now
    redraw(editMarker->getAllBBox());
    forceUpdate();

    editMarker->edit(v*canvasToRef, h);
    editMarker->drawXOR();
  }
}

void FrameBase::markerEditEndCmd()
{
  if (editMarker)
    editMarker->editEnd();
  editMarker = NULL;

  update(PIXMAP);
}

void FrameBase::markerEllipseRadiusCmd(int id, const Vector& radius, 
				       InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	Vector r = mapLenToRef(radius, sys);
	((Ellipse*)m)->setAnnuli(r);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseRadiusCmd(int id, const Vector& radius, 
				   CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	Vector r = ptr->mapLenToRef(radius, sys, format);
	((Ellipse*)m)->setAnnuli(r);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseAnnulusRadiusCmd(int id, const Vector& inner, 
					      const Vector& outer, int num,
					      InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	Vector r1 = mapLenToRef(inner, sys);
	Vector r2 = mapLenToRef(outer, sys);
	((EllipseAnnulus*)(m))->setAnnuli(r1,r2,num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseAnnulusRadiusCmd(int id, 
		      const Vector& inner, const Vector& outer, int num,
		      CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	Vector r1 = ptr->mapLenToRef(inner, sys, format);
	Vector r2 = ptr->mapLenToRef(outer, sys, format);
	((EllipseAnnulus*)(m))->setAnnuli(r1,r2,num);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseAnnulusRadiusCmd(int id, 
					      const char* lev,
					      CoordSystem sys,SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int cnt = 0;
	Vector radii[MAXANNULI];
	string x(lev);
	istringstream str(x);
	while ((cnt<MAXANNULI) && (str >> radii[cnt][0]))
	  str >> radii[cnt++][1];
	
	// verify proper ratios
	for (int i=0; i<cnt; i++)
	  radii[i][1] = radii[i][0]*radii[cnt-1][1]/radii[cnt-1][0];

	// map to ref coord sys
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<cnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], sys, format);

	((EllipseAnnulus*)(m))->setAnnuli(radii,cnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseAnnulusCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((EllipseAnnulus*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEllipseAnnulusDeleteRadiusCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((EllipseAnnulus*)m)->deleteAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerEpandaCreateAnglesCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Epanda*)m)->addAngles(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEpandaCreateRadiusCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	printInteger(((Epanda*)m)->addAnnuli(v*canvasToRef));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEpandaDeleteCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Epanda*)m)->deleteAnglesAnnuli(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerEpandaEditCmd(int id, 
				    double a1, double a2, int an, 
				    const Vector& r1, const Vector& r2, int rn)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Epanda*)m)->setAnglesAnnuli(a1,a2,an,r1,r2,rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEpandaEditCmd(int id, 
				    double a1, double a2, int an, 
				    const Vector& r1, const Vector& r2, int rn,
				    CoordSystem sys, SkyFrame sky)

{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Epanda*)m)->setAnglesAnnuli(mapAngleToRef(a1,sys,sky),
				      mapAngleToRef(a2,sys,sky),an,r1,r2,rn);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerEpandaEditCmd(int id, 
				    const char* a, const char* r,
				    CoordSystem sys, SkyFrame sky, 
				    CoordSystem rsys, SkyFormat rformat)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());

	int acnt = 0;
	double angles[MAXANGLES];
	{
	  string x(a);
	  istringstream astr(x);
	  while ((acnt<MAXANGLES) && (astr >> angles[acnt]))
	    ++acnt;
	}	
	for (int i=0; i<acnt; i++)
	  angles[i] = mapAngleToRef(degToRad(angles[i]),sys,sky);

	int rcnt = 0;
	Vector radii[MAXANNULI];
	{
	  string x(r);
	  istringstream rstr(x);
	  while ((rcnt<MAXANNULI) && (rstr >> radii[rcnt][0]))
	    rstr >> radii[rcnt++][1];
	}	
	// verify proper ratios
	for (int i=0; i<rcnt; i++)
	  radii[i][1] = radii[i][0]*radii[rcnt-1][1]/radii[rcnt-1][0];

	// map to ref coord sys
	FitsImage* ptr = findFits(m->getCenter());
	for (int i=0; i<rcnt; i++)
	  radii[i] = ptr->mapLenToRef(radii[i], rsys, rformat);

	((Epanda*)m)->setAnglesAnnuli(angles,acnt,radii,rcnt);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerFontCmd(const char* f)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerFontCmd(const char* tag, const char* f)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerFontCmd(int id, const char* f)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP, m->getAllBBox());
      m->setFont(f);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerFrontCmd()
{
  Marker* m = markers->head();
  while (m) {
    if (m->isSelected()) {
      Marker* next = markers->extractNext(m);
      markers->insertHead(m);
      update(PIXMAP, m->getAllBBox());
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerFrontCmd(const char* tag)
{
  Marker* m = markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      Marker* next = markers->extractNext(m);
      markers->insertHead(m);
      update(PIXMAP, m->getAllBBox());
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerFrontCmd(int id)
{
  Marker* m = markers->head();
  while (m) {
    if (m->getId() == id) {
      markers->extractNext(m);
      markers->insertHead(m);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerKeyCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected())
      m->key();
    m=m->next();
  }
}

void FrameBase::markerKeyCmd(const Vector& v)
{
  // v is in canvas coords
  Marker* m=markers->head();
  while (m) {
    if (m->isIn(v))
      m->key();
    m=m->next();
  }
}

void FrameBase::markerLineCmd(int id, const Vector& p1, const Vector& p2, 
			      InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Line*)(m))->setPoints(mapToRef(p1,sys),mapToRef(p2,sys));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerLineCmd(int id, const Vector& p1, const Vector& p2, 
			  CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Line*)(m))->setPoints(ptr->mapToRef(p1,sys,sky),
				ptr->mapToRef(p2,sys,sky));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerLineArrowCmd(int id, int p1, int p2)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	((Line*)(m))->setArrows(p1, p2);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerLineWidthCmd(int w)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      m->setLineWidth(w);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerLineWidthCmd(int id, int w)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->setLineWidth(w);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerListCmd(MarkerFormat type, 
			  CoordSystem sys, SkyFrame sky, SkyFormat format,
			  int strip, int listwcs, int select, 
			  unsigned short mask, unsigned short value, 
			  List<Tag>& tags)
{
  int doSys = 1;

  switch (type) {
  case DS9:
  case CIAO:
  case SAOTNG:
    ostringstream str;
    switch (type) {
    case DS9:
      markerListHeader(str, sys, sky, format, strip, listwcs);
      break;
    case CIAO:
      markerListCiaoHeader(str, sys, sky, format, strip, listwcs);
      break;
    case SAOTNG:
      markerListSAOtngHeader(str, sys, sky, format, strip, listwcs);
      break;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
    break;
  }

  Marker* m=markers->head();
  while (m) {
    ostringstream str;
    Tag* t;

    // selected
    if (select) {
      if (!m->isSelected())
	goto next;
    }

    // properties
    if (mask) {
      if (!((m->getProperty() & mask)  == value))
	goto next;
    }

    // tags
    if (t=tags.head()) {
      while (t) {
	if (!m->hasTag(t->tag()))
	  goto next;
	t=t->next();
      }
    }

    // ok, its passed the tests!
    switch (type) {
    case DS9:
      if (doSys) {
	listCoordSystem(str, sys, sky, currentFits->hasWCSEqu(sys));
	str << (strip ? ';' : '\n');
	doSys = 0;
      }
      m->list(str, sys, sky, format, 0, strip);
      break;
    case CIAO:
      m->listCiao(str, sys, 0, strip);
      break;
    case SAOTNG:
      m->listSAOtng(str, sys, sky, format, strip);
      break;
    case SAOIMAGE:
      m->listSAOimage(str, strip);
      break;
    case PROS:
      m->listPros(str, sys, sky, format, strip);
      break;
    case RAWXY:
      m->listXY(str, sys, sky, format, strip);
      break;
    }
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);

  next:
    m=m->next();
  }
}

void FrameBase::markerLoadCmd(MarkerFormat fm, const char* fn)
{
  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  
  parseMarker(fm, str);
}

void FrameBase::markerLoadCmd(MarkerFormat fm, const char* fn, 
			      CoordSystem sys, SkyFrame sky)
{
  markerDefaultSystem = sys;
  markerDefaultSky = sky;

  ifstream str(fn);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  
  parseMarker(fm, str);
}

void FrameBase::markerLoadCmd(MarkerFormat fm, int fd)
{
  boost::fdistream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  
  parseMarker(fm, str);
}

void FrameBase::markerLoadCmd(MarkerFormat fm, int fd, 
			      CoordSystem sys, SkyFrame sky)
{
  markerDefaultSystem = sys;
  markerDefaultSky = sky;

  boost::fdistream str(fd);
  if (!str) {
    result = TCL_ERROR;
    return;
  }  
  parseMarker(fm, str);
}

void FrameBase::markerLoadFitsCmd(const char* fn, const char* color,
				  int width, const char* font)
{
  if (!currentFits)
    return;

  // verify that we have an ext specified
  if (fn && (fn[strlen(fn)-1] != ']')) {
    result = TCL_ERROR;
    return;
  }

  // do we have a WCS?
  FitsImage* mkfits = NULL;
  {
    FitsHist::Function savBinFunction = binFunction_;
    int savBinBufferSize = binBufferSize_;
    int savBinDepth = binDepth_;
    Vector savBinFactor = binFactor_;

    mkfits = new FitsImageFitsMMap(this, fn);
    if (!mkfits || !mkfits->isValid() || !mkfits->isBinTable()) {
      if (mkfits)
	delete mkfits;
      result = TCL_ERROR;
      return;
    }
  }

  // recenter
  if (*channelFits) {
    FitsImage* ptr = *channelFits;
    mkfits->nextHist(ptr->getHistCursor());
  }

  //  int hasWCS = mkfits->hasWCSEqu(WCS);
  int hasWCS = 0;
  FitsFile* mk = mkfits->fitsFile();
  FitsHead* mkh = mk->head();
  FitsTableHDU* mkhdu = (FitsTableHDU*)mkh->hdu();

  // determine x and y column names
  // if image, hard code 'x' and 'y'
  // however, if table, find out the columns used to bin
  FitsColumn* x;
  FitsColumn* y;
  if (*channelFits) {
    FitsImage* ptr = *channelFits;
    if (ptr->isHist()) {
      x = mkhdu->find(ptr->getHistX());
      y = mkhdu->find(ptr->getHistY());
    }
    else {
      x = mkhdu->find("x");
      y = mkhdu->find("y");
    }
  }
  else {
    x = mkhdu->find("x");
    y = mkhdu->find("y");
  }

  FitsColumn* shape = mkhdu->find("shape");
  FitsColumn* r = mkhdu->find("r");
  FitsColumn* ang = mkhdu->find("rotang");
  
  // manatory columns
  if (!x || !y) {
    if (mkfits)
      delete mkfits;
    result = TCL_ERROR;
    return;
  }

  char* ptr = (char*)mk->data();
  int rows = mkhdu->rows();
  int rowlen = mkhdu->width();

  char text[] = "";
  List<Tag> taglist;
  List<CallBack> cblist;

  for (int i=0; i<rows; i++, ptr+=rowlen) {
    char* s1;
    if (shape)
      s1 = toUpper(shape->str(ptr));
    else {
      s1 = new char[7];
      strcpy(s1,"POINT ");
    }

    // look for '!', which sets include/exclude
    char* s2 = s1;
    unsigned short props = Marker::SELECT | Marker::HIGHLITE | Marker::EDIT |
      Marker::MOVE | Marker::ROTATE | Marker::DELETE | Marker::SOURCE;
    if (s2[0]=='!')
      s2++;
    else
      props |= Marker::INCLUDE;

    Vector center(x->value(ptr,0),y->value(ptr,0));
    if (hasWCS)
      center = currentFits->wcs2pix(mkfits->pix2wcs(center,WCS,FK5),WCS,FK5);

    if (!strncmp(s2, "CIRCLE", 6) && r) {
      Vector rr(r->value(ptr),0);
      if (hasWCS)
	rr = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr,WCS),WCS);
      createCircleCmd(currentFits->mapToRef(center, PHYSICAL),
		      currentFits->mapLenToRef(rr[0], PHYSICAL), 
		      color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "ANNULU", 6) && r) {
      Vector rr0(r->value(ptr,0),0);
      Vector rr1(r->value(ptr,1),0);
      if (hasWCS) {
	rr0 = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr0,WCS),WCS);
	rr1 = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr1,WCS),WCS);
      }
      createAnnulusCmd(currentFits->mapToRef(center ,PHYSICAL),
		       currentFits->mapLenToRef(rr1[0], PHYSICAL),
		       currentFits->mapLenToRef(rr0[0], PHYSICAL), 
		       1,
		       color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "BOX   ", 6) && r) {
      Vector rr(r->value(ptr,0),r->value(ptr,1));
      if (hasWCS)
	rr = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr,WCS),WCS);
      createBoxCmd(currentFits->mapToRef(center, PHYSICAL),
		   currentFits->mapLenToRef(rr, PHYSICAL), 
		   0,
		   color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "ROTBOX", 6) && r && ang) {
      Vector rr(r->value(ptr,0),r->value(ptr,1));
      if (hasWCS)
	rr = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr,WCS),WCS);
      createBoxCmd(currentFits->mapToRef(center, PHYSICAL),
		   currentFits->mapLenToRef(rr, PHYSICAL), 
		   degToRad(ang->value(ptr)),
		   color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "RECTAN", 6)) {
      Vector v1(center);
      Vector v2(x->value(ptr,1), y->value(ptr,1));
      if (hasWCS)
	v2 = currentFits->wcs2pix(mkfits->pix2wcs(v2,WCS,FK5),WCS,FK5);

      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(currentFits->mapToRef(c,PHYSICAL), 
		   currentFits->mapLenToRef(d,PHYSICAL), 
		   0,
		   color, width, font, text, props, NULL, taglist,cblist);
    }

    else if (!strncmp(s2, "ROTREC", 6) && ang) {
      Vector v1(center);
      Vector v2(x->value(ptr,1), y->value(ptr,1));
      if (hasWCS)
	v2 = currentFits->wcs2pix(mkfits->pix2wcs(v2,WCS,FK5),WCS,FK5);

      Vector d = v2-v1;
      Vector c = d/2 + v1;

      createBoxCmd(currentFits->mapToRef(c,PHYSICAL), 
		   currentFits->mapLenToRef(d,PHYSICAL), 
		   degToRad(ang->value(ptr)),
		   color, width, font, text, props, NULL, taglist,cblist);
    }

    else if (!strncmp(s2, "ELLIPS", 6) && r && ang) {
      Vector rr(r->value(ptr,0),r->value(ptr,1));
      if (hasWCS)
	rr = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr,WCS),WCS);
      createEllipseCmd(currentFits->mapToRef(center, PHYSICAL),
		       currentFits->mapLenToRef(rr, PHYSICAL),
		       degToRad(ang->value(ptr)),
		       color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "PIE   ", 6) && r && ang) {
      Vector rr0(r->value(ptr,0),0);
      Vector rr1(r->value(ptr,1),0);
      if (hasWCS) {
	rr0 = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr0,WCS),WCS);
	rr1 = currentFits->wcs2pixDist(mkfits->pix2wcsDist(rr1,WCS),WCS);
      }
      createCpandaCmd(currentFits->mapToRef(center,PHYSICAL),
		      degToRad(ang->value(ptr,0)-90), 
		      degToRad(ang->value(ptr,1)-90), 1,
		      currentFits->mapLenToRef(rr1[0],PHYSICAL),
		      currentFits->mapLenToRef(rr0[0],PHYSICAL), 1,
		      color, width, font, text, props, NULL, taglist,cblist);

    }
    else if (!strncmp(s2, "POINT ", 6))
      createBoxCirclePointCmd(currentFits->mapToRef(center,PHYSICAL),
			      POINTSIZE, 
			      color, width, font, text, props, NULL, taglist,cblist);

    else if (!strncmp(s2, "POLYGO", 6)) {
      List<Vertex> list;
      Vertex* cc = new Vertex(currentFits->mapToRef(center, PHYSICAL));
      list.append(cc);

      // this could be a problem, if the end points are not the same
      int i=1;
      while (1) {
	Vector vv(x->value(ptr,i), y->value(ptr,i));
	if (hasWCS)
	  vv = currentFits->wcs2pix(mkfits->pix2wcs(vv,WCS,FK5),WCS,FK5);

	Vertex* n = new Vertex(currentFits->mapToRef(vv, PHYSICAL));
	if (x->value(ptr,i) != x->value(ptr,0) || 
	    y->value(ptr,i) != y->value(ptr,0))
	  list.append(n);
	else {
	  delete n;
	  break;
	}

	i++;
      }

      createPolygonCmd(list, color, width, font, text, props, NULL, 
		       taglist,cblist);
    }

    delete [] s1;
  }

  if (mkfits)
    delete mkfits;
}

void FrameBase::markerMoveCmd(const Vector& v)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveCmd(const char* tag, const Vector& v)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->canMove() && m->hasTag(tag)) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      Vector c = m->getCenter() * refToCanvas;
      update(PIXMAP, m->getAllBBox());
      m->moveTo((c + v) * canvasToRef);
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveCmd(int id, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canMove()) {
	markerUndo(m, MOVE);

	Vector c = m->getCenter() * refToCanvas;
	update(PIXMAP, m->getAllBBox());
	m->moveTo((c + v) * canvasToRef);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerMoveBeginCmd(const Vector& v)
{
  markerBegin = v * canvasToRef;
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }
      m->moveBegin();
    }
    m=m->next();
  }
}

void FrameBase::markerMoveBeginCmd(int id, const Vector& v)
{
  markerBegin = v * canvasToRef;
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canMove()) {
	if (undoMarkers) {
	  undoMarkers->append(m->dup());
	  undoMarkerType = MOVE;
	}
	m->moveBegin();
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerMoveMotionCmd(const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers->head();
  if (m) {
    while (m) {
      if (m->isSelected() && m->canMove())
	redraw(m->getAllBBox());
      m=m->next();
    }

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    m=markers->head();
    while (m) {
      if (m->isSelected() && m->canMove()) {
	m->move(diff);
	m->drawXOR();
      }
      m=m->next();
    }
  }
}

void FrameBase::markerMoveMotionCmd(int id, const Vector& v)
{
  // first, accumulate erase markers

  Marker* m=markers->head(); 
  if (m) {
    while (m) {
      if (m->getId() && m->canMove()) {
	redraw(m->getAllBBox());
	break;
      }
      m=m->next();
    }

    if (!m)
      return; // can't find it

    Marker *ptr = m;

    // and erase now

    forceUpdate();

    // ok, now draw selected markers in new location

    Vector markerEnd = v * canvasToRef;
    Vector diff = markerEnd - markerBegin;
    markerBegin = markerEnd;

    ptr->move(diff);
    ptr->drawXOR();
  }
}

void FrameBase::markerMoveEndCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canMove())
      m->moveEnd();
    m=m->next();
  }

  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void FrameBase::markerMoveEndCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canMove())
	m->moveEnd();
      return;
    }
    m=m->next();
  }

  // update widget since we don't know where the selected markers came from
  update(PIXMAP);
}

void FrameBase::markerMoveToCmd(const Vector& v, InternalSystem sys)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      update(PIXMAP, m->getAllBBox());
      m->moveTo(mapToRef(v,sys));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveToCmd(const Vector& v, CoordSystem sys, SkyFrame sky)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canMove()) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      m->moveTo(ptr->mapToRef(v,sys,sky));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveToCmd(const char* tag, const Vector& v, 
			    InternalSystem sys)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->canMove() && m->hasTag(tag)) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      update(PIXMAP, m->getAllBBox());
      m->moveTo(mapToRef(v,sys));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveToCmd(const char* tag, const Vector& v, 
			    CoordSystem sys, SkyFrame sky)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=markers->head();
  while (m) {
    if (m->canMove() && m->hasTag(tag)) {
      if (undoMarkers) {
	undoMarkers->append(m->dup());
	undoMarkerType = MOVE;
      }

      update(PIXMAP, m->getAllBBox());
      FitsImage* ptr = findFits(m->getCenter());
      m->moveTo(ptr->mapToRef(v,sys,sky));
      update(PIXMAP, m->getAllBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerMoveToCmd(int id, const Vector& v, InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canMove()) {
	markerUndo(m, MOVE);

	update(PIXMAP, m->getAllBBox());
	m->moveTo(mapToRef(v, sys));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerMoveToCmd(int id, const Vector& v, 
				CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canMove()) {
	markerUndo(m, MOVE);

	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	m->moveTo(ptr->mapToRef(v,sys,sky));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerPointSizeCmd(int id, int size)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Point*)m)->setSize(size);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerPolygonCreateVertexCmd(int id, int seg, const Vector& v)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	((Polygon*)(m))->createVertex(seg, v * canvasToRef);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerPolygonDeleteVertexCmd(int id, int h)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Polygon*)(m))->deleteVertex(h);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerPolygonResetCmd(int id, const Vector& size, 
				      InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Polygon*)(m))->reset(mapLenToRef(size, sys));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerPolygonResetCmd(int id, const Vector& size, 
				      CoordSystem sys, SkyFormat format)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Polygon*)(m))->reset(ptr->mapLenToRef(size, sys, format));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerProjectionCmd(int id,const Vector& p1,const Vector& p2, 
				    InternalSystem sys,double width,int method)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Projection*)(m))->set(mapToRef(p1,sys), mapToRef(p2,sys), 
				mapLenToRef(width,sys),	method);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerProjectionCmd(int id, 
				    const Vector& p1, const Vector& p2, 
				    CoordSystem sys, SkyFrame sky,
				    double width, int method,
				    CoordSystem wdsys, SkyFormat wdformat)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Projection*)(m))->set(ptr->mapToRef(p1,sys,sky),
				ptr->mapToRef(p2,sys,sky), 
				ptr->mapLenToRef(width, wdsys, wdformat),
				method);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerProjectionSystemCmd(int id, CoordSystem sys,SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      ((Projection*)m)->setSystem(sys,sky);
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerPropertyCmd(unsigned short prop, int value)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected()) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE || 
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void FrameBase::markerPropertyCmd(const char* tag, unsigned short prop, int value)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE || 
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void FrameBase::markerPropertyCmd(unsigned short prop, int value, const Vector& v)
{
  // v is in canvas coords

  Marker* m=markers->head();
  while (m) {
    if (m->isIn(v)) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE || 
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);
    }
    m=m->next();
  }
}

void FrameBase::markerPropertyCmd(int id, unsigned short prop, int value)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (prop == Marker::FIXED || 
	  prop == Marker::INCLUDE ||
	  prop == Marker::SOURCE) {
	// marker will change bbox, so get before and after
	update(PIXMAP, m->getAllBBox());
	m->setProperty(prop, value);
	update(PIXMAP, m->getAllBBox());
      }
      else
	m->setProperty(prop, value);

      return;
    }
    m=m->next();
  }
}

void FrameBase::markerRotateBeginCmd(int id)
{
  // remember which marker is being edited
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canRotate()) {
	markerUndo(m, EDIT);
	rotateMarker = m;
	rotateMarker->rotateBegin();
      }
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void FrameBase::markerRotateBeginCmd(const Vector& v)
{
  // remember which marker is being edited
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected() && m->canRotate()) {
      markerUndo(m, EDIT);
      rotateMarker = m;
      rotateMarker->rotateBegin();
      return;
    }
    m=m->next();
  }

  rotateMarker = NULL;
}

void FrameBase::markerRotateMotionCmd(const Vector& v, int h)
{
  if (rotateMarker) {

    // erase current marker now
    redraw(rotateMarker->getAllBBox());
    forceUpdate();

    rotateMarker->rotate(v*canvasToRef, h);
    rotateMarker->drawXOR();
  }
}

void FrameBase::markerRotateEndCmd()
{
  if (rotateMarker)
    rotateMarker->rotateEnd();
  rotateMarker = NULL;
  update(PIXMAP);
}

void FrameBase::markerRulerPointCmd(int id, const Vector& p1, const Vector& p2,
				    InternalSystem sys)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Ruler*)(m))->setPoints(mapToRef(p1,sys),mapToRef(p2,sys));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerRulerPointCmd(int id, const Vector& p1, const Vector& p2,
				CoordSystem sys, SkyFrame sky)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Ruler*)(m))->setPoints(ptr->mapToRef(p1,sys,sky),
				 ptr->mapToRef(p2,sys,sky));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerRulerSystemCmd(int id, CoordSystem sys, SkyFrame sky,
				 CoordSystem distsys, SkyFormat distformat)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      // it may shrink
      update(PIXMAP, m->getAllBBox());
      ((Ruler*)(m))->setCoordSystem(sys, sky, distsys, distformat);
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerSaveCmd(const char* fileName, MarkerFormat type, 
			      CoordSystem sys, SkyFrame sky, SkyFormat format,
			      int strip, int listwcs)
{
  int doSys = 1;

  ofstream fn(fileName);
  if (fn) {
    switch (type) {
    case DS9:
      markerListHeader(fn, sys, sky, format, strip, listwcs);
      break;
    case CIAO:
      markerListCiaoHeader(fn, sys, sky, format, strip, listwcs);
      break;
    case SAOTNG:
      markerListSAOtngHeader(fn, sys, sky, format, strip, listwcs);
      break;
    case SAOIMAGE:
      break;
    case PROS:
      break;
    case RAWXY:
      break;
    }

    Marker* m=markers->head();
    while (m) {
      switch (type) {
      case DS9:
	if (doSys) {
	  listCoordSystem(fn, sys, sky, currentFits->hasWCSEqu(sys));
	  fn << (strip ? ';' : '\n');
	  doSys = 0;
	}
	m->list(fn, sys, sky, format, 0, strip);
	break;
      case CIAO:
	m->listCiao(fn, sys, 0, strip);
	break;
      case SAOTNG:
	m->listSAOtng(fn, sys, sky, format, strip);
	break;
      case SAOIMAGE:
	m->listSAOimage(fn, strip);
	break;
      case PROS:
	m->listPros(fn, sys, sky, format, strip);
	break;
      case RAWXY:
	m->listXY(fn, sys, sky, format, strip);
	break;
      }
      m=m->next();
    }
  }
  else {
    Tcl_AppendResult(interp, "Unable to open file ", fileName, NULL);
    result = TCL_ERROR;
  }
}

void FrameBase::markerSaveTemplateCmd(const char* fileName)
{
  Marker* m=markers->head();
  if (currentFits && m) {
    ofstream fn(fileName);
    if (fn) { 

      FitsImage* ptr = currentFits;
      while (ptr) {
	ptr->initWCS0(m->getCenter());
	ptr = ptr->nextMosaic();
      }

      markerListHeader(fn, WCS0, FK5, DEGREES, 0, 0);
      listCoordSystem(fn, WCS0, FK5, currentFits->hasWCSEqu(WCS0));
      fn << endl;

      while (m) {
	m->list(fn, WCS0, FK5, DEGREES, 0, 0);
	m=m->next();
      }

      ptr = *channelFits;
      while (ptr) {
	ptr->resetWCS0();
	ptr = ptr->nextMosaic();
      }
    }
    else {
      Tcl_AppendResult(interp, "Unable to open file ", fileName, NULL);
      result = TCL_ERROR;
    }
  }
}


// Select

void FrameBase::markerSelectAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect()) {
      m->select();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerSelectCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->getId() == id) {
      m->select();
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerSelectCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->hasTag(tag)) {
      m->select();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerSelectToggleCmd()
{
  // toggle the select of the first found
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect()) {
      m->toggleSelect();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerSelectToggleCmd(const Vector& v)
{
  // toggle the select of the first found
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->isIn(v)) {
      m->toggleSelect();
      update(PIXMAP, m->getBBox());
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::markerSelectOnlyCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->getId() == id) {
      if (!m->isSelected()) {
	m->select();
	update(PIXMAP, m->getBBox());
      }
    }
    else {
      if (m->isSelected()) {
	m->unselect();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  }
}

void FrameBase::markerSelectOnlyCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->hasTag(tag))
      if (!m->isSelected()) {
	m->select();
	update(PIXMAP, m->getBBox());
      }
    else
      if (m->isSelected()) {
	m->unselect();
	update(PIXMAP, m->getBBox());
      }
    m=m->next();
  }
}

void FrameBase::markerSelectOnlyCmd(const Vector& v)
{
  // first, check to see if we clicked on an already selected marker
  Marker* m=markers->head();
  while (m) {
    if (m->canSelect() && m->isIn(v) && m->isSelected()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  // ok, now select the first found, and unselect the rest
  int found = 0;

  m=markers->head();
  while (m) {
    if (m->canSelect() && m->isIn(v) && !found) {
      if (!m->isSelected()) {
	m->select();
	update(PIXMAP, m->getBBox());
      }
      found = 1;
    }
    else {
      if (m->isSelected()) {
	m->unselect();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  } 

  if (found)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::markerUnselectAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    m->unselect();
    update(PIXMAP, m->getBBox());
    m=m->next();
  } 
}

void FrameBase::markerUnselectCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->unselect();
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerUnselectCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      m->unselect();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerVectorCmd(int id, const Vector& p, InternalSystem sys,
			      double mag, double ang)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	((Vect*)(m))->setPoints(mapToRef(p,sys),mapLenToRef(mag,sys),ang);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerVectorCmd(int id, 
			      const Vector& p, CoordSystem sys, SkyFrame sky,
			      double mag, CoordSystem dsys, SkyFormat dformat,
			      double ang)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	markerUndo(m, EDIT);

	// it may shrink
	update(PIXMAP, m->getAllBBox());
	FitsImage* ptr = findFits(m->getCenter());
	((Vect*)(m))->setPoints(ptr->mapToRef(p,sys,sky),
				ptr->mapLenToRef(mag,dsys,dformat),
				mapAngleToRef(ang,sys,sky));
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerVectorArrowCmd(int id, int p)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	((Vect*)(m))->setArrow(p);
	update(PIXMAP, m->getAllBBox());
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

// Highlite

void FrameBase::markerHighliteAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite()) {
      m->highlite();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerHighliteCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->getId() == id) {
      m->highlite();
      update(PIXMAP, m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerHighliteCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->hasTag(tag)) {
      m->highlite();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerHighliteToggleCmd(const Vector& v)
{
  // toggle the highlite of the first found
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->isIn(v)) {
      m->toggleHighlite();
      update(PIXMAP, m->getBBox());
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::markerHighliteOnlyCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->getId() == id) { 
      if (!m->isHighlited()) {
	m->highlite();
	update(PIXMAP, m->getBBox());
      }
    }
    else {
      if (m->isHighlited()) {
	m->unhighlite();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  }
}

void FrameBase::markerHighliteOnlyCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->hasTag(tag))
      if (!m->isHighlited()) {
	m->highlite();
	update(PIXMAP, m->getBBox());
      }
    else
      if (m->isHighlited()) {
	m->unhighlite();
	update(PIXMAP, m->getBBox());
      }
    m=m->next();
  }
}

void FrameBase::markerHighliteOnlyCmd(const Vector& v)
{
  // first, check to see if we clicked on an already highlited marker
  Marker* m=markers->head();
  while (m) {
    if (m->canHighlite() && m->isIn(v) && m->isHighlited()) {
      Tcl_AppendResult(interp, "1", NULL);
      return;
    }
    m=m->next();
  }

  // ok, now highlite the first found, and unhighlite the rest
  int found = 0;

  m=markers->head();
  while (m) {
    if (m->canHighlite() && m->isIn(v) && !found) {
      if (!m->isHighlited()) {
	m->highlite();
	update(PIXMAP, m->getBBox());
      }
      found = 1;
    }
    else {
      if (m->isHighlited()) {
	m->unhighlite();
	update(PIXMAP, m->getBBox());
      }
    }
    m=m->next();
  } 

  if (found)
    Tcl_AppendResult(interp, "1", NULL);
  else
    Tcl_AppendResult(interp, "0", NULL);
}

void FrameBase::markerUnhighliteAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    m->unhighlite();
    update(PIXMAP, m->getBBox());
    m=m->next();
  } 
}

void FrameBase::markerUnhighliteCmd(int id)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->unhighlite();
      update(PIXMAP, m->getBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerUnhighliteCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->hasTag(tag)) {
      m->unhighlite();
      update(PIXMAP, m->getBBox());
    }
    m=m->next();
  }
}

void FrameBase::markerShowCmd(int which)
{
  showMarkers = which;
  update(PIXMAP);
}

void FrameBase::markerShowTextCmd(int which)
{
  showMarkersText = which;
  update(PIXMAP);
}

void FrameBase::markerTagCmd(const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->isSelected())
      m->addTag(tag);
    m=m->next();
  }
}

void FrameBase::markerTagCmd(int id, const char* tag)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      m->addTag(tag);
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerTagEditCmd(const char* from, const char* to)
{
  Marker* m=markers->head();
  while (m) {
    m->editTag(from, to);
    m=m->next();
  }
}

void FrameBase::markerTagDeleteCmd(const char* t)
{
  Marker* m=markers->head();
  while (m) {
    m->deleteTag(t);
    m=m->next();
  }
}

void FrameBase::markerTagDeleteAllCmd()
{
  Marker* m=markers->head();
  while (m) {
    m->deleteTags();
    m=m->next();
  }
}

void FrameBase::markerTagUpdateCmd(const char* t)
{
  markerTagDeleteCmd(t);
  markerTagCmd(t);
}

void FrameBase::markerTextCmd(int id, const char* text)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      // things can shrink, so do before and after
      update(PIXMAP,m->getAllBBox());
      m->setText(text);
      update(PIXMAP,m->getAllBBox());
      return;
    }
    m=m->next();
  }
}

void FrameBase::markerTextRotateCmd(int id, int rot)
{
  Marker* m=markers->head();
  while (m) {
    if (m->getId() == id) {
      if (m->canEdit()) {
	((Text*)(m))->setRotate(rot);
	update(PIXMAP);
      }
      return;
    }
    m=m->next();
  }

  result = TCL_ERROR;
}

void FrameBase::markerCopyCmd()
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  pasteMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->isSelected())
      pasteMarkers->append(m->dup());
    m=m->next();
  }
}

void FrameBase::markerCopyCmd(const char* tag)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  pasteMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->hasTag(tag))
      pasteMarkers->append(m->dup());
    m=m->next();
  }
}

void FrameBase::markerCutCmd()
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  pasteMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->isSelected() && m->canDelete()) {
      Marker* next = markers->extractNext(m);
      update(PIXMAP);
      pasteMarkers->append(m);
      m->doCallBack(CallBack::DELETECB);
      m->disableCB();
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerCutCmd(const char* tag)
{
  if (undoMarkers)
    undoMarkers->deleteAll();
  pasteMarkers->deleteAll();
  Marker* m = markers->head();
  while (m) {
    if (m->canDelete() && m->hasTag(tag)) {
      Marker* next = markers->extractNext(m);
      update(PIXMAP);
      pasteMarkers->append(m);
      m->doCallBack(CallBack::DELETECB);
      m->disableCB();
      m = next;
    }
    else
      m=m->next();
  }
}

void FrameBase::markerPasteCmd()
{
  // unselect markers
  {
    Marker* m=markers->head();
    while (m) {
      m->unselect();
      m=m->next();
    }
  }

  if (undoMarkers)
    undoMarkers->deleteAll();
  Marker* m=pasteMarkers->head();
  while (m) {
    Marker* n = m->dup();
    n->newIdentity();
    markers->append(n);

    m = m->next();
  }

  update(PIXMAP);
}


void FrameBase::markerPasteCmd(CoordSystem sys, SkyFrame sky)
{
  MarkerFormat type = DS9;
  SkyFormat format = DEGREES;

  {
    ostringstream str;
    markerListHeader(str, sys, sky, format, 0, 0);
    listCoordSystem(str,sys,sky,currentFits->hasWCSEqu(sys));
    str << endl << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  Marker* m=pasteMarkers->head();
  while (m) {
    ostringstream str;
    m->list(str, sys, sky, format, 0, 0);
    str << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
    m=m->next();
  }
}

void FrameBase::markerUndoCmd()
{
  if (!undoMarkers)
    return;

  Marker* n=undoMarkers->head();
  while (n) {
    Marker* next = n->next();
    undoMarkers->extractPrev(n);
    
    switch (undoMarkerType) {
    case NONE:
      break;
    case DELETE:
      markers->append(n);
      n->updateBBox();
      update(PIXMAP,n->getAllBBox());
      break;
    case EDIT:
    case MOVE:
      {
	Marker* m=markers->head();
	while (m) {
	  if (m->getId() == n->getId()) {
	    n->enableCB();
	    m->updateBBox();
	    update(PIXMAP,m->getAllBBox());

	    markers->insertNext(m,n);
	    markers->extractNext(m);

	    n->updateBBox();
	    update(PIXMAP,n->getAllBBox());
	    switch (undoMarkerType) {
	    case EDIT:
	      n->doCallBack(CallBack::EDITCB);
	      break;
	    case MOVE:
	      n->doCallBack(CallBack::MOVECB);
	      break;
	    }

	    m->disableCB();
	    delete m;
	    break;
	  }
	  m=m->next();
	}
      }
      break;
    }

    n=next;
  }

  undoMarkerType = NONE;
}

// region highlite
void FrameBase::regionHighliteEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers->head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canHighlite())
      m->highlite();
    else
      m->unhighlite();
    m=m->next();
  } 

  update(PIXMAP, bb.expand(2));
}

void FrameBase::regionHighliteShiftEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers->head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canHighlite())
      m->highlite();
    m=m->next();
  }

  update(PIXMAP, bb.expand(2));
}

// region select
void FrameBase::regionSelectBeginCmd(const Vector& v)
{
  regionBegin = v;
  regionEnd = v;
}

void FrameBase::regionSelectMotionCmd(const Vector& v)
{
  if (visible) {

    // erase 

    redrawNow((BBox(regionBegin, regionEnd)).expand(2));

    // and draw to window

    regionEnd = v;
    BBox cc = BBox(regionBegin, regionEnd) * canvasToWindow;
    Vector size = cc.size();

    XDrawRectangle(display, Tk_WindowId(tkwin), selectGCXOR,
		   cc.ll[0], cc.ll[1], size[0], size[1]);
  }
}

void FrameBase::regionSelectEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers->head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canSelect())
      m->select();
    else
      m->unselect();
    m=m->next();
  } 

  update(PIXMAP, bb.expand(2));
}

void FrameBase::regionSelectShiftEndCmd()
{
  BBox bb(regionBegin, regionEnd);

  Marker* m=markers->head();
  while (m) {
    if (bb.isIn(m->getBBox())==4 && m->canSelect())
      m->select();
    m=m->next();
  }

  update(PIXMAP, bb.expand(2));
}

// Marker Support

void FrameBase::markerUndo(Marker* m, UndoMarkerType t)
{
  if (!undoMarkers)
    return;

  undoMarkers->deleteAll();
  undoMarkers->append(m->dup());
  undoMarkerType = t;
}

void FrameBase::parseMarker(MarkerFormat fm, istream& str)
{
  switch (fm) {
  case DS9: 
    {
      mkFlexLexer* ll = new mkFlexLexer(&str);
      mkparse(this, ll);
      delete ll;

      resetCompositeMarker();
    }
    break;
  case CIAO:
    {
      ciaoFlexLexer* ll = new ciaoFlexLexer(&str);
      ciaoparse(this, ll);
      delete ll;
    }
    break;
  case PROS:
    {
      prosFlexLexer* ll = new prosFlexLexer(&str);
      prosparse(this, ll);
      delete ll;
    }
    break;
  case SAOTNG:
    {
      tngFlexLexer* ll = new tngFlexLexer(&str);
      tngparse(this, ll);
      delete ll;
    }
    break;

  case SAOIMAGE:
    {
      saoFlexLexer* ll = new saoFlexLexer(&str);
      saoparse(this, ll);
      delete ll;
    }
    break;

  case RAWXY:
    {
      xyFlexLexer* ll = new xyFlexLexer(&str);
      xyparse(this, ll);
      delete ll;
    }
    break;
  }
}

void FrameBase::psMarkers(List<Marker>* ml, int mode)
{
  // render from back to front
  // bbox is in canvas coords
  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  Marker* m=ml->tail();
  while (m) {
    if (m->isVisible(bb))
      m->ps(mode, showMarkersText);
    m=m->previous();
  }
}

#ifdef _MACOSX
void FrameBase::macosxMarkers(List<Marker>* ml)
{
  // render from back to front
  // bbox is in canvas coords
  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  Marker* m=ml->tail();
  while (m) {
    if (m->isVisible(bb))
      m->macosx(showMarkersText);
    m=m->previous();
  }
}
#endif

#ifdef _WIN32
void FrameBase::win32Markers(List<Marker>* ml)
{
  // render from back to front
  // bbox is in canvas coords
  const BBox bb = BBox(0, 0, options->width-1, options->height-1) * 
    widgetToCanvas;

  Marker* m=ml->tail();
  while (m) {
    if (m->isVisible(bb))
      m->win32(showMarkersText);
    m=m->previous();
  }
}
#endif

void FrameBase::renderMarkers(List<Marker>* ml, const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords
  Marker* m=ml->tail();
  while (m) {
    if (m->isVisible(bb))
      m->draw(pixmap, showMarkersText);
    m=m->previous();
  }
}

void FrameBase::renderMagnifierMarkers(List<Marker>* ml, const Matrix& mx,
				       const BBox& bb)
{
  // render from back to front
  // bbox is in canvas coords
  Marker* m=ml->tail();
  while (m) {
    if (m->isVisible(bb))
      m->drawMagnifier(mx, showMarkersText);
    m=m->previous();
  }
}

void FrameBase::unselectMarkers(List<Marker>* ml)
{
  Marker* m=ml->head();
  while (m) {
    m->unselect();
    m=m->next();
  }
}

void FrameBase::unhighliteMarkers()
{
  Marker* m=markers->head();
  while (m) {
    m->unhighlite();
    m=m->next();
  }
}

void FrameBase::updateCBMarkers()
{
  Marker* m=markers->head();
  while (m) {
    m->doCallBack(CallBack::UPDATECB);
    m=m->next();
  }
}

void FrameBase::updateMarkers(List<Marker>* ml)
{
  Marker* m=ml->head();
  while (m) {
    m->updateBBox();
    m=m->next();
  }
}

void FrameBase::updateMarkerCoords(List<Marker>* ml, const Matrix& mx)
{
  Marker* m=ml->head();
  while (m) {
    m->updateCoords(mx);
    m=m->next();
  }
}

void FrameBase::updateMarkerCBs(List<Marker>* ml)
{
  Marker* m=ml->head();
  while (m) {
    m->doCallBack(CallBack::MOVECB);
    m->doCallBack(CallBack::EDITCB);
    m->doCallBack(CallBack::ROTATECB);
    m=m->next();
  }
}

void FrameBase::markerPrintCoord(const Vector& v, InternalSystem sys)
{
  printVector(mapFromRef(v, sys), DEFAULT);
}

void FrameBase::markerPrintCoord(const Vector& c, const Vector& v,
			     CoordSystem sys, SkyFrame sky, SkyFormat format)
{
  if (!isMosaic()) {
    if (currentFits)
      printFromRef(currentFits, v, sys, sky, format, DEFAULT);
  }
  else {
    if (*channelFits) {
      FitsImage* ptr = *channelFits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(currentScale->scanMode());
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printFromRef(ptr, v, sys, sky, format, DEFAULT);
	  return;
	}

	ptr = ptr->nextMosaic();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case AMPLIFIER:
    case PHYSICAL:
    case IMAGE:
      break;
    default:
      if (currentFits)
	printFromRef(currentFits, v, sys, sky, format, DEFAULT);
      break;
    }
  }
}

void FrameBase::markerPrintDouble(double d, InternalSystem sys)
{
  printDouble(mapLenFromRef(d,sys), DEFAULT);
}

void FrameBase::markerPrintDouble(const Vector& c, double d, 
				  CoordSystem sys, SkyFormat format)
{
  if (!isMosaic()) {
    if (currentFits)
      printDouble(currentFits->mapLenFromRef(d,sys,format), DEFAULT);
  }
  else {
    if (*channelFits) {
      FitsImage* ptr = *channelFits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(currentScale->scanMode());
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printDouble(ptr->mapLenFromRef(d,sys,format), DEFAULT);
	  return;
	}

	ptr = ptr->nextMosaic();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case IMAGE:
    case PHYSICAL:
    case AMPLIFIER:
      break;
    default:
      if (currentFits)
	printDouble(currentFits->mapLenFromRef(d,sys,format), DEFAULT);
      break;
    }
  }
}

void FrameBase::markerPrintDouble(const Vector& p1, const Vector& p2,
				  InternalSystem sys)
{
  printDouble(mapDistFromRef(p1,p2,sys), DEFAULT);
}

void FrameBase::markerPrintDouble(const Vector& c,
				  const Vector& p1, const Vector& p2,
				  CoordSystem sys, SkyFormat format)
{
  if (!isMosaic()) {
    if (currentFits)
      printDouble(currentFits->mapDistFromRef(p1,p2,sys,format), DEFAULT);
  } 
  else {
    if (*channelFits) {
      FitsImage* ptr = *channelFits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(currentScale->scanMode());
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printDouble(ptr->mapDistFromRef(p1, p2, sys, format), DEFAULT);
	  return;
	}

	ptr = ptr->nextMosaic();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case IMAGE:
    case PHYSICAL:
    case AMPLIFIER:
      break;
    default:
      if (currentFits)
	printDouble(currentFits->mapDistFromRef(p1,p2,sys,format), DEFAULT);
      break;
    }
  }
}

void FrameBase::markerPrintVector(const Vector& v, InternalSystem sys)
{
  printVector(mapLenFromRef(v,sys), DEFAULT);
}

void FrameBase::markerPrintVector(const Vector& c, const Vector& v,
			      CoordSystem sys, SkyFormat format)
{
  if (!isMosaic()) {
    if (currentFits)
      printVector(currentFits->mapLenFromRef(v,sys,format), DEFAULT);
  }
  else {
    if (*channelFits) {
      FitsImage* ptr = *channelFits;
      while (ptr) {
	Vector img = c * ptr->getRefToData();
	int* params = ptr->getDataParams(currentScale->scanMode());
	int& xmin = params[1];
	int& xmax = params[2];
	int& ymin = params[3];
	int& ymax = params[4];

	if (img[0]>=xmin && img[0]<xmax && img[1]>=ymin && img[1]<ymax) {
	  printVector(ptr->mapLenFromRef(v,sys,format), DEFAULT);
	  return;
	}

	ptr = ptr->nextMosaic();
      }
    }

    // use default coordinate frame
    switch (sys) {
    case IMAGE:
    case PHYSICAL:
    case AMPLIFIER:
      break;
    default:
      if (currentFits)
	printVector(currentFits->mapLenFromRef(v,sys,format), DEFAULT);
      break;
    }
  }
}

void FrameBase::markerListHeader(ostream& str, CoordSystem sys, 
				 SkyFrame sky, SkyFormat format, 
				 int strip, int listwcs)
{
  // no comments for semicolons
  if (!strip) {
    // header
    str << "# Region file format: DS9 version 4.0" << endl;
    if (currentFits)
      str << "# Filename: " <<  currentFits->getFullBaseFileName() << endl;

    if (listwcs)
      currentFits->listWCS(str, sys);

    str << "global color=green font=\"helvetica 10 normal\" "
	<< "select=1 highlite=1 edit=1 move=1 delete=1 include=1 fixed=0 source" 
	<< endl;
  }
}

void FrameBase::markerListCiaoHeader(ostream& str, CoordSystem sys, 
				     SkyFrame sky, SkyFormat format, 
				     int strip, int listwcs)
{
  // no comments for semicolons
  if (!strip)
    str << "# Region file format: CIAO version 1.0" << endl;
}

void FrameBase::markerListSAOtngHeader(ostream& str, CoordSystem sys,
				       SkyFrame sky, SkyFormat format, 
				       int strip, int listwcs)
{
  // no comments for semicolons
  if (strip)
    return;

  if (currentFits)
    str << "# filename: " << currentFits->getRootFileName() << endl;

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case AMPLIFIER:
  case DETECTOR:
    str << "# format: pixels (physical)" << endl;
    break;

  default:
    str << "# format: ";
    switch (sky) {
    case FK4:
    case FK5:
    case ICRS:
    case ECLIPTIC:
    case GALACTIC:
      switch (format) {
      case DEGREES:
	str << "degrees (";
	break;
      case SEXAGESIMAL:
      case HMS:
	str << "hms (";
	break;
      }

      switch (sky) {
      case FK4:
	str << "fk4";
	break;
      case FK5:
	str << "fk5";
	break;
      case ICRS:
	str << "icrs";
	break;
      case GALACTIC:
	str << "galactic";
	break;
      case ECLIPTIC:
	str << "ecliptic";
	break;
      }

      str << ')' << endl;
      break;

    default:
      str << "linear";
    }
  }
}
