// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include <iostream>
#include <math.h>

#include "inversescale.h"
#include "framebase.h"
#include "util.h"

#include "NaN.h"

InverseScale::InverseScale(int s) : size_(s)
{
  level_ = new double[size_];

  for (int i=0; i<size_; i++)
    level_[i] = 0;
}

InverseScale::InverseScale(int s, double* l) : size_(s)
{
  level_ = new double[size_];

  for (int i=0; i<size_; i++)
    level_[i] = l[i];
}

InverseScale::~InverseScale()
{
  if (level_)
    delete [] level_;
}

ostream& operator<<(ostream& s, const InverseScale& c)
{
  for (int i=0; i<c.size_; i++)
    if (!isnand(c.level_[i]))
      s << c.level_[i] << ' ';
    else
      s << 0 << ' ';
  return s;
}

// Note: this an inverse algorithm. We have y, we need x

LinearInverseScale::LinearInverseScale(int s, double low, double high) 
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  double dd = high-low;
  for (int i=0; i<size_; i++) {
    double a = double(i)/(size_-1);
    level_[i] = a*dd +low;
  }
}

LogInverseScale::LogInverseScale(int s, double low, double high) 
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  CheckScaleLog(&low,&high);
  for (int i=0; i<size_; i++)
    level_[i] = pow(high/low, double(i)/(size_-1)) *low;
}

PowInverseScale::PowInverseScale(int s, double low, double high) 
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  CheckScaleLog(&low,&high);
  double dd = high-low;
  for (int i=0; i<size_; i++) {
    double a = log10(high/low*double(i)/(size_-1) +1)/log10(high/low +1);
    level_[i] =  a*dd  +low;
  }
}

SqrtInverseScale::SqrtInverseScale(int s, double low, double high) 
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  double dd = high-low;
  for (int i=0; i<size_; i++) {
    double a = double(i)/(size_-1);
    level_[i] = (a*a)*dd +low;
  }
}

SquaredInverseScale::SquaredInverseScale(int s, double low, double high)
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  double dd = high-low;
  for (int i=0; i<size_; i++) {
    double a = sqrt(double(i)/(size_-1));
    level_[i] = a*dd +low;
  }
}

HistEquInverseScale::HistEquInverseScale(int s, double low, double high, 
					 double* hist, int histsize) 
  : InverseScale(s)
{
  if (size_==1) {
    level_[0] = high;
    return;
  }

  double dd = high-low;
  if (!hist) {
    for (int i=0; i<size_; i++) {
      double a = double(i)/(size_-1);
      level_[i] = a*dd +low;
    }
  }
  else {
    for (int i=0; i<size_; i++) {
      double v = double(i)/(size_-1);
      int j=0;
      while (j<histsize-1) {
	if (hist[j]>v)
	  break;
	j++;
      }
      double a = double(j)/histsize;
      level_[i] = a*dd +low;
    }
  }
}

