// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "polygon.h"
#include "framebase.h"
#include "fitsimage.h"
#include "util.h"

Polygon::Polygon(const Polygon& a) : Marker(a)
{
  vertex = a.vertex;
}

Polygon::Polygon(FrameBase* p, const Vector& ctr,
		 const Vector& b,
		 const char* clr, int wth, 
		 const char* fnt, const char* txt,
		 unsigned short prop, const char* cmt,
		 const List<Tag>& tg, const List<CallBack>& cb)
  : Marker(p, ctr, 0, clr, wth, fnt, txt, prop, cmt, tg, cb)
{
  angle = 0;
  strcpy(type, "polygon");

  Vector bb = b;
  vertex.append(new Vertex(-bb[0],-bb[1]));
  vertex.append(new Vertex( bb[0],-bb[1]));
  vertex.append(new Vertex( bb[0], bb[1]));
  vertex.append(new Vertex(-bb[0], bb[1]));

  updateBBox();
}

Polygon::Polygon(FrameBase* p, const List<Vertex>& v, 
		 const char* clr, int wth, 
		 const char* fnt, const char* txt,
		 unsigned short prop, const char* cmt,
		 const List<Tag>& tg, const List<CallBack>& cb)
  : Marker(p, Vector(0,0), 0, clr, wth, fnt, txt, prop, cmt, tg, cb)
{
  // Vertex list is in ref coords
  angle = 0;
  strcpy(type, "polygon");
  vertex = v;

  // check to see if the first and last node are the same
  if (vertex.head()->vector[0] == vertex.tail()->vector[0] &&
      vertex.head()->vector[1] == vertex.tail()->vector[1])
    delete vertex.pop();

  // find center
  center = Vector(0,0);
  vertex.head();
  do
    center += vertex.current()->vector;
  while (vertex.next());
  center /= vertex.count();

  // vertices are relative
  vertex.head();
  do
    vertex.current()->vector *= Translate(-center) * flip_; // no rotation
  while (vertex.next());

  updateBBox();
}

void Polygon::renderX(Drawable drawable, const Matrix& mx, RenderMode mode)
{
  GC lgc = renderXGC(mode);

  Matrix mm = fwdRefMatrix() * mx;

  vertex.head();
  Vector v1;
  Vector v2 = (vertex.current()->vector * mm).round();
  int done = 0;

  do {
    if (!vertex.next()) {
      done = 1;
      vertex.head();
    }
    v1 = v2;
    v2 = (vertex.current()->vector * mm).round();
    XDrawLine(display, drawable, lgc, (int)v1[0], (int)v1[1], 
	      (int)v2[0], (int)v2[1]);
  } while (!done);

  // remember, bbox is in canvas coords
  if (!(properties & INCLUDE)) {
    Vector ll = (handle[0] * parent->getCanvasToRef() * mx).round();
    Vector ur = (handle[2] * parent->getCanvasToRef() * mx).round();

    if (mode==SRC)
      XSetForeground(display, gc, parent->getColor("red"));

    XDrawLine(display, drawable, lgc, (int)ll[0], (int)ll[1], 
	      (int)ur[0], (int)ur[1]);    
  }
}

void Polygon::renderPS(int mode)
{
  renderPSGC(mode);

  Matrix mm = fwdCanvasMatrix();

  vertex.head();
  int first = 1;
  do {
    ostringstream str;
    Vector v =  vertex.current()->vector * mm;
    if (first) {
      str << "newpath " << endl
	  << v.TkCanvasPs(parent->canvas) << " moveto" << endl << ends;
      first = 0;
    }
    else
      str << v.TkCanvasPs(parent->canvas) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  } while (vertex.next());

  ostringstream str;
  str << "closepath stroke" << endl << ends;
  Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);

  // remember, bbox is in canvas coords
  if (!(properties & INCLUDE)) {
    renderPSColor(mode, parent->getXColor("red"));

    ostringstream str;
    str << "newpath " 
	<< handle[0].TkCanvasPs(parent->canvas) << "moveto"
	<< handle[2].TkCanvasPs(parent->canvas) << "lineto"
	<< " stroke" << endl << ends;
    Tcl_AppendResult(parent->interp, str.str().c_str(), NULL);
  }
}

#ifdef _MACOSX
void Polygon::renderMACOSX()
{
  renderMACOSXGC();

  Matrix mm = fwdCanvasMatrix();

  vertex.head();
  Vector v1;
  Vector v2 = vertex.current()->vector * mm;
  int done = 0;

  do {
    if (!vertex.next()) {
      done = 1;
      vertex.head();
    }
    v1 = v2;
    v2 = vertex.current()->vector * mm;

    macosxDrawLine(v1,v2);
  } while (!done);

  // remember, bbox is in canvas coords
  if (!(properties & INCLUDE)) {
    macosxColor(parent->getXColor("red"));
    macosxDrawLine(handle[0], handle[2]);
  }
}
#endif

#ifdef _WIN32
void Polygon::renderWIN32()
{
  renderWIN32GC();

  Matrix mm = fwdCanvasMatrix();

  vertex.head();
  Vector v1;
  Vector v2 = vertex.current()->vector * mm;
  int done = 0;

  do {
    if (!vertex.next()) {
      done = 1;
      vertex.head();
    }
    v1 = v2;
    v2 = vertex.current()->vector * mm;

    win32DrawLine(v1,v2);
  } while (!done);

  // remember, bbox is in canvas coords
  if (!(properties & INCLUDE)) {
    win32Color(parent->getXColor("red"));
    win32DrawLine(handle[0], handle[2]);
  }
}
#endif

void Polygon::updateBBox()
{
  // generate handles
  numHandle = 4 + vertex.count();
  if (handle)
    delete [] handle;
  handle = new Vector[numHandle];

  // the first four are our control handles
  BBox bb;
  vertex.head();
  do
    bb.bound(vertex.current()->vector);
  while (vertex.next());

  Vector zz = parent->zoom();
  float r = 10/zz.length();
  bb.expand(r); // give us more room

  Matrix mm = fwdCanvasMatrix();

  handle[0] = bb.ll   * mm;
  handle[1] = bb.lr() * mm;
  handle[2] = bb.ur   * mm;
  handle[3] = bb.ul() * mm;

  // and the rest are vertices
  int i=4;
  vertex.head();
  do
    handle[i++] = vertex.current()->vector * mm;
  while (vertex.next());

  // now, calc bbox
  Vector cc = center * parent->refToCanvas;
  bbox = BBox(cc);
  for (int i=0; i<numHandle; i++)
    bbox.bound(handle[i]);
  bbox.expand(3); // give us room for control handles

  // calculate overall bbox
  calcAllBBox();
}

void Polygon::updateCoords(const Matrix& mx)
{
  Scale s(mx);
  vertex.head();
  do
    vertex.current()->vector *= s;
  while (vertex.next());
  
  Marker::updateCoords(mx);
}

void Polygon::edit(const Vector& v, int h)
{
  if (h < 5) {
    Vector s1 = v * bckCanvasMatrix();
    Vector s2 = handle[h-1] * bckRefMatrix();

    if (s1[0] != 0 && s1[1] != 0 && s2[0] != 0 && s2[1] != 0) {
      double a = fabs(s1[0]/s2[0]);
      double b = fabs(s1[1]/s2[1]);
      double s = a > b ? a : b;

      vertex.head();
      do
	vertex.current()->vector *= Scale(s);
      while (vertex.next());
    }

    updateBBox();
    doCallBack(CallBack::EDITCB);
  }
  else {
    moveVertex(v,h);

    updateBBox();
    doCallBack(CallBack::EDITCB);
    doCallBack(CallBack::MOVECB); // center can change
  }
}

void Polygon::rotate(const Vector& v, int h)
{
  if (h < 5)
    Marker::rotate(v,h);
  else {
    // we need to check this here, because we are really rotating
    if (canEdit()) { 
      moveVertex(v,h);

      updateBBox();
      doCallBack(CallBack::EDITCB);
      doCallBack(CallBack::MOVECB); // center can change
    }
  }
}

void Polygon::reset(const Vector& b)
{
  angle = 0;
  vertex.deleteAll();

  Vector bb = b;
  vertex.append(new Vertex(-bb[0],-bb[1]));
  vertex.append(new Vertex( bb[0],-bb[1]));
  vertex.append(new Vertex( bb[0], bb[1]));
  vertex.append(new Vertex(-bb[0], bb[1]));

  updateBBox();
}

int Polygon::isIn(const Vector& vv)
{
  /*
    v[0]-- x value of point being tested
    v[1]-- y value of point being tested

    This algorithm is from "An Introduction to Ray Tracing", Academic Press,
    1989, edited by Andrew Glassner, pg 53
    -- a point lies in a polygon if a line is extended from the point to 
    infinite in any direction and the number of intersections with the 
    polygon is odd.
    This is valid for both concave and convex polygons.
    Points on a vertex are considered inside.
    Points on a edge are considered inside.
  */

  Matrix mm = bckRefMatrix();

  Vector v = vv * mm; // xform to ref coords
  int crossings = 0;   // number of crossings

  vertex.head();
  Vector v1;
  Vector v2 = vertex.current()->vector - v;

  int sign = ((v2[1])>=0) ? 1 : -1; // init sign

  // for all edges

  int done = 0;

  do {
    // look at next two vertices
    v1 = v2;
    if (!vertex.next()) {
      done = 1;
      vertex.head();
    }

    v2 = vertex.current()->vector - v;

    int nextSign = (v2[1]>=0) ? 1 : -1; // sign holder for p2

    if (sign != nextSign) {
      if (v1[0]>0 && v2[0]>0)
	crossings++;
      else if (v1[0]>0 || v2[0]>0) {
	if (v1[0]-(v1[1]*(v2[0]-v1[0])/(v2[1]-v1[1])) > 0)
	  crossings++;
      }
      sign = nextSign;
    }
  } while (!done);

  return fmod(float(crossings),float(2)) ? 1 : 0; // if odd, point is inside
}

void Polygon::createVertex(int which, const Vector& v)
{
  // which segment (1 to n)
  // v is in ref coords

  Matrix mm = bckCanvasMatrix();

  int seg = which-1;
  if (seg>=0 && seg<vertex.count()) {
    Vertex* n = new Vertex(v * mm);
    vertex.insert(seg,n);

    recalcCenter();

    updateBBox();
    doCallBack(CallBack::EDITCB);
    doCallBack(CallBack::MOVECB); // center can change
  }
}

void Polygon::deleteVertex(int h)
{
  if (h>4) {
    int hh = h-4-1;

    if (vertex.count() > 3) {
      Vertex* v = vertex[hh];
      if (v) {
	vertex.extractNext(v);
	delete v;

	recalcCenter();

	updateBBox();
	doCallBack(CallBack::EDITCB);
	doCallBack(CallBack::MOVECB); // center can change
      }  
    }
  }
}

int Polygon::getSegment(const Vector& v)
{
  // v is in canvas coords

  Matrix mm = fwdRefMatrix();

  vertex.head();
  Vector v1;
  Vector v2 = vertex.current()->vector * mm;
  int done = 0;
  int i = 1;
  do {
    v1 = v2;
    if (!vertex.next()) {
      vertex.head();
      done = 1;
    }
    v2 = vertex.current()->vector * mm;

    Vector l1 = v1 * parent->refToCanvas;
    Vector l2 = v2 * parent->refToCanvas;
    double a = parent->isIIS() ? -(l2-l1).angle() : (l2-l1).angle();
    Matrix mx = Translate(-l1) * flip_ * Rotate(-a); 
    Vector end = l2*mx;
    Vector vv = v*mx;
    
    if (vv[0]>0 && vv[0]<end[0] && vv[1]>-3 && vv[1]<3)
      return i;

    i++;
  } while (!done);

  return 0;
}

void Polygon::moveVertex(const Vector& v, int h)
{
  Matrix mm = bckCanvasMatrix();

  if (vertex[h-5])
    vertex.current()->vector = v * mm;

  recalcCenter();
}

void Polygon::recalcCenter()
{
  // recalculate center

  Vector nc;
  vertex.head();
  do
    nc += vertex.current()->vector * Rotate(angle) * flip_;
  while (vertex.next());
  nc /= vertex.count();

  center += nc;

  // update all vertices

  vertex.head();
  do
    vertex.current()->vector -= nc * flip_ * Rotate(-angle);
  while (vertex.next());
}

// list

void Polygon::list(ostream& str, CoordSystem sys, SkyFrame sky,
		   SkyFormat format, int conj, int strip)
{
  FitsImage* ptr = parent->findFits(center);
  listPre(str, sys, sky, ptr, strip, 0);

  Matrix mm = fwdRefMatrix();
  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      str << type << '(';
      int first=1;
      vertex.head();
      do {
	if (!first)
	  str << ',';
	first=0;

	Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys);
	str << setprecision(8) << v[0] << ',' << v[1];
      }
      while (vertex.next());
      str << ')';
    }
    break;
  default:
    if (ptr->hasWCS(sys)) {
      if (ptr->hasWCSEqu(sys)) {
	switch (format) {
	case DEGREES:
	  {
	    str << type << '(';
	    int first=1;
	    vertex.head();
	    do {
	      if (!first)
		str << ',';
	      first=0;

	      Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys,sky);
	      str << setprecision(8) << v[0] << ',' << v[1];
	    }
	    while (vertex.next());
	    str << ')';
	  }
	  break;
	case SEXAGESIMAL:
	  {
	    char buf[64];
	    char ra[16];
	    char dec[16];

	    str << type << '(';
	    int first=1;
	    vertex.head();
	    do {
	      if (!first)
		str << ',';
	      first=0;

	      ptr->mapFromRef(vertex.current()->vector*mm,sys,sky,format,buf,64);
	      string x(buf);
	      istringstream wcs(x);
	      wcs >> ra >> dec;
	      str << ra << ',' << dec;
	    }
	    while (vertex.next());
	    str << ')';
	  }
	  break;
	}
      }
      else {
	str << type << '(';
	int first=1;
	vertex.head();
	do {
	  if (!first)
	    str << ',';
	  first=0;

	  Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys);
	  str << setprecision(8) << v[0] << ',' << v[1];
	}
	while (vertex.next());
	str << ')';
      }
    }
  }

  listPost(str, conj, strip);
}

void Polygon::listCiao(ostream& str, CoordSystem sys, int conj, int strip)
{
  FitsImage* ptr = parent->findFits(1);
  Matrix mm = fwdRefMatrix();
  listCiaoPre(str);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      str << type << '(';
      int first=1;
      vertex.head();
      do {
	if (!first)
	  str << ',';
	first=0;

	Vector v = ptr->mapFromRef(vertex.current()->vector * mm,PHYSICAL);
	str << setprecision(8) << v[0] << ',' << v[1];
      }
      while (vertex.next());
      str << ')';
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      char buf[64];
      char ra[16];
      char dec[16];

      str << type << '(';
      int first=1;
      vertex.head();
      do {
	if (!first)
	  str << ',';
	first=0;

	ptr->mapFromRef(vertex.current()->vector*mm,sys,FK5,SEXAGESIMAL,buf,64);
	string x(buf);
	istringstream wcs(x);
	wcs >> ra >> dec;
	str << ra << ',' << dec;
      }
      while (vertex.next());
      str << ')';
    }
  }

  listCiaoPost(str, conj, strip);
}

void Polygon::listPros(ostream& str, CoordSystem sys, SkyFrame sky,
		       SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits(1);
  Matrix mm = fwdRefMatrix();

  switch (sys) {
  case IMAGE:
  case DETECTOR:
  case AMPLIFIER:
    sys = IMAGE;
  case PHYSICAL:
    {
      listProsCoordSystem(str,sys,sky);
      str << "; " << type;
      vertex.head();
      do {
	Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys);
	str << setprecision(8) << v;
      }
      while (vertex.next());
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      listProsCoordSystem(str,sys,sky);
      str << "; " << type << ' ';

      switch (format) {
      case DEGREES:
	{
	  vertex.head();
	  do {
	    Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys,sky);
	    str << setprecision(8) << v[0] << "d " << v[1] << "d ";
	  }
	  while (vertex.next());
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  char ra[16];
	  char dec[16];

	  vertex.head();
	  do {
	    ptr->mapFromRef(vertex.current()->vector*mm,sys,sky,format,buf,64);
	string x(buf);
	istringstream wcs(x);
	    wcs >> ra >> dec;
	    if (dec[0]=='+')
	      str << ' ' << ra << ' ' << dec+1;
	    else
	      str << ' ' << ra << ' ' << dec;
	  }
	  while (vertex.next());
	}
	break;
      }
    }
  }

  listProsPost(str, strip);
}

void Polygon::listSAOtng(ostream& str, CoordSystem sys, SkyFrame sky,
			 SkyFormat format, int strip)
{
  FitsImage* ptr = parent->findFits(1);
  Matrix mm = fwdRefMatrix();
  listSAOtngPre(str, strip);

  switch (sys) {
  case IMAGE:
  case PHYSICAL:
  case DETECTOR:
  case AMPLIFIER:
    {
      str << type << '(';
      int first=1;
      vertex.head();
      do {
	if (!first)
	  str << ',';
	first=0;

	Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys);
	str << setprecision(8) << v[0] << ',' << v[1];
      }
      while (vertex.next());
      str << ')';
    }
    break;
  default:
    if (ptr->hasWCSEqu(sys)) {
      switch (format) {
      case DEGREES:
	{
	  str << type << '(';
	  int first=1;
	  vertex.head();
	  do {
	    if (!first)
	      str << ',';
	    first=0;

	    Vector v = ptr->mapFromRef(vertex.current()->vector*mm,sys,sky);
	    str << setprecision(8) << v[0] << ',' << v[1];
	  }
	  while (vertex.next());
	  str << ')';
	}
	break;
      case SEXAGESIMAL:
	{
	  char buf[64];
	  char ra[16];
	  char dec[16];

	  str << type << '(';
	  int first=1;
	  vertex.head();
	  do {
	    if (!first)
	      str << ',';
	    first=0;

	    ptr->mapFromRef(vertex.current()->vector*mm,sys,sky,format,buf,64);
	    string x(buf);
	    istringstream wcs(x);
	    wcs >> ra >> dec;
	    str << ra << ',' << dec;
	  }
	  while (vertex.next());
	  str << ')';
	}
	break;
      }

    }
  }

  listSAOtngPost(str, strip);
}

void Polygon::listSAOimage(ostream& str, int strip)
{
  FitsImage* ptr = parent->findFits(1);
  listSAOimagePre(str);
  Matrix mm = fwdRefMatrix();

  str << type << '(';
  int first=1;
  vertex.head();
  do {
    if (!first)
      str << ',';
    first=0;

    Vector v = ptr->mapFromRef(vertex.current()->vector*mm,IMAGE);
    str << setprecision(8) << v[0] << ',' << v[1];
  }
  while (vertex.next());

  str << ')';
  listSAOimagePost(str, strip);
}

