// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "framebase.h"
#include "fitsimage.h"
#include "contour.h"
#include "tkpostscript.h"
#include "util.h"

int FrameBase::postscriptProc(int prepass)
{
  TkPostscriptInfo* psInfo = (TkPostscriptInfo*)(((TkCanvas*)canvas)->psInfo);

  if (!visible)
    return TCL_OK;

  if (prepass)
    return TCL_OK;

  Vector org = psOrigin();

  // Image
  if (*channelFits) {
    // save graphics state
    Tcl_AppendResult(interp, "gsave\n", NULL);

    double scale = psInfo->scale * psResolution / 72.;
    int width = (int)(options->width*scale);
    int height = (int)(options->height*scale);

    // scale
    {
      ostringstream str;
      str << org[0] << ' ' << org[1] << " translate" << endl
	  << setprecision(12)
	  << 1/scale << ' ' << 1/scale << " scale" << endl
	  << ends;
      Tcl_AppendResult(interp, str.str().c_str(), NULL);
    }

    // image
    switch (psLevel) {
    case 1:
      switch (psColorSpace) {
      case BW:
      case GRAY:
	psLevel1(GRAY, width, height, scale);
	break;
      case RGB:
      case CMYK:
	psLevel1(RGB, width, height, scale);
	break;
      }
      break;
    case 2:
      psLevel2(psColorSpace, width, height, scale);
      break;
    }

    // restore gstate
    Tcl_AppendResult(interp, "grestore\n", NULL);
  }

  // Markers & Contours & Grids
  // clip path
  {
    ostringstream str;
    str << org[0] << ' ' << org[1] << ' ' 
    	<< options->width << ' ' << options->height
    	<< " rectclip" << endl << ends;
    Tcl_AppendResult(interp, str.str().c_str(), NULL);
  }

  switch (psLevel) {
  case 1:
    switch (psColorSpace) {
    case BW:
    case GRAY:
      psContours(GRAY);
      if (grid)
	grid->ps(GRAY);

      if (showMarkers) {
	psMarkers(&userMarkers, GRAY);
	psMarkers(&catalogMarkers, GRAY);
      }

      if (useCrosshair)
	psCrosshair(GRAY);
      break;
    case RGB:
    case CMYK:
      psContours(RGB);
      if (grid)
	grid->ps(RGB);

      if (showMarkers) {
	psMarkers(&userMarkers, RGB);
	psMarkers(&catalogMarkers, RGB);
      }

      if (useCrosshair)
	psCrosshair(RGB);
      break;
    }
    break;
  case 2:
    psContours(psColorSpace);
    if (grid)
      grid->ps(psColorSpace);

    if (showMarkers) {
      psMarkers(&userMarkers, psColorSpace);
      psMarkers(&catalogMarkers, psColorSpace);
    }

    if (useCrosshair)
      psCrosshair(psColorSpace);
    break;
  }

  return TCL_OK;
}

Matrix FrameBase::psMatrix(float psZoom, int width, int height)
{
  Matrix m;
  if (!isIIS()) {
    // Note: imageCenter() is in IMAGE coords
    Vector center = imageCenter() * imageToData;
    m = Translate(-center) * FlipY() * Translate(center);
  }

  return m *
    Translate(-cursor) *                // translate to cursor position
    wcsOrientationMatrix *              // flip x/y axis about center
    Rotate(wcsRotation) *               // rotate about center position
    orientationMatrix *                 // flip x/y axis about cursor position
    Scale(zoom_) *                      // scale about cursor position
    Scale(psZoom) *
    Rotate(rotation) *                  // rotate about cursor position

    FlipY() *
    Translate(Vector(width,height)/2.);
}

void FrameBase::psContours(PSColorSpace cs)
{
  // render back to front
  // aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->ps(cs);
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->ps(cs);
}

#ifdef _MACOSX
void FrameBase::macosxPrintCmd()
{
  if (!visible)
    return;

  // init
  macosxBegin();

  // clip rect
  XRectangle* rr = rectWidget;
  Vector v1 = Vector(rr->x, rr->y) * widgetToCanvas;
  Vector v2 = Vector(rr->x+rr->width, rr->y+rr->height) * widgetToCanvas;
  macosxClip(v1,v2-v1);

  // scale
  //  double scale = psResolution / 72.;
  double scale = 150 / 72.;
  int width = (int)(options->width*scale);
  int height = (int)(options->height*scale);

  // image
  macosx(scale, width, height, v1, v2-v1);

  // contours
  //   render back to front
  //   aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->macosx();
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->macosx();

  // grid
  if (grid)
    grid->macosx();

  // markers
  if (showMarkers) {
    macosxMarkers(&userMarkers);
    macosxMarkers(&catalogMarkers);
  }

  if (useCrosshair)
    macosxCrosshair();

  // cleanup
  macosxEnd();
}
#endif

#ifdef _WIN32
void FrameBase::win32PrintCmd()
{
  if (!visible)
    return;

  // init
  win32Begin();

  // clip rect
  XRectangle* rr = rectWidget;
  Vector v1 = Vector(rr->x, rr->y) * widgetToCanvas;
  Vector v2 = Vector(rr->x+rr->width, rr->y+rr->height) * widgetToCanvas;
  win32Clip(v1,v2-v1);

  // scale
  //  double scale = psResolution / 72.;
  // double scale = 150 / 72.;
  double scale = 1.0;
  int width = (int)(options->width*scale);
  int height = (int)(options->height*scale);

  // image
  win32(scale, width, height, v1, v2-v1);

  // contours
  //   render back to front
  //   aux contours
  Contour* ptr=auxcontours.tail();
  while (ptr) {
    ptr->win32();
    ptr=ptr->previous();
  }

  // main contour
  if (*currentcontour)
    (*currentcontour)->win32();

  // grid
  if (grid)
    grid->win32();

  // markers
  if (showMarkers) {
    win32Markers(&userMarkers);
    win32Markers(&catalogMarkers);
  }

  if (useCrosshair)
    win32Crosshair();

  // cleanup
  win32End();
}
#endif

