// Copyright (C) 1999-2008
// Smithsonian Astrophysical Observatory, Cambridge, MA, USA
// For conditions of distribution and use, see copyright notice in "copyright"

#include "gridbase.h"
#include "widget.h"
#include "rotstr.h"
#include "util.h"

GridBase* astGridPtr = NULL;

extern "C" {
  #include "ast.h"
  #include "grf.h"
}

static char* black = "black";
static char* white = "white";
static char* red = "red";
static char* green = "green";
static char* blue = "blue";
static char* cyan = "cyan";
static char* magenta = "magenta";
static char* yellow = "yellow";

Attribute::Attribute(Widget* p) : parent(p)
{
  style_ = SOLID;
  width_ = 1;

  font_ = HELVETICA;
  size_ = 10;
  tkfont_ = Tk_GetFont(parent->getInterp(), parent->getTkwin(), "helvetica 10 normal");

  colour_ = WHITE;
  colorName_ = white;
  setColour(colour_);
}

Attribute::~Attribute()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);
}

void Attribute::setStyle(double v)
{
  switch ((int)v) {
  case SOLID:
  case DASH:
    break;
  default:
    return;
  }

  style_ = (Style)((int)v);
}

void Attribute::setWidth(double v)
{
  if (v>0)
    width_ = v;
  else
    width_ = 1;
}

void Attribute::setSize(double v)
{
  if (v >= 1) {
    size_ = (int)v;
    initFont();
  }
}

void Attribute::setFont(double v)
{
  switch ((Font)((int)v)) {
  case DEFAULT:
  case TIMES:
  case HELVETICA:
  case SYMBOL:
  case COURIER:

  case DEFAULTBOLD:
  case TIMESBOLD:
  case HELVETICABOLD:
  case SYMBOLBOLD:
  case COURIERBOLD:

  case DEFAULTITALIC:
  case TIMESITALIC:
  case HELVETICAITALIC:
  case SYMBOLITALIC:
  case COURIERITALIC:
    font_ = (Font)((int)(v));
    initFont();
    return;

  default:
    return;
  }
}

void Attribute::initFont()
{
  if (tkfont_)
    Tk_FreeFont(tkfont_);
  tkfont_ = NULL;

  ostringstream str;
  switch (font_) {
  case DEFAULT:
    str << "system " << size_ << " normal" << ends;
    break;
  case TIMES:
    str << "times " << size_ << " normal" << ends;
    break;
  case HELVETICA:
    str << "helvetica " << size_ << " normal" << ends;
    break;
  case SYMBOL:
    str << "symbol " << size_ << " normal" << ends;
    break;
  case COURIER:
    str << "courier " << size_ << " normal" << ends;
    break;

  case DEFAULTBOLD:
    str << "system " << size_ << " bold" << ends;
    break;
  case TIMESBOLD:
    str << "times " << size_ << " bold" << ends;
    break;
  case HELVETICABOLD:
    str << "helvetica " << size_ << " bold" << ends;
    break;
  case SYMBOLBOLD:
    str << "symbol " << size_ << " bold" << ends;
    break;
  case COURIERBOLD:
    str << "courier " << size_ << " bold" << ends;
    break;

  case DEFAULTITALIC:
    str << "system " << size_ << " italic" << ends;
    break;
  case TIMESITALIC:
    str << "times " << size_ << " italic" << ends;
    break;
  case HELVETICAITALIC:
    str << "helvetica " << size_ << " italic" << ends;
    break;
  case SYMBOLITALIC:
    str << "symbol " << size_ << " italic" << ends;
    break;
  case COURIERITALIC:
    str << "courier " << size_ << " italic" << ends;
    break;
  default:
    return;
  }

  tkfont_ = Tk_GetFont(parent->getInterp(), parent->getTkwin(), str.str().c_str());
}

void Attribute::setColour(double v)
{
  // check for color change
  if (colour_ == (Colour)((int)(v)))
    return;

  switch ((int)v) {
  case BLACK:
    colorName_ = black;
    break;
  case WHITE:
    colorName_ = white;
    break;
  case RED:
    colorName_ = red;
    break;
  case GREEN:
    colorName_ = green;
    break;
  case BLUE:
    colorName_ = blue;
    break;
  case CYAN:
    colorName_ = cyan;
    break;
  case MAGENTA:
    colorName_ = magenta;
    break;
  case YELLOW:
    colorName_ = yellow;
    break;
  default:
    return;
  }

  color_ = parent->getColor(colorName_);
  colour_ = (Colour)((int)(v));
}

GridBase::GridBase(Widget* p, const char* o) : parent(p)
{
  option_ = dupstr(o);

  renderMode = X11;
  line = new Attribute(parent);
  text = new Attribute(parent);

  gc_ = NULL;
  pixmap_ = 0;
  byteorder_ = 0;

  mode_ = Widget::RGB;
}

GridBase::~GridBase()
{
  if (option_)
    delete [] option_;

  if (line)
    delete line;
  if (text)
    delete text;
}

// Generic

int GridBase::gAttr(int which, double value, double* old, int prim)
{
  if (value == AST__BAD) {
    if (old)
      *old = 0;
    return 1;
  }

  Attribute* attr;
  switch (prim) {
  case GRF__TEXT:
    attr = text;

    switch (which) {
    case GRF__STYLE:
      break;
    case GRF__WIDTH:
      break;
    case GRF__SIZE:
      if (old)
	*old = attr->size();
      attr->setSize(value);
      break;
    case GRF__FONT:
      if (old)
	*old = attr->font();
      attr->setFont(value);
      break;
    case GRF__COLOUR:
      if (old)
	*old = attr->colour();
      attr->setColour(value);
      break;
    }

    break;
  case GRF__LINE:
    attr = line;

    switch (which) {
    case GRF__STYLE:
      if (old)
	*old = attr->style();
      attr->setStyle(value);
      break;
    case GRF__WIDTH:
      if (old)
	*old = attr->width();
      attr->setWidth(value);
      break;
    case GRF__SIZE:
      break;
    case GRF__FONT:
      break;
    case GRF__COLOUR:
      if (old)
	*old = attr->colour();
      attr->setColour(value);
      break;
    }

    break;
  }

  return 1;
}

int GridBase::gLine(int n, float* x, float* y)
{
  switch (renderMode) {
  case X11:
    x11Line(n,x,y);
    break;
  case PS:
    psLine(n,x,y);
    break;
#ifdef _MACOSX
  case MACOSX:
    macosxLine(n,x,y);
    break;
#endif
#ifdef _WIN32
  case GWIN32:
    win32Line(n,x,y);
    break;
#endif
  }
}

int GridBase::gQch(float* chv, float* chh)
{
  if (text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);

    *chv = (float)metrics.linespace;
    *chh = (float)metrics.linespace;
    return 1;
  }
  else {
    *chv = *chh = 0;
    return 0;
  }
}

int GridBase::gText(const char* txt, float x, float y, const char* just, 
		    float upx, float upy)
{
  switch (renderMode) {
  case X11:
    x11Text(txt,x,y,just,upx,upy);
    break;
  case PS:
    psText(txt,x,y,just,upx,upy);
    break;
#ifdef _MACOSX
  case MACOSX:
    macosxText(txt,x,y,just,upx,upy);
    break;
#endif
#ifdef _WIN32
  case GWIN32:
    win32Text(txt,x,y,just,upx,upy);
    break;
#endif
  }
}

int GridBase::gTxExt(const char* txt, float x, float y, const char* just,
		 float upx, float upy, float* xb, float* yb)
{
  if (txt && text->tkfont()) {
    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    // Translate from CC to CC justification
    Matrix m1,m2;
    if (just) {
      switch (just[0]) {
      case 'T':
	m1 = Translate(0,metrics.linespace/2);
	break;
      case 'C':
	break;
      case 'B':
	m1 = Translate(0,-(metrics.ascent-metrics.linespace/2));
	break;
      case 'M':
	m1 = Translate(0,-metrics.linespace/2);
      }

      switch (just[1]) {
      case 'L':
	m2 = Translate(width/2,0);
	break;
      case 'C':
	break;
      case 'R':
	m2 = Translate(-width/2,0);
	break;
      }
    }

    Vector c = Vector(x,y) *  m1 * m2;

    BBox bb = BBox(c*Translate(-width/2,-metrics.linespace/2), 
    		   c*Translate(width/2,metrics.linespace/2));

    xb[0] = bb.ll[0];
    yb[0] = bb.ll[1];

    xb[1] = bb.ur[0];
    yb[1] = bb.ll[1];

    xb[2] = bb.ur[0];
    yb[2] = bb.ur[1];

    xb[3] = bb.ll[0];
    yb[3] = bb.ur[1];
  }
  else {
    xb[0] = xb[1] = xb[2] = xb[3] = 0;
    yb[0] = yb[1] = yb[2] = yb[3] = 0;
  }

  return 1;
}

int GridBase::gScales(float* alpha, float* beta)
{
  /*
  *alpha = float(XDisplayWidthMM(parent->getDisplay(),parent->screenNumber)) / 
    float(XDisplayWidth(parent->getDisplay(),parent->screenNumber));

  *beta = float(XDisplayHeightMM(parent->getDisplay(),parent->screenNumber)) / 
    float(XDisplayHeight(parent->getDisplay(),parent->screenNumber));
  */
  return 1;
}

int GridBase::gCap(int cap, int value)
{
  switch (cap) {
  case GRF__SCALES:
    return 0;
  case GRF__MJUST:
    return 1;
  case GRF__ESC:
    return 0;
  }
}

// X11 Render functions

int GridBase::x11Line(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  XSetForeground(parent->getDisplay(), gc_, line->color());
  int w = (int)line->width();
  if (w<1)
    w = 1;
  switch (line->style()) {
  case Attribute::SOLID:
    XSetLineAttributes(parent->getDisplay(), gc_, w, 
		       LineSolid, CapButt, JoinMiter);
    break;
  case Attribute::DASH:
    XSetLineAttributes(parent->getDisplay(), gc_, w, 
		       LineOnOffDash, CapButt, JoinMiter);
    char dlist[] = {8,3};
    XSetDashes(parent->getDisplay(), gc_, 0, dlist, 2);
    break;
  }

  for (int i=0; i<n-1; i++) {
    Vector s = Vector(x[i],y[i]).round();
    Vector e = Vector(x[i+1],y[i+1]).round();
    XDrawLine(parent->getDisplay(), pixmap_, gc_, 
	      (int)s[0],(int)s[1],(int)e[0],(int)e[1]);
  }

  return 1;
}

int GridBase::x11Text(const char* txt, float x, float y, 
		   const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    XSetFont(parent->getDisplay(), gc_, Tk_FontId(text->tkfont()));
    XSetForeground(parent->getDisplay(), gc_, text->color());

    Vector v = Vector(x,y);
    double angle = calcTextAngle(Vector(upx,upy).angle());

    if ((angle>-.01 && angle<+.01))
      drawString(txt, v, just);
    else
      drawRotString(txt, v, just, angle);
    return 1;
  }

  return 0;
}

// PS Render functions

int GridBase::psLine(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  psColor(line);
  psWidth(line);
  psStyle(line);

  for (int i=0; i<n; i++) {
    Vector v = Vector(x[i],y[i]) * matrix_;

    ostringstream str;
    if (i == 0) {
      str << "newpath " << endl;
      str << v.TkCanvasPs(parent->getCanvas()) << " moveto" << endl << ends;
    }
    else
      str << v.TkCanvasPs(parent->getCanvas()) << " lineto" << endl << ends;

    Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
  }

  ostringstream str;
  str << "stroke" << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);

  return 1;
}

int GridBase::psText(const char* txt, float x, float y, 
		const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    double angle = calcTextAngle(Vector(upx,upy).angle());

    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(text->tkfont(), &psFont);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1,m2;
    switch (just[0]) {
    case 'T':
      m1 = Translate(0,metrics.ascent);
      break;
    case 'C':
      m1 = Translate(0,metrics.ascent-metrics.linespace/2);
      break;
    case 'B':
      break;
    case 'M':
      m1 = Translate(0,-metrics.descent);
    }

    switch (just[1]) {
    case 'L':
      break;
    case 'C':
      m2 = Translate(-width/2.,0);
      break;
    case 'R':
      m2 = Translate(-width,0);
      break;
    }

    Vector v = Vector(x,y) * matrix_;
    Matrix mm = Translate(-v)*Rotate(-angle)*m1*m2*Rotate(angle)*Translate(v);
    
    ostringstream str;
    str << '/' << Tcl_DStringValue(&psFont) 
	<< " findfont " << psSize << " scalefont setfont" << endl;

    Tcl_DStringFree(&psFont);

    psColor(text);

    Vector vv = v*mm;
    str << "gsave " 
	<< vv.TkCanvasPs(parent->getCanvas()) << " moveto" << endl
      	<< radToDeg(angle) << " rotate " 
	<< '(' << psQuote(txt) << ')' << " show"
	<< " grestore" << endl << ends;

    Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
    return 1;
  }

  return 0;
}

#ifdef _MACOSX
int GridBase::macosxLine(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  macosxColor(parent->getXColor(line->colorName()));
  macosxWidth(line->width());
  switch (line->style()) {
  case Attribute::SOLID:
    macosxDash(NULL,0);
    break;
  case Attribute::DASH:
    float dlist[] = {8,3};
    macosxDash(dlist,2);
    break;
  }

  Vector v[n];
  for (int i=0; i<n; i++)
    v[i] = Vector(x[i],y[i]) * matrix_;

  macosxDrawLines(v, n);
  
  return 1;
}

int GridBase::macosxText(const char* txt, float x, float y, 
			 const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    double angle = calcTextAngle(Vector(upx,upy).angle());

    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(text->tkfont(), &psFont);

    macosxFont(Tcl_DStringValue(&psFont),psSize); 
    Tcl_DStringFree(&psFont);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1,m2;
    switch (just[0]) {
    case 'T':
      m1 = Translate(0,metrics.ascent);
      break;
    case 'C':
      m1 = Translate(0,metrics.ascent-metrics.linespace/2);
      break;
    case 'B':
      break;
    case 'M':
      m1 = Translate(0,-metrics.descent);
    }

    switch (just[1]) {
    case 'L':
      break;
    case 'C':
      m2 = Translate(-width/2.,0);
      break;
    case 'R':
      m2 = Translate(-width,0);
      break;
    }

    Vector v = Vector(x,y) * matrix_;
    Matrix mm = Translate(-v)*Rotate(-angle)*m1*m2*Rotate(angle)*Translate(v);
    
    macosxColor(parent->getXColor(text->colorName()));

    Vector vv = v*mm;
    macosxDrawText(vv, angle, txt);

    return 1;
  }

  return 0;
}
#endif

#ifdef _WIN32
int GridBase::win32Line(int n, float* x, float* y)
{
  if (n<2 || !x || !y)
    return 1;

  win32Color(parent->getXColor(line->colorName()));
  win32Width(line->width());
  switch (line->style()) {
  case Attribute::SOLID:
    win32Dash(NULL,0);
    break;
  case Attribute::DASH:
    float dlist[] = {8,3};
    win32Dash(dlist,2);
    break;
  }

  Vector v[n];
  for (int i=0; i<n; i++)
    v[i] = Vector(x[i],y[i]) * matrix_;

  win32DrawLines(v, n);
  
  return 1;
}

int GridBase::win32Text(const char* txt, float x, float y, 
			 const char* just, float upx, float upy)
{
  if (txt && txt[0] && text->tkfont() && just) {
    double angle = calcTextAngle(Vector(upx,upy).angle());

    Tcl_DString psFont;
    Tcl_DStringInit(&psFont);
    int psSize = Tk_PostscriptFontName(text->tkfont(), &psFont);

    win32Font(Tcl_DStringValue(&psFont),psSize); 
    Tcl_DStringFree(&psFont);

    Tk_FontMetrics metrics;
    Tk_GetFontMetrics(text->tkfont(), &metrics);
    int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

    Matrix m1,m2;
    switch (just[0]) {
    case 'T':
      m1 = Translate(0,metrics.ascent);
      break;
    case 'C':
      m1 = Translate(0,metrics.ascent-metrics.linespace/2);
      break;
    case 'B':
      break;
    case 'M':
      m1 = Translate(0,-metrics.descent);
    }

    switch (just[1]) {
    case 'L':
      break;
    case 'C':
      m2 = Translate(-width/2.,0);
      break;
    case 'R':
      m2 = Translate(-width,0);
      break;
    }

    Vector v = Vector(x,y) * matrix_;
    Matrix mm = Translate(-v)*Rotate(-angle)*m1*m2*Rotate(angle)*Translate(v);
    
    win32Color(parent->getXColor(text->colorName()));

    Vector vv = v*mm;
    win32DrawText(vv, angle, txt);

    return 1;
  }

  return 0;
}
#endif

// X11 support

void GridBase::drawString(const char* txt, Vector v, const char* just)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));

  // Translate from CC to left,baseline justification
  Matrix m1,m2;
  switch (just[0]) {
  case 'T':
    m1 = Translate(0,metrics.ascent);
    break;
  case 'C':
    m1 = Translate(0,metrics.ascent-metrics.linespace/2);
    break;
  case 'B':
    break;
  case 'M':
    m1 = Translate(0,-metrics.descent);
  }

  switch (just[1]) {
  case 'L':
    break;
  case 'C':
    m2 = Translate(-width/2,0);
    break;
  case 'R':
    m2 = Translate(-width,0);
    break;
  }

  Vector c = v *  m1 * m2;

  XDrawString(parent->getDisplay(), pixmap_, gc_,
	      (int)c[0], (int)c[1], txt, strlen(txt));
}

void GridBase::drawRotString(const char* txt, Vector v, 
			 const char* just, double angle)
{
  Tk_FontMetrics metrics;
  Tk_GetFontMetrics(text->tkfont(), &metrics);
  int width = Tk_TextWidth(text->tkfont(), txt, strlen(txt));
  int height = metrics.linespace;

  Matrix m1,m2;
  switch (just[0]) {
  case 'T':
    m1 = Translate(0,metrics.ascent);
    break;
  case 'C':
    m1 = Translate(0,metrics.ascent-metrics.linespace/2);
    break;
  case 'B':
    break;
  case 'M':
    m1 = Translate(0,-metrics.descent);
  }

  switch (just[1]) {
  case 'L':
    break;
  case 'C':
    m2 = Translate(-width/2.,0);
    break;
  case 'R':
    m2 = Translate(-width,0);
    break;
  }

  Matrix mm = Translate(-v)*Rotate(-angle)*m1*m2*Rotate(angle)*Translate(v);
  XDrawRotString(parent->getDisplay(), pixmap_, gc_, text->tkfont(), 
		 angle, v*mm, txt);
}

// PS Support

void GridBase::psColor(Attribute* attr)
{
  ostringstream str;
  switch ((Widget::PSColorSpace)mode_) {
  case Widget::BW:
  case Widget::GRAY:
    psColorGray(parent->getXColor(attr->colorName()), str);
    str << " setgray";
    break;
  case Widget::RGB:
    psColorRGB(parent->getXColor(attr->colorName()), str);
    str << " setrgbcolor";
    break;
  case Widget::CMYK:
    psColorCMYK(parent->getXColor(attr->colorName()), str);
    str << " setcmykcolor";
    break;
  }
  str << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
}

void GridBase::psWidth(Attribute* attr)
{
  ostringstream str;
  str << attr->width() << " setlinewidth" << endl << ends;
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
}

void GridBase::psStyle(Attribute* attr)
{
  ostringstream str;
  switch (attr->style()) {
  case Attribute::SOLID:
    str << "[] 0 setdash" << endl << ends;
    break;
  case Attribute::DASH:
    str << "[8 3] 0 setdash" << endl << ends;
    break;
  }
  Tcl_AppendResult(parent->getInterp(), str.str().c_str(), NULL);
}

double GridBase::calcTextAngle(double a)
{
  // our angle is 90 off from ast's, and the other direction
  double r = -(a - M_PI/2);

  // and add any image rotation

  // normalize
  if (r>0)
    while (r>2*M_PI)
      r -= 2*M_PI;
  else 
    while (r<0)
      r += 2*M_PI;

  // and flip if upside down
  // we only want to do this if the number is really, really upside down
  // you might have the case of a curved vertical line
  if (r > M_PI/2 && r <= 3*M_PI/2)
    r -= M_PI;

  return r;
}

int GridBase::matrixMap(void* fs, Matrix& mx, const char* str)
{
  AstFrameSet* frameSet = (AstFrameSet*)fs;

  double ss[] = {mx.matrix(0,0),mx.matrix(1,0),mx.matrix(0,1),mx.matrix(1,1)};
  double tt[] = {mx.matrix(2,0),mx.matrix(2,1)};

  AstMatrixMap* mm;
  if (!(mm= astMatrixMap(2, 2, 0, ss, "")))
    return 0;

  AstShiftMap* sm;
  if (!(sm = astShiftMap(2, tt, "")))
    return 0;

  AstCmpMap* cmp;
  if (!(cmp = astCmpMap(mm, sm, 1, "")))
    return 0;

  astAddFrame(frameSet, AST__CURRENT, cmp, astFrame(2, str));
}

